/*
 * tserver.c: test program
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * $Id: tserver.c 208 2011-08-03 17:45:07Z meaton $
 */

#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

// QNX includes
#ifndef USE_FC
#	include <sys/kernel.h>
#	include <sys/proxy.h>
#	include <sys/name.h>
#else
#	include <fcqnx.h>
#endif

#define debug 0

int		count	= 0;
int		size	= 4;
int		rtime	= 10;
int		slow	= 0;
//int		debug	= 0;
int*	msg;
int		child	= -1;
int		parent	= -1;

static int		debug_was_sent = 0;

int timedout;
int reaped;

/*
 * sleep() and alarm() are not safe to use together, so this is a different
 * kind of sleep
 */
void Sleep(int usec)
{
	usleep(usec);
}

void Kill(pid_t pid, int signo)
{
	if(debug)
		printf("%d Kill(pid %d, signo %d)\n", getpid(), pid, signo);
	
	if(kill(pid, signo) == -1) {
		printf("kill() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	}
}


void Timer(int rtime)
{
	timedout = 0;
	alarm(rtime);
}

void SigNullHandler(int signo)
{
	if(debug)
		printf("%d SigNullHandler(%d)\n", getpid(), signo);
}

void SigAlarmHandler(int signo)
{
	signo = signo;

	if(debug)
		printf("%d SigAlarmHandler()\n", getpid());

	timedout = 1;
}

void SigChildHandler(int signo)
{
	int status;
	pid_t pid = wait(&status);

	signo = signo;

	assert(parent == getpid());

	if(pid == -1) {
		printf("wait() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	}
	if(WIFEXITED(status)) {
		printf("unexpected child death: pid %d exit status %d\n",
			pid, WEXITSTATUS(status));
		exit(1);
	}
	assert(WIFSIGNALED(status));

	if(WTERMSIG(status) != SIGKILL) {
		printf("unexpected child death: pid %d signal %d\n",
			pid, WTERMSIG(status));
		exit(1);
	}

	/* otherwise we commited infanticide */
	if(debug)
		printf("%d SigChildHandler() pid %d died\n", getpid(), pid);
	reaped = 1;
}

void AtExit(void)
{
	if(parent != getpid()) {
		if(debug)
			printf("%d AtExit() parent %d\n", getpid(), getppid());
		return;
	}

	if(debug)
		printf("%d AtExit() child %d\n", getpid(), child);

	if(child != -1) {
		signal(SIGHUP, SIG_IGN);
		Kill(child, SIGKILL);
	}
	child = -1;
}

/*
 * test send/receive/reply
 */
int Receiver(int* msg, size_t size)
{
	pid_t client = getpid();
	pid_t pid = fork();

//	printf ("My pid is %d\n", getpid());
//	sleep (10);
	
	if(pid == -1) {
		printf("fork() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	} else if(pid > 0) {
		return pid;
	}

	if(Send(getppid(), 0, 0, 0, 0) == -1) {
		printf("%d Receiver initial Send(%d) failed: [%d] %s\n",
			getpid(), getppid(), errno, strerror(errno));
		exit(1);
	}

//	printf ("Receive size is %d\n", size);
	
	while(1) {
		if((client = Receive(0, msg, size+4)) == -1) {
			if (errno == ESRCH) continue;
			printf("%d Receiver Receive(0) failed: [%d] %s\n",
				getpid(), errno, strerror(errno));
			exit(1);
		}
//		if (*msg != size)
//			printf ("Got message of %d, expect %d\n", *msg, size);
//		else
//			printf ("Got message with size %d\n", *msg);

		if(Reply(client, msg, *msg) == -1) {
			/* not a fatal error, client timedout during a send
			printf("%d Receiver Reply(%d) failed: [%d] %s\n",
				getpid(), client, errno, strerror(errno));
			*/
		}
	}
}

void TestSrr()
{
	int tsize = size;

	child = Receiver(msg, tsize);

	if (debug)
		printf ("S/R/R test: wait for child to come ready\n");
	
	// wait for child to start and register with module
	if(Receive(0, 0, 0) == -1) {
		printf("Receive() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	}
	Reply(child, 0, 0);

	if(debug)
		printf("S/R/R test: parent/client %d child/server %d\n",
			getpid(), child);

	if (debug)
		printf ("Child pid is %d\n", child);
	
	while(1) {
		count = 0;

		Timer(rtime);

		while(!timedout) {
			*msg = tsize; // set Reply() size

			if(Send(child, msg, msg, tsize, tsize) == -1 && errno != EINTR) {
				printf("srr Send() failed: [%d] %s\n", errno, strerror(errno));
				exit(1);
			}
			count++;
			if ((count > 1) && debug_was_sent)
			{
				SrrDebug (2);
				debug_was_sent = 0;
			}
			
			
			if(slow)
				Sleep(slow);
		}

		/* Looking for a bug here.  If we turned on debugging last
		   time, then we need to sleep here to allow the user to look
		   at the state of the SRR module */
		if (debug_was_sent)
		{
			char	tempbuf[10];
			
			printf ("Blocked so you can debug. Press return to continue");
			fflush (stdout);
			scanf ("%s", tempbuf);
		}
		/* Looking for a bug here.  Turn on debugging output if we see
		   a count of less than 2. */
		if (count < 2)
		{
			SrrDebug (3);
			debug_was_sent = 1;
		}
		
		printf("  srr: %4u bytes %7u times %3u sec %7u Hz %11u bytes/sec\n",
			tsize, count, rtime, count/rtime, 2*(count/rtime)*tsize);
			
		if(tsize == 0)
			break;

		tsize /= 4;
	}

	Kill(child, SIGKILL);
	
	child = -1;

	while(!reaped)
		Sleep(1); // wait for child to die
	reaped = 0;
}

/*
 * test polling
 */
int PollReceiver(int* msg, size_t size)
{
	pid_t client = getpid();
	pid_t pid = fork();
	fd_set readfds;
	int ndesc;

	if(debug) {
		printf("fork: child %d parent %d\n", getpid(), getppid());
	}

	if(pid == -1) {
		printf("fork() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	} else if(pid > 0) {
		return pid;
	}

	if(Send(getppid(), 0, 0, 0, 0) == -1) {
		printf("%d Send(%d) failed: [%d] %s\n",
			getpid(), getppid(), errno, strerror(errno));
		exit(1);
	}

	while(1) {
		FD_ZERO (&readfds);
		FD_SET (SrrFd(), &readfds);
		ndesc = select (SrrFd() + 1, &readfds, NULL, NULL, NULL);
		if (ndesc > 0)
		{
			if((client = Receive(0, msg, size)) == -1) {
				printf("%d Receive(0) failed: [%d] %s\n",
					   getpid(), errno, strerror(errno));
				exit(1);
			}
			if(Reply(client, msg, *msg) == -1 && errno != ESRCH) {
				printf("%d Reply(%d) failed: [%d] %s\n",
				   getpid(), client, errno, strerror(errno));
				exit(1);
			}
		}
		else if (ndesc == 0)
		{
			printf ("select() timed out: not possible\n");
		}
		else
		{
			printf ("select() failed: [%d] %s\n", errno, strerror (errno));
			exit(1);
		}
	}
}

void TestPoll()
{
	int			tsize = size, sender;
	
	child = PollReceiver(msg, tsize);

	// wait for child to start and register with module
	if((sender = Receive(0, 0, 0)) == -1) {
		printf("Receive() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	}
	if (sender != child)
	{
		printf ("Got a message from %d when expecting child %d\n",
				sender, child);
	}
	Reply(child, 0, 0);

	if(debug)
		printf("S/R/R poll test: parent/client %d child/server %d\n",
			getpid(), child);

	while(1) {
		*msg = tsize; // set Reply() size

		count = 0;

		Timer(rtime);

		while(!timedout) {
			if(Send(child, msg, msg, tsize, tsize) == -1 && errno != EINTR) {
				printf("poll Send() failed: [%d] %s\n", errno, strerror(errno));
				printf("count currently %d\n", count);
				exit(1);
			}
			count++;
			
			if(slow)
				Sleep(slow);
		}

		printf(" poll: %4u bytes %7u times %3u sec %7u Hz %11u bytes/sec\n",
			tsize, count, rtime, count/rtime, 2*(count/rtime)*tsize);
			
		if(tsize == 0)
			break;

		tsize /= 4;
	}

	Kill(child, SIGKILL);
	
	child = -1;

	while(!reaped)
		Sleep(1); // wait for child to die
	reaped = 0;
}

/*
 * test trigger
 */
#ifndef USE_FC

int Triggerer(void)
{
	pid_t pid = fork();
	pid_t proxy = -1;
	
	if(pid == -1) {
		printf("fork() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	} else if(pid > 0) {
		return pid;
	}
	
	if(Send(getppid(), 0, 0, 0, 0) == -1) {
		printf("%d Send(%d) failed: [%d] %s\n",
			getpid(), getppid(), errno, strerror(errno));
		exit(1);
	}
	//Reply(child, 0, 0);

	if((proxy = qnx_proxy_attach(getppid(), 0, 0, -1)) == -1) {
		printf("%d qnx_proxy_attach(%d) failed: [%d] %s\n",
			getpid(), getppid(), errno, strerror(errno));
		exit(1);
	}
	
	while(1) {
		if(Receive(0, 0, 0) == -1) {
			if(errno != EINTR) {
				printf("%d Receive(%d) failed: [%d] %s\n",
					getpid(), 0, errno, strerror(errno));
				exit(1);
			}
			continue;
		}
		if(Trigger(proxy) == -1) {
			/* not a fatal error, client timedout during a send
			printf("%d Trigger(%d) failed: [%d] %s\n",
				getpid(), proxy, errno, strerror(errno));
			*/
		}
	}
}

void TestTrigger()
{
	pid_t proxy;
	int		nextra=0;

	child = Triggerer();

	// wait for child to start and register with module
	if(Receive(0, 0, 0) == -1) {
		printf("Receive() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	}
	Reply(child, 0, 0);

	if((proxy = qnx_proxy_attach(child, 0, 0, -1)) == -1) {
		printf("qnx_proxy_attach(%d) failed: [%d] %s\n",
			child, errno, strerror(errno));
		exit(1);
	}

	if(debug)
		printf("Trigger test: parent/client %d proxy %d child/server %d\n",
			getpid(), proxy, child);

	count = 0;
	Timer(rtime);

	while(!timedout) {
		if(Trigger(proxy) == -1) {
			if(errno == EINTR)
				continue;

			printf("Trigger(%d) failed: [%d] %s\n",
				proxy, errno, strerror(errno));
			exit(1);
		}
		if(Receive(0, 0, 0) == -1 && errno != EINTR) {
			printf("Receive(%d) failed: [%d] %s\n",
				0, errno, strerror(errno));
			exit(1);
		}
		count += 2;

		if(slow)
			Sleep(slow);
	}
			
	printf(" proxy: %6u times %3u sec %5u Hz\n", count, rtime, count/rtime);

	Kill(child, SIGKILL);
	child = -1;

	while(!reaped)
		Sleep(1); // wait for child to die 
	reaped = 0;

	while (Creceive(0,0,0) != -1) // Receive possible pending proxy
		nextra++;
	//	if (nextra)
	//		printf ("Received %d extra proxies\n", nextra);
}	
#endif

/*
 * test timers: use signals or proxies
 */

void TestTimer(int);

void TestTimedProxy()
{
	pid_t proxy = qnx_proxy_attach(0, 0, 0, -1);
	if(proxy == -1) {
		printf ("qnx_proxy_attach() failed: [%d] %s\n", errno, strerror(errno));
		exit (1);
	}

	TestTimer(-proxy);

	if(qnx_proxy_detach(proxy) == -1) {
		printf ("qnx_proxy_detach()) failed: [%d] %s\n", errno, strerror(errno));
		// exit(1); not yet implemented...
	}
}

void TestTimedSignal()
{
	TestTimer(SIGUSR1);
}

void TestTimer(int signo)
{
	struct itimerspec	value;
	struct itimerspec	ovalue;
	timer_t				tid;
	struct sigevent		event;
	const char*			test = "proxy";

	count = 0;

	event.sigev_signo = signo;

	if(signo >= 0)
	{
		signal(signo, SigNullHandler);
		test = "signal";
	}

	if ((tid = timer_create (CLOCK_REALTIME, &event)) == -1)
	{
		printf ("timer_create() failed: [%d] %s\n", errno, strerror(errno));
		exit (1);
	}

	value.it_value.tv_sec = 0;
	value.it_value.tv_nsec = 1;
	value.it_interval.tv_sec = 0;
	value.it_interval.tv_nsec = 1;

	if(timer_settime (tid, TIMER_ADDREL, &value, &ovalue) == -1)
	{
		printf ("timer_settime() failed: [%d] %s\n", errno, strerror(errno));
		exit (1);
	}

	Timer (rtime);
	
	while(!timedout) {
		if(Receive(0, 0, 0) == -1)
		{
			if (errno != EINTR)
			{
				printf("Receive() failed: [%d] %s\n", errno, strerror(errno));
				exit(1);
			}
		}
		count += 1;
	}

	if(timer_delete(tid) == -1)
	{
		printf("timer_delete() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	}

	printf("timer w/%s: %6u times, %3u sec, %5u Hz, period %g us\n",
		   test, count, rtime, count/rtime, 1000000 * (double)rtime/(double)count );
}

void EmitClockResolution(void)
{
	struct timespec ts;
	if (SrrClockGetres (CLOCK_REALTIME, &ts) != -1)
	{
		printf ("Clock resolution: %ld sec, %ld nsec (%g Hz)\n",
				ts.tv_sec, ts.tv_nsec,
				1.0E9/(ts.tv_sec * 1.0E9 + ts.tv_nsec));
	}
	else
	{
		printf ("Error reading clock resolution: %s\n", strerror(errno));
	}
}

char USAGE[] =
	"usage: tserver [-h] [-s size] [-t sec] [-d] [-D usec] [-T m|p|s|t|k]\n";

char HELP[] =
	"    -h    print this helpful message\n"
	"    -s    the size to use for msg passing\n"
	"    -t    the time (in seconds) to perform the test for\n"
	"    -d    debug output\n"
	"    -D    micro-seconds of delay between each test loop (default is none)\n"
	"    -T    test only\n"
	"            message passing: 'm',\n"
	"            proxy triggers: 'p',\n"
	"            select: 's',\n"
	"            timer proxies: 't'\n"
	"            timer signals: 'k'\n"
	"            \n"
	" The test server for the QNX-style API loadable module\n"
	;

int main(int argc, char* argv[])
{
	int		opt		= 0;
	char	tests[16]	= { 0 };
	int		ntests = 0;
  int meaningless;

	setvbuf(stdout, 0, _IOLBF, 0);

	parent = getpid();

	while((opt = getopt(argc, argv, "uhs:t:d:D:T:")) != -1) {
		switch(opt) {
		case 'h':
			printf("%s%s", USAGE, HELP);
			exit(0);

		case 's':
			size = atoi(optarg);
			break;

		case 't':
			rtime = atoi(optarg);
			break;

		case 'd':
			meaningless = atoi(optarg);
			break;

		case 'D':
			slow = atoi(optarg);
			break;

		case 'T':
			if(*optarg == 'a')
				ntests = 0;
			else
				tests[ntests++] = *optarg;
			break;

		case 'u':
			SrrUnload();
			exit (0);
			break;

		default:
			break;
		}
	}
	
	/* register S/R/R - NOTE: this is unnecessary, but useful for debugging */
#ifdef SRR_SRR_H
	if(SrrReg() == -1) {
		printf("SrrReg() failed: [%d] %s\n", errno, strerror(errno));
		exit(1);
	}
#endif

	/* alloc message buffer */
	msg = (int*) malloc(size);

	if(!msg) {
		printf("malloc() failed\n");
		exit(1);
	}
	memset(msg, 0, size);

	/* set up signal handling */
	signal(SIGALRM, SigAlarmHandler);
	signal(SIGCHLD, SigChildHandler);
	atexit(AtExit);

	SrrDebug (2);

	if(ntests == 0 || strchr (tests, 'm'))
		TestSrr();

#ifndef USE_FC
	if(ntests == 0 || strchr (tests, 'p'))
		TestTrigger();

	if(ntests == 0 || strchr (tests, 's'))
		TestPoll();

	if(ntests == 0 || strchr (tests, 't') || strchr (tests, 'k'))
		EmitClockResolution();
	
	if(ntests == 0 || strchr (tests, 't'))
		TestTimedProxy();

	if(ntests == 0 || strchr (tests, 'k'))
		TestTimedSignal();

#endif

	return(0);
}

