/*
 * libsrr.c: implementation of send/receive/reply IPC library calls.
 *
 *  Linux Loadable Module Implementation of Send/Receive/Reply
 *	message passing and related functions, similar to what is used
 *  by QNX4.
 *
 *	QNX is a trademark of QNX Software Systems Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; see the file COPYING.LIB.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * $Id: srr_lib.c 213 2011-08-16 15:07:17Z meaton $
 */

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <inttypes.h>
#ifdef SRR_USE_PTHREADS
#include <pthread.h>
#endif
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

#include "srr.h"
#include "srr_if.h"

static int global_fd = -1;
#ifdef SRR_USE_PTHREADS
static pthread_key_t pthread_fd_key = -1;
#endif

#ifdef SRR_USE_PTHREADS
static int using_pthreads = 1;
#else
static int using_pthreads = 0;
#endif

char *SrrDevice = "/dev/srripc";

#define SRR_REG  if(SrrLocalFd() == -1 && SrrReg() == -1) return -1;

/* We can use Sendmx, Replymx and Receivemx to implement
   Send, Reply and Receive.  It turns out to have about a 12% impact
   on performance! */
#undef SEND_AS_MX


/* We can set to use pthreads, but only once.  If pthreads are
   enabled, they cannot be disabled. */

int SrrLocalFd ()
{
#ifdef SRR_USE_PTHREADS
	if (using_pthreads)
	{
		int		*data = (int*)pthread_getspecific(pthread_fd_key);
		return (data ? *data : -1);
	}
	else
#endif
		return (global_fd);
}

#ifdef SRR_USE_PTHREADS
static void SrrFreeThreadFd (void* ptr)
{
	close (*(int*)ptr);
	free (ptr);
}
#endif

int SrrUsePthreads (int enable)
{
#ifdef SRR_USE_PTHREADS
	if (enable)
	{
		if (pthread_fd_key == -1)
		{
			int	result;
			
			result = pthread_key_create(&pthread_fd_key, SrrFreeThreadFd);
			if (result == 0)
				using_pthreads = 1;
		}
	}
	else
#endif
		pthread_fd_key = 0;

	return (using_pthreads);
}

int SrrSurviveExec (void)
{
	int			fd;
	
	/* Make a call to ensure that the FD we use is correct for
	   this thread/process. */
	SrrGetpid();

	/* Get this FD */
	fd = SrrFd();
	
	return (fcntl(fd, F_SETFD, 0));
}

int SrrReg(void)
{
	char		*device;
	int			fd;
#ifdef SRR_USE_PTHREADS
	int			*ptr = NULL;
#endif

#ifdef SRR_USE_PTHREADS
	if (pthread_fd_key == -1)
	{
		SrrUsePthreads (using_pthreads);
	}
	
	if (using_pthreads)
	{
		if (!(ptr = pthread_getspecific (pthread_fd_key)))
		{
			int		result;
			
			ptr = (int*) malloc (sizeof(int));
			*ptr = -1;
			result = pthread_setspecific (pthread_fd_key, ptr);
		}
		else if (*ptr != -1)
			return (0);
	}
	else
#endif
		if (global_fd != -1)
	{
		return 0;
	}

	if ((device = getenv ("SRRDEV")))
		SrrDevice = device;
	
	fd = open(SrrDevice, O_RDWR);
	if (fd != -1)
	{
		/*
		 * Upon further consideration, this is wrong. Execing shouldn't
		 * unblock somebody who is blocked on a send to us, we're the same
		 * process still.  OK - that's true, but most people would prefer
		 * that the FD no survive the exec.  It causes problems with closing
		 * the connection in the kernal.  The child can hold open names and
		 * resources.  Make it default to closing on exec().  If people want
		 * otherwise, they can call SrrSurviveExec() for each thread using
		 * SRR.
		 */

		if(fcntl(fd, F_SETFD, FD_CLOEXEC) == -1) {
			close(fd);
			fd = -1;
		}
	}

#ifdef SRR_USE_PTHREADS
	if (using_pthreads)
	{
		*ptr = fd;
	}
	else
#endif
		global_fd = fd;
	
	return (fd == -1) ? -1 : 0;
}

int SrrUnreg(void)
{
	int err = 0;

#ifdef SRR_USE_PTHREADS
	if (using_pthreads)
	{
		int		*ptr;
		if ((ptr = pthread_getspecific (pthread_fd_key)) &&
			*ptr != -1)
		{
			err = close(*ptr);
			*ptr = -1;
		}
	}
	else
#endif
	{
		if(global_fd == -1)
			return 0;
		err = close(global_fd);
		global_fd = -1;
	}
	return err;
}

static int DoIt(int ioc, void *s, int uninterruptible)
{
	int	err = -1;

	do
	{
		if (SrrLocalFd() != -1 || SrrReg() != -1)
		{
			do
			{
				err = ioctl (SrrLocalFd(), ioc, s);
      }
			while ((uninterruptible && err == -1 && errno == EINTR));
		}
		if (err == -1 && errno == EBUSY)
		{
			SrrUnreg();
			errno = EBUSY;
		}
	} while (err == -1 && errno == EBUSY);
	
	/* ioctl set errno on failure and returns -1.
	   It returns a non-negative value on success.  We force this
	   to -1 or 0.
	   2002/03/14: That's just stupid.  We need this output to
	   get the correct PID, etc. back from some functions */
	// if (err > 0)
	//     err = 0;
    return err;
}

int SrrUnload()
{
	int err;

	SRR_REG
	
	err = ioctl(SrrLocalFd(), SRR_IOCUNLOAD);

	SrrUnreg();

	return err;
}

int SrrFd(void)
{
	SRR_REG

	return SrrLocalFd();
}

int SrrDebug(int level)
{
	struct SrrTraceArgs s;

	s.level = level;
    return DoIt(SRR_IOCTRACE, &s, 1);
}

int SrrKill(pid_t pid, int signo)
{
	SrrKillArgs s;

	s.pid	= pid;
	s.signo	= signo;
    return DoIt(SRR_IOCKILL, &s, 1);
}

/*
 * Note that these will fail with EBUSY if SrrReg() can't be
 * called successfully.
 */
int SrrSend(pid_t pid, void* smsg, void* rmsg, size_t ssize, size_t rsize)
{
#ifdef SEND_AS_MX
	struct _mxfer_entry smsgmx;
	struct _mxfer_entry rmsgmx;

    smsgmx.addr = smsg;
    smsgmx.mxfer_len = ssize;
    rmsgmx.addr = rmsg;
    rmsgmx.mxfer_len = rsize;
    return Sendmx(pid, 1, 1, &smsgmx, &rmsgmx);
#else
	struct SrrSendArgs s;
	int		result;

	s.pid	= pid;
	s.ssize	= ssize;
	s.rsize	= rsize;
	s.smsg	= smsg;
	s.rmsg	= rmsg;

	result = DoIt(SRR_IOCSEND, &s, 0);
	if (result != -1)
	  result = s.rsize;
	return (result);
#endif
}

sendTimeoutStruct SrrSendTimeout(pid_t pid, void* smsg, void* rmsg, size_t ssize, size_t rsize, unsigned long time)
{
#ifdef SEND_AS_MX
	struct _mxfer_entry smsgmx;
	struct _mxfer_entry rmsgmx;

    smsgmx.addr = smsg;
    smsgmx.mxfer_len = ssize;
    rmsgmx.addr = rmsg;
    rmsgmx.mxfer_len = rsize;
    return Sendmx(pid, 1, 1, &smsgmx, &rmsgmx);
#else
	struct SrrSendTimeoutArgs s;
	int		result;

	s.pid	= pid;
	s.ssize	= ssize;
	s.rsize	= rsize;
	s.smsg	= smsg;
	s.rmsg	= rmsg;
  s.time  = time;

	result = DoIt(SRR_IOCSENDTIME, &s, 0);
	if (result != -1)
	result = s.rsize;
	sendTimeoutStruct ret;
  ret.rsize = result;
  ret.time = s.time;
  return (ret);
#endif
}


int SrrSendmx(pid_t pid, size_t sparts, size_t rparts,
			  struct _mxfer_entry *smsg, struct _mxfer_entry *rmsg)
{
	struct SrrSendmxArgs s;
	int		result, i;

	s.pid	 	= pid;
	s.sparts	= sparts;
	s.rparts	= rparts;
	s.smsgmx	= smsg;
	s.rmsgmx	= rmsg;

	result = DoIt(SRR_IOCSENDMX, &s, 0);
	if (result != -1)
	  {
		result = 0;
		for (i=0; i<rparts; i++)
		  result += rmsg[i].mxfer_len;
	  }
	return (result);
}

sendTimeoutStruct SrrSendmxTimeout(pid_t pid, size_t sparts, size_t rparts,
			  struct _mxfer_entry *smsg, struct _mxfer_entry *rmsg,
        unsigned long time)
{
	struct SrrSendmxTimeoutArgs s;
	int		result, i;

	s.pid	 	= pid;
	s.sparts	= sparts;
	s.rparts	= rparts;
	s.smsgmx	= smsg;
	s.rmsgmx	= rmsg;
  s.time    = time;
	result = DoIt(SRR_IOCSENDMXTIME, &s, 0);
	if (result != -1)
	  {
		result = 0;
		for (i=0; i<rparts; i++)
		  result += rmsg[i].mxfer_len;
	  }
  sendTimeoutStruct ret;
  ret.rsize = result;
  ret.time = s.time;
	return (ret);
}

pid_t SrrReceive(pid_t pid, void* msg, size_t *size)
{
	pid_t		result;
   struct timespec __attribute__( (aligned) )pause;
   int     count = 0;
   
#ifdef SEND_AS_MX
	struct _mxfer_entry msgmx;

    msgmx.addr = msg;
    msgmx.mxfer_len = *size;
    result = Receivemx(pid, 1, &msgmx);
	if (result != -1)
	  *size = msgmx.mxfer_len;
#else
	struct SrrRecvArgs s;

	do
	{
		s.pid		= pid;
		s.nowait	= 0;
		s.msg		= msg;
		s.size		= *size;

		result = DoIt(SRR_IOCRECV, &s, 0);
      if (( result == -1) &&  (errno == EAGAIN ) && ( count++> 5 ))
      {                                                                                           
         pause.tv_sec = 0L;
         pause.tv_nsec = (10*1000000L);
         nanosleep( &pause, NULL );
      }
      
	} while ((result == -1) && (errno == EAGAIN) && (count < 100));
	
	if (result != -1)
	  *size = s.size;
#endif
	return (result);
}

receiveTimeoutStruct SrrReceiveTimeout(pid_t pid, void* msg, size_t *size, unsigned long time)
{
	pid_t		result;
   struct timespec __attribute__( (aligned) )pause;
   int     count = 0;
   
#ifdef SEND_AS_MX
	struct _mxfer_entry msgmx;

    msgmx.addr = msg;
    msgmx.mxfer_len = *size;
    result = Receivemx(pid, 1, &msgmx);
	if (result != -1)
	  *size = msgmx.mxfer_len;
#else
	struct SrrRecvTimeoutArgs s;

	do
	{
		s.pid		= pid;
		s.nowait	= 0;
		s.msg		= msg;
		s.size		= *size;
		s.time	= time;

		result = DoIt(SRR_IOCRECVTIME, &s, 0);
      if (( result == -1) &&  (errno == EAGAIN ) && ( count++> 5 ))
      {                                                                                           
         pause.tv_sec = 0L;
         pause.tv_nsec = (10*1000000L);
         nanosleep( &pause, NULL );
      }
      
	} while ((result == -1) && (errno == EAGAIN) && (count < 100));
	
	if (result != -1)
	  *size = s.size;
#endif
	receiveTimeoutStruct ret;
  ret.pid = result;
  ret.time = s.time;
  
  return (ret);
}

int SrrCreceive(pid_t pid, void* msg, size_t *size)
{
	int		result;
   struct timespec __attribute__( (aligned) )pause;
   int     count = 0;

#ifdef SEND_AS_MX
	struct _mxfer_entry msgmx;
	
    msgmx.addr = msg;
    msgmx.mxfer_len = *size;
    result = Creceivemx(pid, 1, &msgmx);
	if (result != -1)
	  *size = msgmx.mxfer_len;
#else
	struct SrrRecvArgs s;

	do
	{
		s.pid		= pid;
		s.nowait	= 1;
		s.msg		= msg;
		s.size		= *size;

		result = DoIt(SRR_IOCRECV, &s, 0);
      if (( result == -1) &&  (errno == EAGAIN ) && ( count++> 5 ))
      {                                                                                           
         pause.tv_sec = 0L;
         pause.tv_nsec = (10*1000000L);
         nanosleep( &pause, NULL );
      }
      
	} while ((result == -1) && (errno == EAGAIN) && (count < 100));
		
	if (result != -1)
	  *size = s.size;
#endif
	return (result);
}

receiveTimeoutStruct SrrReceivemxTimeout(pid_t pid, size_t parts, struct _mxfer_entry *msgmx, unsigned long time)
{
	struct SrrRecvmxTimeoutArgs s;
	pid_t		result;
   struct timespec __attribute__( (aligned) )pause;
   int     count = 0;

	do
	{
		s.pid		= pid;
		s.nowait	= 0;
		s.parts		= parts;
		s.msgmx		= msgmx;
    s.time    = time;

		result = DoIt(SRR_IOCRECVMXTIME, &s, 0);
      if (( result == -1) &&  (errno == EAGAIN ) && ( count++> 5 ))
      {                                                                                           
         pause.tv_sec = 0L;
         pause.tv_nsec = (10*1000000L);
         nanosleep( &pause, NULL );
      }
 
   } while ((result == -1) && (errno == EAGAIN) && (count < 100));
  receiveTimeoutStruct ret;
  ret.pid = result;
  ret.time = s.time;
	return (ret);
}

/* John Green 7/10/2000 */
pid_t SrrReceivemx(pid_t pid, size_t parts, struct _mxfer_entry *msgmx)
{
	struct SrrRecvmxArgs s;
	pid_t		result;
   struct timespec __attribute__( (aligned) )pause;
   int     count = 0;

	do
	{
		s.pid		= pid;
		s.nowait	= 0;
		s.parts		= parts;
		s.msgmx		= msgmx;

		result = DoIt(SRR_IOCRECVMX, &s, 0);
      if (( result == -1) &&  (errno == EAGAIN ) && ( count++> 5 ))
      {                                                                                           
         pause.tv_sec = 0L;
         pause.tv_nsec = (10*1000000L);
         nanosleep( &pause, NULL );
      }
 
   } while ((result == -1) && (errno == EAGAIN) && (count < 100));

	return (result);
}

/* John Green 7/10/2000 */
pid_t SrrCreceivemx(pid_t pid, size_t parts, struct _mxfer_entry *msgmx)
{
	struct SrrRecvmxArgs s;
	int			result;
   struct timespec __attribute__( (aligned) )pause;
   int     count = 0;

	do
	{
		s.pid		= pid;
		s.nowait	= 1;
		s.parts		= parts;
		s.msgmx		= msgmx;

		result = DoIt(SRR_IOCRECVMX, &s, 0);
      if (( result == -1) &&  (errno == EAGAIN ) && ( count++> 5 ))
      {                                                                                           
         pause.tv_sec = 0L;
         pause.tv_nsec = (10*1000000L);
         nanosleep( &pause, NULL );
      }

   } while ((result == -1) && (errno == EAGAIN) && (count < 100));

	return (result);
}

int SrrReply(pid_t pid, void* msg, size_t size)
{
#ifdef SEND_AS_MX
	struct _mxfer_entry msgmx;

    msgmx.addr = msg;
    msgmx.mxfer_len = size;
    return Replymx(pid, 1, &msgmx);
#else
	struct SrrReplyArgs s;

	s.pid	= pid;
	s.msg	= msg;
	s.size	= size;

	return DoIt(SRR_IOCREPLY, &s, 1);
#endif
}

/* John Green 7/10/2000 */
int SrrReplymx(pid_t pid, size_t parts, struct _mxfer_entry *msgmx)
{
	struct SrrReplymxArgs s;

	s.pid		= pid;
	s.parts		= parts;
	s.msgmx		= msgmx;

	return DoIt(SRR_IOCREPLYMX, &s, 1);
}

/* 2000/10/15 John Green based on 2000/02/17 - Alessandro Sala */
int SrrReadmsgmx(pid_t pid, size_t offset, size_t parts,
				 struct _mxfer_entry *msgmx)
{
	struct SrrReadmsgmxArgs s;

	s.pid = pid;
	s.offset = offset;
	s.parts = parts;
	s.msgmx = msgmx;

	return (DoIt(SRR_IOCREADMSGMX, &s, 1));
}

/* 2000/10/15 John Green based on 2000/03/26 - Alessandro Sala */
int SrrWritemsgmx(pid_t pid, size_t offset, size_t parts,
				  struct _mxfer_entry *msgmx)
{
	struct SrrWritemsgmxArgs s;

	s.pid = pid;
	s.offset = offset;
	s.parts = parts;
	s.msgmx = msgmx;

	return (DoIt (SRR_IOCWRITEMSGMX, &s, 1));
}

/* 2000/02/17 - Alessandro Sala */
int SrrReadmsg(pid_t pid, size_t offset, void* msg, size_t size)
{
	struct SrrReadmsgArgs s;

	s.pid = pid;
	s.offset = offset;
	s.msg = msg;
	s.size   = size;

	return DoIt(SRR_IOCREADMSG, &s, 1);
}

/* 2000/03/26 - Alessandro Sala */
int SrrWritemsg(pid_t pid, size_t offset, void* msg, size_t size)
{
	struct SrrWritemsgArgs s;

	s.pid = pid;
	s.offset = offset;
	s.msg = msg;
	s.size   = size;

	return DoIt (SRR_IOCWRITEMSG, &s, 1);
}

/* 2005/09/05 - Len Meakin */
int SrrRelay(pid_t src, pid_t tgt)
{
	struct SrrRelayArgs s;

	s.srcpid	= src;
	s.tgtpid	= tgt;
	return DoIt(SRR_IOCRELAY, &s, 1);
}

int SrrNameAttach(nid_t n, const char* name)
{
	struct SrrNameAttachArgs s;
	int err;

	n = n;

	SRR_REG

	if(!name || strlen(name) == 0)
	{
		errno = EINVAL;
		return -1;
	}

	s.name = (char*) name;
	s.size = strlen(name);

	err = DoIt(SRR_IOCNAME_A, &s, 1);
	if ((err == -1) && (errno == EEXIST))
		errno = EBUSY;
	return err;
}

pid_t SrrNameLocate(nid_t n, const char* name, unsigned sz, unsigned* copies)
{
	struct SrrNameLocateArgs s;
	int err=ENOSYS;

	n = n; sz = sz;

	SRR_REG

	if(!name || strlen(name) == 0)
	{
		errno = EINVAL;
		return -1;
	}
	
	s.name = (char*) name;
	s.size = strlen(name);

	while(((SrrLocalFd() != -1) || (SrrReg() != -1)) &&
		  ((err = ioctl(SrrLocalFd(), SRR_IOCNAME_L, &s)) == -1)
		  && (errno == EBUSY))
	{
		SrrUnreg();
	}

	if(copies)
	{
		if(err > 0)
			*copies = 1;
		else
			*copies = 0;
	}

	return (pid_t) err;
}

int SrrNameDetach(nid_t n, int name_id)
{
	struct SrrNameDetachArgs s;

	n = n;

	s.name_id = name_id;

	return DoIt(SRR_IOCNAME_D, &s, 1);
}

pid_t SrrProxyAttach(pid_t pid, char* data, int nbytes, int priority)
{
	struct SrrProxyAttachArgs s;
  struct timespec errorTimer;
  errorTimer.tv_sec = 0;
  errorTimer.tv_nsec = 10000;
  int ret = 0, tries = 0;
	priority = priority;

	s.pid = pid;
	s.data = data;
	s.nbytes = nbytes;
  do{
    ret = ((pid_t) DoIt(SRR_IOCPROXY_A, &s, 1));
    tries++;
    if(ret == -1 && errno == EAGAIN)
      nanosleep(&errorTimer, NULL);
  } while (ret == -1 && errno == EAGAIN && tries < 10);
  return ret;
}

int SrrProxyDetach(pid_t pid)
{
	SrrProxyDetachArgs s;

	s.pid = pid;

	return DoIt(SRR_IOCPROXY_D, &s, 1);
}

pid_t SrrProxyTrigger(pid_t pid)
{
	SrrTriggerArgs s;

	s.pid = pid;

	return (pid_t) DoIt(SRR_IOCTRIGGER, &s, 1);
}


int SrrPflags (unsigned short flags, unsigned short mask, void* dum1,
			   void* dum2)
{
	SrrPFlagsArgs s;

	dum1 = dum2;

	s.pid	= getpid();
	s.flags = flags;
	s.mask	= mask;
	
	return DoIt(SRR_IOCPFLAGS, &s, 1);
}


pid_t SrrProxyRemAttach (nid_t nid, pid_t pid)
{
	return pid;
}

int SrrProxyRemDetach (nid_t nid, pid_t pid)
{
	return SrrProxyDetach(pid);
}

int SrrVcDetach (pid_t pid)
{
	return 0;
}

pid_t SrrVcAttach (nid_t nid, pid_t pid, int size, int flags)
{
	return pid;
}

timer_t SrrTimerCreate (int clock, struct sigevent *event)
{
	SrrTimerCreateArgs s;
	struct sigevent e = { sigev_signo: SIGALRM, };

	s.clock = clock;
	s.event = (event && (event->sigev_signo != 0)) ? *event : e;
	
	return DoIt(SRR_IOCTCREATE, &s, 1);
}

int SrrTimerDelete (timer_t tid)
{
	SrrTimerDeleteArgs s;

	s.tid = tid;
	
	return DoIt(SRR_IOCTDELETE, &s, 1);
}

int SrrTimerSettime (timer_t tid, int flags,
					 const struct itimerspec *value, struct itimerspec *ovalue)
{
	SrrTimerSetArgs s;

	s.tid	= tid;
	s.flags	= flags;
	s.value = value;
	s.ovalue = ovalue;

	return DoIt(SRR_IOCTSET, &s, 1);
}

int SrrTimerGettime (timer_t tid, struct itimerspec *value)
{
	SrrTimerGetArgs s;

	s.tid	= tid;
	s.value = value;

	return DoIt(SRR_IOCTGET, &s, 1);
}

int SrrClockGetres (timer_t tid, struct timespec *tp)
{
	SrrClockGetresArgs s;

	s.tid	= tid;
	s.value = tp;

	return DoIt(SRR_IOCCGETRES, &s, 1);
}

int SrrClockGettime (clockid_t clock_id, struct timespec *tp)
{
	struct timeval	tv;
	int				result;
	
	if ((result = gettimeofday (&tv, NULL)) == 0)
	{
		tp->tv_sec = tv.tv_sec;
		tp->tv_nsec = tv.tv_usec * 1000;
	}
	return (result);
}

int SrrClockSettime (clockid_t clock_id, struct timespec *tp)
{
	struct timeval	tv;
	tv.tv_sec = tp->tv_sec;
	tv.tv_usec = tp->tv_nsec / 1000;
	return (settimeofday (&tv, NULL));
}

int SrrGetpid (void)
{
	SrrGetpidArgs	args;
	return DoIt(SRR_IOCGETPID, &args, 1);
}

/* ---------- QNX4 compatibility. ----------- */

/* The QNX4 calls for Receive, Send, Receivemx and Sendmx do not alter
   the message sizes in the arguments. */
   
pid_t Receive(pid_t pid, void* msg, size_t size)
{
  size_t		n = size;
  return (SrrReceive (pid, msg, &n));
}

receiveTimeoutStruct ReceiveTimeout(pid_t pid, void* msg, size_t size, unsigned long time)
{
  size_t    n = size;
  return (SrrReceiveTimeout (pid, msg, &n, time));
}

pid_t Creceive(pid_t pid, void* msg, size_t size)
{
  size_t		n = size;
  return (SrrCreceive (pid, msg, &n));
}

receiveTimeoutStruct ReceivemxTimeout (pid_t pid, size_t parts, 
    struct _mxfer_entry *msgmx, unsigned long time)
{
	struct _mxfer_entry	*tmpmx;
	int					len = parts * sizeof(struct _mxfer_entry);
	receiveTimeoutStruct				result;
	
	tmpmx = (struct _mxfer_entry*)alloca(len);
	memcpy (tmpmx, msgmx, len);
	result = SrrReceivemxTimeout (pid, parts, tmpmx, time);
	return (result);
}


pid_t Receivemx (pid_t pid, size_t parts, struct _mxfer_entry *msgmx)
{
	struct _mxfer_entry	*tmpmx;
	int					len = parts * sizeof(struct _mxfer_entry);
	pid_t				result;
	
	tmpmx = (struct _mxfer_entry*)alloca(len);
	memcpy (tmpmx, msgmx, len);
	result = SrrReceivemx (pid, parts, tmpmx);
	return (result);
}

pid_t Creceivemx (pid_t pid, size_t parts, struct _mxfer_entry *msgmx)
{
	struct _mxfer_entry	*tmpmx;
	int					len = parts * sizeof(struct _mxfer_entry);
	pid_t				result;
	
	tmpmx = (struct _mxfer_entry*)alloca(len);
	memcpy (tmpmx, msgmx, len);
	result = SrrCreceivemx (pid, parts, tmpmx);
	return (result);
}

int Sendmx (pid_t pid, size_t sparts, size_t rparts,
			struct _mxfer_entry *smsg, struct _mxfer_entry *rmsg)
{
	struct _mxfer_entry	*rtmpmx, *stmpmx;
	int					rlen = rparts * sizeof(struct _mxfer_entry);
	int					slen = sparts * sizeof(struct _mxfer_entry);
	int					result;
	
	rtmpmx = (struct _mxfer_entry*)alloca(rlen);
	stmpmx = (struct _mxfer_entry*)alloca(slen);
	memcpy (rtmpmx, rmsg, rlen);
	memcpy (stmpmx, smsg, slen);
	result = SrrSendmx (pid, sparts, rparts, stmpmx, rtmpmx);
	return (result == -1 ? -1 : 0);
}

sendTimeoutStruct SendmxTimeout (pid_t pid, size_t sparts, size_t rparts,
			struct _mxfer_entry *smsg, struct _mxfer_entry *rmsg,
      unsigned long time)
{
	struct _mxfer_entry	*rtmpmx, *stmpmx;
	int					rlen = rparts * sizeof(struct _mxfer_entry);
	int					slen = sparts * sizeof(struct _mxfer_entry);
	sendTimeoutStruct					result;
	
	rtmpmx = (struct _mxfer_entry*)alloca(rlen);
	stmpmx = (struct _mxfer_entry*)alloca(slen);
	memcpy (rtmpmx, rmsg, rlen);
	memcpy (stmpmx, smsg, slen);
	result = SrrSendmxTimeout (pid, sparts, rparts, stmpmx, rtmpmx, time);
	return (result);
}

