/*
 * srr_mod.c: the bare srr char module
 *
 *  Linux Loadable Module Implementation of QNX Send/Receive/Reply
 *	message passing, and related function.
 *
 *	QNX is a trademark of QSSL, Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software in a file called 'COPYING'; if not, write to:
 * 	Free Software Foundation, Inc.,
 * 	59 Temple Place, Suite 330,
 * 	Boston, MA 
 * 	02111-1307  USA
 *
 * Tested with 2.0.x (x86 and alpha) and 1.2.13 (x86)
 *
 * $Id: srr_mod.c 226 2011-08-22 03:38:58Z meaton $
 * $Id: srr_mod.c 226 2011-08-22 03:38:58Z meaton $
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif

#ifndef MODULE
#  define MODULE
#endif

#define DECLARE_MODULE_VERSION /* define kernel version only once, here. */
#include "module_versions.h"

#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>
#include <linux/miscdevice.h>
#include <linux/spinlock.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <asm/system.h>
#include <asm/segment.h>

#include "srr_mod.h"
#include "srr_sys.h"
#include "srr_time.h"
#include "srr_util.h"

#if	LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#	include <linux/poll.h>
#endif

#if	LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/compiler.h>
#include <linux/device.h>
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
  #include <linux/devfs_fs_kernel.h>
  #endif
#endif
 
/*
#if LINUX_VERSION_CODE > KERNEL_VERSION(1,99,3)
*/
	/* 1.99.4 exported the needed symbols */
#	define SRR_USE_PROC
/*
#endif
*/

/*
 * I don't use static symbols here, because register_symtab is called
 */

#define SRR_MINOR	171

int srr_major	= MISC_MAJOR;
int srr_minor	= SRR_MINOR;
int srr_dbglvl	= 1;

struct SrrTm* srr_taskmap;

#ifdef MODULE_AUTHOR
    MODULE_AUTHOR("Andrew Thomas <andrew@cogent.ca> [Cogent Real-Time Systems] et al.");
#endif
#ifdef MODULE_DESCRIPTION
#ifdef TASK_SPECIFIC_LOCKS
    MODULE_DESCRIPTION("Send/Receive/Reply IPC module.  Task-specific version.");
#else
    MODULE_DESCRIPTION("Send/Receive/Reply InterProcess Communication module");
#endif
#endif
#ifdef MODULE_LICENSE
    MODULE_LICENSE("GPL");
#endif
#ifdef MODULE_VERSION
    MODULE_VERSION( VERSION );
#endif
#ifdef MODULE_ALIAS_MISCDEV
	MODULE_ALIAS_MISCDEV(SRR_MINOR);
#endif

#ifdef NOTDEF
#ifdef MODULE_DEVICE_TABLE
#include <linux/pci.h>
static struct pci_device_id srr_pci_tbl[] =
{
 	{ 0x0001, 0x0001, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0 },
 	{ 0, },	/* terminate list */
};
MODULE_DEVICE_TABLE (pci, srr_pci_tbl);
#endif
#endif

#ifdef MODULE_ALIAS_MISCDEV
MODULE_ALIAS_MISCDEV(SRR_MINOR);
#endif
	
/*
 * The proc filesystem: function to read and entry
 */
#ifdef SRR_USE_PROC /* don't waste space if unused */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
#define READ_U	,int unused
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
#define READ_U	,int *eof, void* data
#else
#define READ_U
#endif

#ifdef  SRR_TRACEPOINT
#include "srr_trace.h"

EXPORT_TRACEPOINT_SYMBOL(srripc_start_event);

static void probe_subsys_event( pid_t pid)
{
	if ( pid == -1 )
	printk( KERN_INFO "event_start  pid %d \n", pid);
}
EXPORT_TRACEPOINT_SYMBOL(srripc_stop_event);

static void probe_subsys_eventb( pid_t pid)
{
	if (pid == -1 )
	printk( KERN_INFO "event_stop  pid %d \n", pid);
}

#endif
#ifdef PROC_SRR_TARGET
int srr_read_procmem(char* buf, char** start, off_t offset, int len READ_U)
{
	int		totlen=0;
#if 0
	len = 0; /* why...? */
	len += sprintf(buf, "My name is Inigo Montoya, ...");
	return len;
#endif

	return SrrDumpState(srr_taskmap, buf, PAGE_SIZE, &totlen);
}
#endif

int srr_read_sin (char* buf, char** start, off_t offset, int len READ_U)
{
	int		totlen=0;
	return SrrDumpSin(srr_taskmap, buf, PAGE_SIZE, &totlen);
}

int srr_read_proxy (char* buf, char** start, off_t offset, int len READ_U)
{
	int		totlen=0;
	return SrrDumpProxy(srr_taskmap, buf, PAGE_SIZE, &totlen);
}

int srr_read_version (char* buf, char** start, off_t offset, int len READ_U)
{
	int		totlen=0;
	return SrrDumpVersion(srr_taskmap, buf, PAGE_SIZE, &totlen);
}

int srr_read_blocks (char* buf, char** start, off_t offset, int len READ_U)
{
	int		totlen=0;
	return SrrDumpTaskBlocks(srr_taskmap, buf, PAGE_SIZE, &totlen);
}

int srr_read_timers (char* buf, char** start, off_t offset, int len READ_U)
{
	int		totlen=0;
	return SrrDumpTimers(srr_taskmap, buf, PAGE_SIZE, &totlen);
}

#ifndef NO_IRQ	
int srr_read_irq (char* buf, char** start, off_t offset, int len READ_U)
{
	int		totlen=0;
	return SrrDumpIrq(srr_taskmap, buf, PAGE_SIZE, &totlen);
}
#endif

int srr_read_names (char* buf, char** start, off_t offset, int len READ_U)
{
	int		totlen=0;
	return SrrDumpNames(srr_taskmap, buf, PAGE_SIZE, &totlen);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
struct proc_dir_entry srr_proc_entry = {
		0,					/* low_ino: the inode -- dynamic */
		6, "srripc",		/* len of name; name */
		S_IFDIR | S_IRUGO | S_IXUGO,	/* mode */
		2, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		NULL,				/* function used to read data */
		/* nothing more */
	};

#ifdef PROC_SRR_TARGET
struct proc_dir_entry srr_info_entry = {
		0,					/* low_ino: the inode -- dynamic */
		3, "srr",			/* len of name; name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		&srr_read_procmem,	/* function used to read data */
		/* nothing more */
	};
#endif

struct proc_dir_entry srr_sin_entry = {
		0,					/* low_ino: the inode -- dynamic */
		3, "sin",			/* len of name; name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		&srr_read_sin,		/* function used to read sin output */
		/* nothing more */
	};

struct proc_dir_entry srr_proxy_entry = {
		0,					/* low_ino: the inode -- dynamic */
		5, "proxy",			/* len of name; name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		&srr_read_proxy,	/* function used to read sin output */
		/* nothing more */
	};

#ifndef NO_IRQ	
struct proc_dir_entry srr_irq_entry = {
		0,					/* low_ino: the inode -- dynamic */
		3, "irq",			/* len of name; name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		&srr_read_irq,	/* function used to read sin output */
		/* nothing more */
	};
#endif

struct proc_dir_entry srr_version_entry = {
		0,					/* low_ino: the inode -- dynamic */
		7, "version",		/* len of name; name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		&srr_read_version,	/* function used to read sin output */
		/* nothing more */
	};

struct proc_dir_entry srr_blocks_entry = {
		0,					/* low_ino: the inode -- dynamic */
		6, "blocks",		/* len of name; name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		&srr_read_blocks,	/* function used to read sin output */
		/* nothing more */
	};

struct proc_dir_entry srr_timers_entry = {
		0,					/* low_ino: the inode -- dynamic */
		7, "timers",		/* len of name; name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		&srr_read_timers,	/* function used to read sin output */
		/* nothing more */
	};

struct proc_dir_entry srr_names_entry = {
		0,					/* low_ino: the inode -- dynamic */
		5, "names",			/* len of name; name */
		S_IFREG | S_IRUGO,	/* mode */
		1, 0, 0,			/* nlinks; owner; group */
		0, NULL,			/* size - unused; operations -- use default */
		&srr_read_names,	/* function used to read sin output */
		/* nothing more */
	};
#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0) */
static struct proc_dir_entry	*srr_proc_root = NULL;
#ifdef PROC_SRR_TARGET
static struct proc_dir_entry	*srr_proc_info = NULL;
#endif
static struct proc_dir_entry	*srr_proc_names = NULL;
static struct proc_dir_entry	*srr_proc_proxy = NULL;
static struct proc_dir_entry	*srr_proc_sin = NULL;
#ifndef NO_IRQ	
static struct proc_dir_entry	*srr_proc_irq = NULL;
#endif
static struct proc_dir_entry	*srr_proc_version = NULL;
static struct proc_dir_entry	*srr_proc_blocks = NULL;
static struct proc_dir_entry	*srr_proc_timers = NULL;
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0) */
#endif /* SRR_USE_PROC */

/*
 * Open and close
 */

int srr_open (struct inode *inode, struct file *filp)
{
	PDEBUG("srr_open() pid %d\n", current->pid);

	/* currently don't use minors, so verify it's 0 */

	if(MINOR(inode->i_rdev) != srr_minor)
		return -ENODEV;

	/* This file is already open */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,35)
	if (filp->private_data)
	{
		Oops("srr_open() pid %d attempted to open already opened file\n",
			 current->pid);
		return -EACCES;
	}
#endif	
	/* use filp->private_data to point to the task struct */
	filp->private_data = (void*) SrrTm_Alloc(srr_taskmap, current->pid, 0);

	if(!filp->private_data)
		return -ENOMEM;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	try_module_get (THIS_MODULE);
#else
	MOD_INC_USE_COUNT;
#endif

	return 0;		  /* success */
}

/* In version 2.2 kernels, we had to use the flush as evidence that
   the file descriptor was closing.  This is not the case with 2.4,
   where we can use the release instead. */

/* Flush is called once by each task that opens a file descriptor.
   After the last close (flush) is called, then release() is called
   once on the file descriptor and that's the end of it. */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
int srr_flush(struct file *filp
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
			  , fl_owner_t id
#endif
	)
{
	SrrTask* task;

	task = (SrrTask*) filp->private_data;

	PDEBUG("srr_flush(%p) current %d task %d\n",
		   filp, current->pid, task ? task->pid : 0);

	/* If task is us, we did the open and we've now called close.  We
	   would like to do this in the release(), but that might not be
	   called by the correct process, since fds get copied to new
	   child processes and threads and they might be last to close
	   the fd.  */
	if(task && task->pid == current->pid)
	{
		SrrTm_Free(srr_taskmap, current->pid);
		filp->private_data = 0;
	}
	else if (!task)
	{
		/* This is not an oops.  It means that the owner of the fd has
		   died and exercised the above clause, and the child of the
		   owner has subsequently died.  The order should not matter
		   because the child is never supposed to free the SRR info
		   anyway. */
		PDEBUG ("Task %d called flush on deleted connection.\n", current->pid);
	}
	else
	{
		/* This is not an oops.  This is a child process closing the
		   connection. */
		PDEBUG ("Task %d is closing connection, but %d opened it.\n",
				current->pid, task->pid);
	}
	return 0;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
int
#else
void
#endif
srr_release (struct inode *inode, struct file *filp)
{
	SrrTask* task = (SrrTask*) filp->private_data;

	PDEBUG("srr_release(%p) current %d task %d\n",
		   filp, current->pid, task ? task->pid : 0);

	if (task)
	{
		if (current->pid != task->pid)
		{
			/* This is an oops.  This is a child process being the
			   last to close the file descriptor, but the original
			   owner has not freed the task structure.  The original
			   owner of the FD is the one that should do the
			   SrrTm_Free in the srr_flush() call.  The library calls
			   close() on the fd when a thread exits to ensure that
			   the flush/release pair are called by the thread, not by
			   the parent. */
			/* In this case the death of task->pid will not correctly
			   get announced to PPF_INFORMed processes.  */
			/* OK, this is *not* an oops.  The linux kernel allows
			   this case to happen - one task opens a FD, and another
			   closes the FD on behalf of the opener.  This means that
			   in order to send task death notifications we need to
			   masquerade as the original opener, so this is just a
			   normal occurrence when using pthreads. */
			PDEBUG("srr_release() task %d being released by process %d\n",
				   task->pid, current->pid);
		}
		else
		{
			Oops("srr_release() but %d failed to flush!\n", task->pid);
		}
		/* free the task in the filp - it's us or an exited process */
		SrrTm_Free(srr_taskmap, task->pid);
	}

	filp->private_data = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
	module_put (THIS_MODULE);
#else	
	MOD_DEC_USE_COUNT;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	return 0;
#endif
}

/*
 * The ioctl() implementation
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
long srr_ioctl( struct file *flip,
      unsigned int cmd, unsigned long arg )

#else
int srr_ioctl (struct inode *inode, struct file *filp,
		 unsigned int cmd, unsigned long arg)

#endif
{
	int		err		= 0;
	int		size	= _IOC_SIZE(cmd); /* the size bitfield in cmd */
	SrrTask* task	= 0;

	/* force the usage count to 1 (for the fd doing the ioctl()), so that
	 * the module can be unloaded. */
	if(cmd == SRR_IOCUNLOAD) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
		while(MOD_IN_USE)
			MOD_DEC_USE_COUNT;
		MOD_INC_USE_COUNT;
#endif
		return 0;
	}

	PDEBUG("srr_ioctl() size %d, dir %x, type %c, number %d\n",
		size, _IOC_DIR(cmd), _IOC_TYPE(cmd), _IOC_NR(cmd));
	
	task	= (struct SrrTask*)filp->private_data;

	/*
	 * Processes using srr cannot share connections, so if the task pid is
	 * wrong, or if it has been zero'ed (which can happen after a flush),
	 * return an error to force the process to open its own connection.
	 */
	if(!task || current->pid != task->pid) {
		PDEBUG("ioctl() from %d on fd belonging to %d!\n",
			current->pid, task ? task->pid : 0);

		return -EBUSY;
	}

	/*
	 * extract the type and number bitfields, and don't decode
	 * wrong cmds: return EINVAL before verify_area()
	 */
	if (_IOC_TYPE(cmd) != SRR_IOC_TYPE) {
		Oops("ioctl() cmd %#x has type %#x, not %#x!\n",
			cmd, _IOC_TYPE(cmd), SRR_IOC_TYPE);
		return -EINVAL;
	}
	if (_IOC_NR(cmd) > SRR_IOC_MAXNR) {
		Oops("ioctl() cmd %#x has number %d, greater than %d!\n",
			cmd, _IOC_NR(cmd), SRR_IOC_MAXNR);
		return -EINVAL;
	}

	/*
	 * the direction is a bitmask, and VERIFY_WRITE catches R/W
	 * transfers. `Type' is user-oriented, while
	 * verify_area is kernel-oriented, so the concept of "read" and
	 * "write" is reversed
	 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
		/* under 2.2 kernels this functionality is wrapped in the copyies */
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = verify_area(VERIFY_WRITE, (void *)arg, size);
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err =  verify_area(VERIFY_READ, (void *)arg, size);
	
	if (err) {
		Oops("ioctl() cmd %#x failed verify_area!\n", cmd);
		return err;
	}
#endif
	
	if(cpy_u2k(&task->args, (void*)arg, size) != 0)
	{
		PDEBUG("Memory copy fault");
		return -EFAULT;
	}


	/* Poll information is inexact.  If we enter an ioctl in
	 * state POLL_BLOCKED, we must have exited a poll some time
	 * in the past. */
	if (task->flags & _PPF_POLLING)
	{
		PDEBUG("Task %d got ioctl %#x while POLL_BLOCKED\n",
			   task->pid, cmd);
		task->state = READY;
		task->flags &= ~_PPF_POLLING;
	}

	task->last_ioctl = cmd;
	
	PDEBUG("switching on %d\n", _IOC_NR(cmd));

	switch(_IOC_NR(cmd)) {
			
		case SRR_SEND: {
			
			err = SrrModSend(srr_taskmap, task);
#ifdef RETURN_SIZES
#ifdef TASK_SPECIFIC_LOCKS
#ifdef LOCK_VIA_SPINLOCKS
      spin_lock_irqsave(&task->mutex.lock, task->mutex.flags);
#else
      mutex_lock(&task->mutex.mutex);
#endif
#endif
			cpy_k2u((void*)arg, &task->args, size);
#ifdef TASK_SPECIFIC_LOCKS
#ifdef LOCK_VIA_SPINLOCKS
      spin_unlock_irqrestore(&task->mutex.lock, task->mutex.flags);
#else
      mutex_unlock(&task->mutex.mutex);
#endif
#endif
#endif	
    } break;
			
		case SRR_RECV: {
			
			err = SrrRecv(srr_taskmap, task);
			/* Return the actual number of bytes transferred */
#ifdef RETURN_SIZES
#ifdef TASK_SPECIFIC_LOCKS
#ifdef LOCK_VIA_SPINLOCKS
      spin_lock_irqsave(&task->mutex.lock, task->mutex.flags);
#else
      mutex_lock(&task->mutex.mutex);
#endif
#endif
      cpy_k2u((void*)arg, &task->args, size);
#ifdef TASK_SPECIFIC_LOCKS
#ifdef LOCK_VIA_SPINLOCKS
      spin_unlock_irqrestore(&task->mutex.lock, task->mutex.flags);
#else
      mutex_unlock(&task->mutex.mutex);
#endif
#endif
#endif /* RETURN_SIZES */
			} break;
			
		case SRR_REPLY: {
			
			err = SrrModReply(srr_taskmap, task);
			
			} break;
			
		case SRR_NAME_A: {
			
			err = SrrModNameAttach(srr_taskmap, task);
			
			} break;
			
		case SRR_NAME_L: {

			err = SrrModNameLocate(srr_taskmap, task);

			} break;

		case SRR_NAME_D: {

			err = SrrModNameDetach(srr_taskmap, task);

			} break;

		case SRR_PROXY_A: {

			err = SrrModProxyAttach(srr_taskmap, task);
			
      } break;

		case SRR_PROXY_D: {

			err = SrrModProxyDetach(srr_taskmap, task);

			} break;

		case SRR_TRIGGER: {

			err = SrrModTrigger(srr_taskmap, task);

			} break;

		case SRR_TRACE: {
			
			PDEBUG("set trace level to %d from %d\n",
				   task->args.trace.level, err);
			
			err = srr_debug_trace(task->args.trace.level);
			
			PDEBUG("set trace level to %d from %d\n",
				   task->args.trace.level, err);
			
			} break;
			
		case SRR_PFLAGS: {

			err = SrrPFlags(srr_taskmap, task);

			} break;

		case SRR_TCREATE: {
			PDEBUG ("ioctl: timer create\n");
			err = SrrModTimerCreate(srr_taskmap, task);

			} break;

		case SRR_TDELETE: {

			err = SrrModTimerDelete(srr_taskmap, task);

			} break;

		case SRR_TSET: {

			err = SrrModTimerSet(srr_taskmap, task);

			} break;

		case SRR_TGET: {

			err = SrrModTimerGet(srr_taskmap, task);

			} break;

		case SRR_CGETRES: {

			err = SrrModClockGetres(srr_taskmap, task);

			} break;

		case SRR_KILL: {
			
			err = SrrTm_Kill(srr_taskmap, task->args.kill.pid, task->args.kill.signo);
			
			} break;

		/* 2000/02/17 - Alessandro Sala */
		case SRR_READMSG: {
	
			err = SrrModReadmsg(srr_taskmap, task);

			} break;
			
		/* 2000/03/26 - Alessandro Sala */
		case SRR_WRITEMSG: {
	
			err = SrrModWritemsg(srr_taskmap, task);

			} break;

		case SRR_SENDMX: {
			
			err = SrrModSendmx(srr_taskmap, task);
#ifdef RETURN_SIZES
			cpy_k2u((void*)arg, &task->args, size);
#endif /* RETURN_SIZES */
			} break;
			
		case SRR_RECVMX: {
			
			err = SrrRecvmx(srr_taskmap, task);
#ifdef RETURN_SIZES
			cpy_k2u((void*)arg, &task->args, size);
#endif /* RETURN_SIZES */
			} break;
			
		case SRR_REPLYMX: {
			
			err = SrrModReplymx(srr_taskmap, task);
			
			} break;
			
		/* 2000/10/15 - John Green thanks Alessandro Sala */
		case SRR_READMSGMX: {
	
			err = SrrModReadmsgmx(srr_taskmap, task);

			} break;
			
		/* 2000/10/15 - John Green thanks Alessandro Sala */
		case SRR_WRITEMSGMX: {
	
			err = SrrModWritemsgmx(srr_taskmap, task);

			} break;

#ifndef NO_IRQ
		/* 2002/04/18 - Andrew Thomas */
		case SRR_IRQ: {
	
			err = SrrIrq_DoIRQCommand(srr_taskmap, task);
			if (err == 0)
				cpy_k2u((void*)arg, &task->args, size);
				
			} break;
#endif

	    case SRR_GETPID: {
			err = current->pid;
		} break;
		
		/* 2005/09/05 - Len Meakin */
		case SRR_RELAY: {
			
			err = SrrModRelay(srr_taskmap, task);
			
			} break;
		case SRR_RECVTIME: {
			err = SrrRecvTimeout(srr_taskmap, task);
			/* Return the actual number of bytes transferred */
#ifdef RETURN_SIZES
			cpy_k2u((void*)arg, &task->args, size);
#endif /* RETURN_SIZES */
         }
          break;
			
    case SRR_SENDTIME: {
       err = SrrModSendTimeout(srr_taskmap, task);
#ifdef RETURN_SIZES
		 	 cpy_k2u((void*)arg, &task->args, size);
#endif /* RETURN_SIZES */
      }
      break;
     
    case SRR_RECVMXTIME: {
        MXDEBUG("Calling SRR_RECVMXTIME\n");    
			err = SrrRecvmxTimeout(srr_taskmap, task);
#ifdef RETURN_SIZES
			cpy_k2u((void*)arg, &task->args, size);
#endif /* RETURN_SIZES */
      }
      break;
    
    case SRR_SENDMXTIME: {
        MXDEBUG("Calling SRR_SENDMXTIME\n");    
			err = SrrModSendmxTimeout(srr_taskmap, task);
#ifdef RETURN_SIZES
			cpy_k2u((void*)arg, &task->args, size);
#endif /* RETURN_SIZES */
      }
      break;
    
    default:  /* paranoid, as cmd was checked against MAXNR */
			err = -ENOSYS;
	}

	PDEBUG("srr_ioctl(%d) return %d\n", _IOC_NR(cmd), err);

	if(task->state != READY) {
		Oops("ioctl() pid %d cmd %d return %d, but state %s!\n",
			current->pid, cmd, err, SrrTaskState2Str(task, task->state));
		task->state = READY;
	}

	return err;
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37)

long srr_unlocked_ioctl( struct file *filp, unsigned int cmd, unsigned long arg )
{
  //  static int cnt;
    long status;
    struct inode *inode = filp->f_dentry->d_inode;

 //   if( cnt == 0 ) Oops("ioctl() unlocked ioctl called\n" );
//    cnt++;

    status = (long)srr_ioctl( inode, filp, cmd, arg );

    return( status );
}

#endif
/*
 * select()/poll() is used by clients to wait for a condition on a "normal" file
 * descriptor, or for a QNX message or proxy to become available. The "read"
 * condition is interpreted as "message available", and the write condition is
 * always considered true.
 *
 * Note that processes using srr cannot (currently) share connections, so
 * if the task pid is wrong, we do not own the fd. All other calls are made from
 * one of the library wrappers, so returning an error code forces a reopen.
 * select() is different as it is called directly by clients. Regardless, we
 * can't indicate that Receive() won't block, and we can't block the client till a
 * message is available (because one never will be) so we have to return an error.
 *
 * The select/poll mechanism has changed significantly from kernel
 * version 2.0 -> 2.2, so there are two versions of thes system call.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)

unsigned int srr_poll (struct file* filp, poll_table* table)
{
	SrrTask* task	= 0;
	int		 result = 0;

	PDEBUG("srr_poll()\n");

	task = (struct SrrTask*)filp->private_data;

	if(current->pid != task->pid)
	{
		Oops("srr_poll() from %d on fd belonging to %d!\n",
			current->pid, task->pid);
		result = -EAGAIN;
	}
	else
	{
		poll_wait(filp, &task->poll_q, table);
		
		result = POLLOUT | POLLWRNORM;
	
		if (SrrRq_NonEmpty(task))
			result |= POLLIN | POLLRDNORM;

		PDEBUG ("  Poll result is 0x%x\n", result);

		task->flags |= _PPF_POLLING;
	}
	
	return (result);
}

#else

int srr_poll (struct inode* inode, struct file* filp, int mode,
			  select_table* table)
{
	SrrTask* task	= 0;
	int		 result = 0;

	PDEBUG("srr_select() mode %#x\n", mode);

	task = (struct SrrTask*)filp->private_data;

	if(current->pid != task->pid) {
		Oops("srr_poll() from %d on fd belonging to %d!\n",
			current->pid, task->pid);
		result = -EAGAIN;

	} else if(mode == SEL_OUT) {
		result = 1;

	} else if(mode == SEL_IN) {
		if (SrrRq_NonEmpty(task))
			result = 1;
		else
		{
			select_wait(&task->poll_q, table);
			task->flags |= _PPF_POLLING;
		}
	}

	return (result);
}

#endif

/*
 * The file operations structure
 */

static struct file_operations srr_fops = {
	.owner = 	THIS_MODULE,
	.poll =		srr_poll,		/* poll/select */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
   .unlocked_ioctl =srr_ioctl,
#else  
	.ioctl =	srr_ioctl,
#endif  
	.open =		srr_open,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	.flush =	srr_flush,		/* flush */
#endif
	.release =	srr_release,	/* release/close */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37)
        .unlocked_ioctl = srr_unlocked_ioctl
#endif		
};

/*
 * Finally, the module stuff
 */

/*
 * The minor device expects us to create and hang onto the struct
 * miscdevice, so it must be either global or allocated on the heap.
 */
static struct miscdevice SrrMiscDevice =
{
	.minor = 			SRR_MINOR,	// int minor number
	.name =				"srripc",	// const char* name
	.fops =				&srr_fops,	// struct file_operations *fops;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
	.devfs_name =		"srripc",	// char devfs_name[64];
#endif
};


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/*
 * Conform to 2.6.0 module definitions
 */
static int __init srr_module_init(void)
#else
 int init_module(void)
#endif
{
	int						result;

#ifdef SRR_TRACEPOINT
        result = register_trace_srripc_start_event( probe_subsys_eventb);
  	WARN_ON(result);
	result = register_trace_srripc_stop_event( probe_subsys_event);
  	WARN_ON(result);
#endif
	printk(KERN_DEBUG
		   "SRRIPC (C) 2001-2006 Cogent Real-Time Systems Inc.  All Rights Reserved.\n");
	printk(KERN_DEBUG
		   "    SRRIPC version %s initializing as misc major %d, minor %d\n",
		   VERSION, srr_major, srr_minor);
    // add notice on patches
	//	printk(KERN_DEBUG
	//		   "SRRIPC - this is a patched version supporting big messages (>64K) [compiled on "__DATE__" "__TIME__"].\n");
	/*
	 * Allocate the taskmap, our root data structure
	 */
	
	srr_taskmap = SrrTm_new();
	if(!srr_taskmap) {
		printk(KERN_DEBUG
		   "SRRIPC task map allocation failed.  Module not installed.\n");
		return -ENOMEM;
	}

	/*
	 * By assigning the minor number with srr_minor, we can make the
	 * minor number a module parameter.  This is useful if testing more
	 * than one copy of the module on a machine.
	 */

	SrrMiscDevice.minor = srr_minor;
	result = misc_register (&SrrMiscDevice);
	
	if (result < 0) {
		printk(KERN_WARNING "srr: can't get major %d, minor %d\n",
			   srr_major, srr_minor);
		return result;
	}

	/*
	 * Do more module initialization stuff...
	 */
#ifndef SRR_DEBUG
#  if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,0)
	register_symtab(NULL); /* otherwise, leave global symbols visible */
#  endif
#endif

#ifdef SRR_USE_PROC /* only when available */
	/* this is the last line in init_module */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,0)
	srr_proc_root = proc_mkdir ("srripc", PROC_ROOT);
#ifdef PROC_SRR_TARGET
	srr_proc_info = srr_create_proc_read_entry ("srr", S_IFREG | S_IRUGO,
											srr_proc_root, srr_read_procmem,
											NULL);
#endif
	srr_proc_names = srr_create_proc_read_entry ("names", S_IFREG | S_IRUGO,
											 srr_proc_root, srr_read_names,
											 NULL);
	srr_proc_proxy = srr_create_proc_read_entry ("proxy", S_IFREG | S_IRUGO,
											 srr_proc_root, srr_read_proxy,
											 NULL);
	srr_proc_sin = srr_create_proc_read_entry ("sin", S_IFREG | S_IRUGO,
										   srr_proc_root, srr_read_sin,
										   NULL);
#ifndef NO_IRQ	
	srr_proc_irq = srr_create_proc_read_entry ("irq", S_IFREG | S_IRUGO,
										   srr_proc_root, srr_read_irq,
										   NULL);
#endif
	srr_proc_version = srr_create_proc_read_entry ("version", S_IFREG | S_IRUGO,
											   srr_proc_root, srr_read_version,
											   NULL);
	srr_proc_blocks = srr_create_proc_read_entry ("blocks", S_IFREG | S_IRUGO,
											  srr_proc_root, srr_read_blocks,
											  NULL);
	srr_proc_timers = srr_create_proc_read_entry ("timers", S_IFREG | S_IRUGO,
											  srr_proc_root, srr_read_timers,
											  NULL);
#else
#  if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	proc_register(PROC_ROOT, &srr_proc_entry);
#ifdef PROC_SRR_TARGET
	proc_register(&srr_proc_entry, &srr_info_entry);
#endif
	proc_register(&srr_proc_entry, &srr_sin_entry);
#ifndef NO_IRQ	
	proc_register(&srr_proc_entry, &srr_irq_entry);
#endif
	proc_register(&srr_proc_entry, &srr_version_entry);
	proc_register(&srr_proc_entry, &srr_blocks_entry);
	proc_register(&srr_proc_entry, &srr_timers_entry);
	proc_register(&srr_proc_entry, &srr_proxy_entry);
	proc_register(&srr_proc_entry, &srr_names_entry);
#  else
	proc_register_dynamic(PROC_ROOT, &srr_proc_entry);
#  endif
#endif
#else  // !SRR_USE_PROC
	printk (KERN_DEBUG "SRRIPC: /proc filesystem support not present.\n");
#endif // SRR_USE_PROC

	return 0; /* success! */
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/*
 * Conform to 2.6.0 module definitions
 */
  static void __exit srr_module_cleanup(void)
#else
  void cleanup_module(void)
#endif
{
#ifdef SRR_TRACEPOINT
	unregister_trace_srripc_stop_event( probe_subsys_eventb);
	unregister_trace_srripc_start_event( probe_subsys_event);
        tracepoint_synchronize_unregister();
#endif
	printk(KERN_DEBUG
		   "SRRIPC module shutting down: misc major %d, minor %d\n",
		   srr_major, srr_minor);

	misc_deregister (&SrrMiscDevice);

	SrrTm_delete(srr_taskmap);

#ifdef SRR_USE_PROC
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
	proc_unregister(&srr_proc_entry, srr_names_entry.low_ino);
	proc_unregister(&srr_proc_entry, srr_proxy_entry.low_ino);
	proc_unregister(&srr_proc_entry, srr_sin_entry.low_ino);
#ifdef PROC_SRR_TARGET
	proc_unregister(&srr_proc_entry, srr_info_entry.low_ino);
#endif
#ifndef NO_IRQ	
	proc_unregister(&srr_proc_entry, srr_irq_entry.low_ino);
#endif
	proc_unregister(&srr_proc_entry, srr_version_entry.low_ino);
	proc_unregister(&srr_proc_entry, srr_blocks_entry.low_ino);
	proc_unregister(&srr_proc_entry, srr_timers_entry.low_ino);
	proc_unregister(PROC_ROOT, srr_proc_entry.low_ino);
#else
#ifdef PROC_SRR_TARGET
	remove_proc_entry ("srr", srr_proc_root);
#endif
	remove_proc_entry ("names", srr_proc_root);
	remove_proc_entry ("proxy", srr_proc_root);
	remove_proc_entry ("sin", srr_proc_root);
#ifndef NO_IRQ	
	remove_proc_entry ("irq", srr_proc_root);
#endif
	remove_proc_entry ("version", srr_proc_root);
	remove_proc_entry ("blocks", srr_proc_root);
	remove_proc_entry ("timers", srr_proc_root);
	remove_proc_entry ("srripc", PROC_ROOT);
#endif
#endif
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/*
 * Conform to 2.6.0 module definitions
 */
 module_init( srr_module_init );
 module_exit( srr_module_cleanup );


#endif
