/*
 * task.c: implementation of the task structures
 *
 *  Linux Loadable Module Implementation of QNX Send/Receive/Reply
 *	message passing, and related function.
 *
 *	QNX is a trademark of QSSL, Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software in a file called 'COPYING'; if not, write to:
 * 	Free Software Foundation, Inc.,
 * 	59 Temple Place, Suite 330,
 * 	Boston, MA 
 * 	02111-1307  USA
 *
 * $Id: srr_sys.c 237 2011-09-29 22:02:10Z mstormer $
 */

/* There is a real mystery here in the inclusion order of headers.
   If you include linux/modversions.h in some distributions, you
   will get all kinds of duplicate (and incompatible) definitions.
   I have NO idea how to solve this.  So we put up with undefined
   symbol complaints from depmod.  If anybody has any ideas how to
   do this "right" for 2.2 and 2.4 kernels, please let me know at
   andrew@cogent.ca.
   */

#include "module_versions.h"
#include <linux/spinlock.h>
#include <linux/jiffies.h>
#include <linux/rtmutex.h>
#include "sysdep.h"

#include "srr_mod.h"
#include "srr_util.h"
#include "srr_tt.h"
#include "srr_time.h"
#include "srr_sys.h"
#ifdef  SRR_TRACEPOINT
#include "srr_trace.h"
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
/* Remove any previous definitions.  */
#undef	alloca
extern void *alloca (size_t __size);
#ifdef	__GNUC__
# define alloca(size)	__builtin_alloca (size)
#endif /* GCC.  */
#else
#include <alloca.h>
#endif /* Kernel version 2.6.0 onward. */

spinlock_t rcv_q_lock = SPIN_LOCK_UNLOCKED;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37)
DEFINE_SEMAPHORE(srr_module_mutex)
#else
#ifndef USE_RT_MUTEX
DECLARE_MUTEX(srr_module_mutex);
#else
DEFINE_RT_MUTEX(srr_module_mutex);
#endif
#endif

#ifdef SRR_TRACEPOINT
DEFINE_TRACE(srripc_start_event);
DEFINE_TRACE(srripc_stop_event);
#endif
/*
 * Task wake and sleep: wrappers useful for debugging.
 */

void SrrTaskWake(SrrTask* task, int state, const char* whence)
{
	PDEBUG("TaskWake %d from %s to %s\n",
		task->pid,
		SrrTaskState2Str(task, task->state),
		SrrTaskState2Str(NULL, state)
		);

	if (task->state == state)
	   Oops ("Task %d was already in state %s on wake\n", task->pid,
            SrrTaskState2Str(task, task->state));
	
	task->state = state;
	/* AST: current should be replaced with effective task id of
	   the waking task.  We need this for Sends during death that come
	   from the wrong task.  It happens and we need to deal with it. */
	/* It doesn't matter who wakes a dying task.  This task ID of the
	   waker is only used by Recv, which should never be called by a
	   dying task. */
	/* The Relay call can set the waker before calling SrrTaskWake
	   so we have to let it override the current task id */
	if (task->waker == -1)
		task->waker = current->pid;

	if (task->waker == -1)
	{
		Oops ("Wake task %d from state %s in function %s has no waker\n",
			  task->pid, SrrTaskState2Str(task, task->state), whence);
	}
	
	wake_up_interruptible(&task->wq);

	/* This is not an oops.  It is the result of scheduling. */
	/*
	if (task->state != state)
		Oops ("Task %d changed state during wakeup from %s to %s\n",
			  task->pid, SrrTaskState2Str(task, state),
			  SrrTaskState2Str(task, task->state));
	*/
}

int SrrWakeFromPoll(SrrTask* task, int state)
{
	if (task->flags & _PPF_POLLING)
	{
		PDEBUG("TaskWakeFromPoll %d from %s to %s\n",
			   task->pid,
			   SrrTaskState2Str(task, task->state),
			   SrrTaskState2Str(NULL, state)
			   );

		// We cannot check for a wrong state here, as POLL is not
		// a possible value of task->state.  It is a flag (_PPF_POLLING)
		// on the task, with its task->state set to READY.
		// if (task->state == state)
		//	Oops ("Task %d was already in state %s on poll wake\n",
		//		  task->pid, SrrTaskState2Str(task, state));
		
		task->state = state;
		task->flags &= ~_PPF_POLLING;
		wake_up_interruptible (&task->poll_q);
		
		/* This is not an oops.  It is the result of scheduling. */
		/*
		if (task->state != state)
			Oops ("Task %d changed state during poll wakeup to %s\n",
				  task->pid, SrrTaskState2Str(task, task->state));
		*/
		return 1;
	}
	return 0;
}

void SrrClearPendingSignals (struct task_struct *task)
{
	if (task)
	{
		flush_signals (task);

		/* The argument list to recalc_sigpending() changed somewhere
		   along the way, and in a way that is not consistent from one
		   distribution to another.  We now try to determine whether
		   recalc_sigpending takes a task argument or not using a grep
		   in the top-level makefile.  The previous code is commented
		   out here as THIS_IS_WHAT_WE_USED_TO_USE */

#if RECALC_TSK
		recalc_sigpending (task);
#else
		recalc_sigpending ();
#endif

#ifdef THIS_IS_WHAT_WE_USED_TO_USE
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
		recalc_sigpending (task);
#else
		recalc_sigpending ();
#endif
#endif
	
		if (signal_pending(task))
			Oops ("Signal pending even after trying to clear\n");
	}
}

#ifdef TASK_SPECIFIC_LOCKS
void SrrTaskSleepAny(SrrTask* task, int state, int waitonqueue,
					 SPIN_FLAG_TYPE _spinflags, const char* whence, SrrTask* sleepOn)
#else
void SrrTaskSleepAny(SrrTask* task, int state, int waitonqueue,
					 SPIN_FLAG_TYPE _spinflags, const char* whence)
#endif

{
	int count = 0;

	PDEBUG("TaskSleep %d from %s to %s\n",
		task->pid,
		SrrTaskState2Str(task, task->state),
		SrrTaskState2Str(NULL, state));

	/* AST: If a death send is being performed on behalf of another
	   task, this is bound to happen.  The case exists for programs
	   using the pthread library. */
	if(task->pid != current->pid) {
		PDEBUG("current %d trying to sleep as task %d!\n",
			 current->pid, task->pid);
	}

	while(count<1000) {
		int		waitresult = 0;
		
		if (task->flags & _PPF_TERMING)
		{
			/* Remove all pending signals on tasks that are terminating */
			/* AST: The current task is not necessarily the terminating
			   task.  This is a problem introduced with pthreads. */
			/* AST: However, task->kern_task may not be valid.  See
			   CHANGES.pod for v 1.4.25.  So, we cannot clear pending
			   signals if the task is already gone. */

			task->kern_task = find_task_by_pid(task->pid);
			SrrClearPendingSignals (task->kern_task);

			/* We also need to clear the pending signals on the current
			   task, since it is standing in for the dying task.
			   I think. (AT) */
			SrrClearPendingSignals (current);
		}

		if (task->state != state)
			Oops ("Task state changed from %d to %d during sleep attempt\n",
				  state, task->state);

#ifdef TASK_SPECIFIC_LOCKS
		if (sleepOn != NULL){
			LKPRINT("Sleep returning sleepOn lock to %d", sleepOn->pid);
			Srr_ReturnTaskLock(sleepOn);
		}
		LKDEBUG("Unlocking process %d for sleep...\n", task->pid);
		Srr_TaskUnlock();
#else
		Srr_SpinUnlock();
#endif
		if (waitonqueue)      
			waitresult = wait_event_interruptible (task->wq,
									  task->recv_q != NULL ||
									  task->state != state);
		else
			waitresult = wait_event_interruptible (task->wq,task->state != state);
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("Locking process %d from sleep...\n", task->pid);
		Srr_TaskLock();
		if(sleepOn != NULL)
		{
			LKDEBUG("Relocking sleepOn %d after sleep\n", sleepOn->pid);
			Srr_WaitTaskLock(task, sleepOn);
		}
#else
		Srr_SpinLock();
#endif

		count++;

		if (waitresult == 0)
		{
				/* If the task transitioned into REPLY from SEND before we
			 * had a chance to block, update the target state and block
			 * anyway.  A transition from SEND to REPLY is not a reason
			 * to abandon the sleep.
			 */
			if (task->state == REPLY_BLOCKED && state == SEND_BLOCKED)
			{
				/* Oops("Task transitioned from SEND to REPLY during sleep\n"); */
				state = REPLY_BLOCKED;
				continue;
			}
		}
		else if (waitresult == -ERESTARTSYS)
		{
			//Oops ("Wait failed due to an interrupt\n");
			break;
		}

		/* The waitresult is 0 if the condition was already met.  In that
		   case another thread ran after the spin lock was released to
		   make the wait unnecessary.  We only worry about cases where
		   an error seems to have occurred. */
		if(waitresult != 0 &&
		   task->waker == -1 &&
		   task->kern_task && 
		   !signal_pending(task->kern_task))
		{
			if (waitonqueue)
			{
				/* We got woken up before we went to sleep */
				task->state = READY;
				KDEBUG ("No need to sleep in receive\n");
			}
			else
			{
				Oops("sleep failed: %s, wait %d, state %s, args %d, waker %d, sig %#lx of %#lx, # %d\n",
					 whence, waitresult,
					 SrrTaskState2Str(task, task->state),
					 task->args.pid, task->waker,
					 signal_set(current), signal_blocked(current), count);
				continue;
			}
		}
		break; /* Wait condition was  satisfied */
	}

	PDEBUG("TaskSleep woken by %d into state %s # %d\n",
		task->waker, SrrTaskState2Str(task, task->state), count);
}

int SrrTaskSleepRecvTimeout(SrrTask* task, int state, int waitonqueue,
					 SPIN_FLAG_TYPE _spinflags, const char* whence, int mx
#ifdef TASK_SPECIFIC_LOCKS
					 , SrrTask* sleepOn)
#else
					)
#endif
{
	int count = 0;
   long waitresult = 0;
	PDEBUG("TaskSleepTimeout %d from %s to %s\n",
		task->pid,
		SrrTaskState2Str(task, task->state),
		SrrTaskState2Str(NULL, state));

	/* AST: If a death send is being performed on behalf of another
	   task, this is bound to happen.  The case exists for programs
	   using the pthread library. */
	if(task->pid != current->pid) {
		PDEBUG("current %d trying to timeout sleep as task %d!\n",
			 current->pid, task->pid);
	}

	while(count < 1000)
   {
		waitresult = 0;
		
		if (task->flags & _PPF_TERMING)
		{
			/* Remove all pending signals on tasks that are terminating */
			/* AST: The current task is not necessarily the terminating
			   task.  This is a problem introduced with pthreads. */
			/* AST: However, task->kern_task may not be valid.  See
			   CHANGES.pod for v 1.4.25.  So, we cannot clear pending
			   signals if the task is already gone. */

			task->kern_task = find_task_by_pid(task->pid);
			SrrClearPendingSignals (task->kern_task);

			/* We also need to clear the pending signals on the current
			   task, since it is standing in for the dying task.
			   I think. (AT) */
			SrrClearPendingSignals (current);
		}

		if (task->state != state)
			Oops ("Task state changed from %d to %d during timeout sleep attempt\n",
				  state, task->state);
#ifdef TASK_SPECIFIC_LOCKS
		if(sleepOn != NULL)
		{
			LKDEBUG("Unlocking sleepOn for %d\n", sleepOn->pid);
			Srr_ReturnTaskLock(sleepOn);
		}
		LKDEBUG("Unlocking Task %d for RecvTimeout sleep...\n", task->pid);
		Srr_TaskUnlock();
#else
		Srr_SpinUnlock();
#endif
		if (waitonqueue)
			waitresult = wait_event_interruptible_timeout (task->wq,
									  task->recv_q != NULL ||
									  task->state != state,
									  msecs_to_jiffies((unsigned long)task->args.recvTimeout.time));
    else
			waitresult = wait_event_interruptible_timeout (task->wq,
												   task->state != state,
												   msecs_to_jiffies((unsigned long)task->args.recvTimeout.time));

    TSPRINT("%d msecs left after RecvTimeout sleep\n", jiffies_to_msecs(waitresult));
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("Relocking after RecvTimeout sleep\n");
		Srr_TaskLock();
		if(sleepOn != NULL)
		{
			LKDEBUG("Relocking sleepOn %d after sleep\n", sleepOn->pid);
			Srr_WaitTaskLock(task, sleepOn);
		}
#else
		Srr_SpinLock();
#endif
		count++;
    if(!mx)
    {
       task->args.recvTimeout.time = jiffies_to_msecs(waitresult);
    }
    else
    {
       task->args.recvmxTimeout.time = jiffies_to_msecs(waitresult);
    }
  
    if (waitresult > 0 ) /*( Condition occured */
	 {
			/* If the task transitioned into REPLY from SEND before we
			 * had a chance to block, update the target state and block
			 * anyway.  A transition from SEND to REPLY is not a reason
			 * to abandon the sleep.
			 */
			if (task->state == REPLY_BLOCKED && state == SEND_BLOCKED)
			{
				/* Oops("Task transitioned from SEND to REPLY during sleep\n"); */
				state = REPLY_BLOCKED;
				continue;
			}
         break;
		}
		else if (waitresult == -ERESTARTSYS)
		{
			//Oops ("Wait failed due to an interrupt\n");
			break;
	 }
#ifdef SRR_DEBUG
    if(waitresult > 0) {
      TSPRINT("RecvTimeout() awakened by event.\n");
    } else if(waitresult < 0) {
      TSPRINT("RecvTimeout() interrupted!\n");
    } else {
      TSPRINT("RecvTimeout() timed out.\n");
    }
#endif
		/* The waitresult is 0 if the condition was already met.  In that
		   case another thread ran after the spin lock was released to
		   make the wait unnecessary.  We only worry about cases where
		   an error seems to have occurred. */
		if(waitresult < 0 &&
		   task->waker == -1 &&
		   task->kern_task && 
		   !signal_pending(task->kern_task))
		{
			if (waitonqueue)
			{
				/* We got woken up before we went to sleep */
				task->state = READY;
				KDEBUG ("No need to sleep in receiveTimeout (wakeup before sleep)\n");
			}
			else
			{
				Oops("sleep failed: %s, wait %ld, state %s, args %d, waker %d, sig %#lx of %#lx, # %d\n",
					 whence, waitresult,
					 SrrTaskState2Str(task, task->state),
					 task->args.pid, task->waker,
					 signal_set(current), signal_blocked(current), count);
				continue;
			}
		}
      /* We must have timed out */
		break;
	}

	PDEBUG("TaskSleep woken by %d into state %s # %d\n",
		task->waker, SrrTaskState2Str(task, task->state), count);
  return waitresult;
}

int SrrTaskSleepSendTimeout(SrrTask* task, int state, int waitonqueue,	SPIN_FLAG_TYPE _spinflags,
							const char* whence, int mx
#ifdef TASK_SPECIFIC_LOCKS
							, SrrTask* sleepOn)
#else
							)
#endif

{
	int count = 0;
   long waitresult = 0;

	PDEBUG("TaskSleepTimeout %d from %s to %s\n",
		task->pid,
		SrrTaskState2Str(task, task->state),
		SrrTaskState2Str(NULL, state));

	/* AST: If a death send is being performed on behalf of another
	   task, this is bound to happen.  The case exists for programs
	   using the pthread library. */
	if(task->pid != current->pid)
   {
		PDEBUG("current %d trying to timeout sleep as task %d!\n",
			 current->pid, task->pid);
	}

	while(count <1000) {
		waitresult = 0;
		
		if (task->flags & _PPF_TERMING)
		{
			/* Remove all pending signals on tasks that are terminating */
			/* AST: The current task is not necessarily the terminating
			   task.  This is a problem introduced with pthreads. */
			/* AST: However, task->kern_task may not be valid.  See
			   CHANGES.pod for v 1.4.25.  So, we cannot clear pending
			   signals if the task is already gone. */

			task->kern_task = find_task_by_pid(task->pid);
			SrrClearPendingSignals (task->kern_task);

			/* We also need to clear the pending signals on the current
			   task, since it is standing in for the dying task.
			   I think. (AT) */
			SrrClearPendingSignals (current);
		}

		if (task->state != state)
			Oops ("Task state changed from %d to %d during timeout sleep attempt\n",
				  state, task->state);
#ifdef TASK_SPECIFIC_LOCKS
		if(sleepOn != NULL){
			LKDEBUG("Unlocking sleepOn %d for SendTimeout sleep\n", sleepOn->pid);
			Srr_ReturnTaskLock(sleepOn);
		}
		LKDEBUG("Unlocking task %d for SendTimeout sleep\n", task->pid);
		Srr_TaskUnlock();
#else
		Srr_SpinUnlock();
#endif
		if (waitonqueue)
			waitresult = wait_event_interruptible_timeout (task->wq,
									  task->recv_q != NULL ||
									  task->state != state,
									  msecs_to_jiffies((unsigned long)task->args.sendTimeout.time));
    else
			waitresult = wait_event_interruptible_timeout (task->wq,
												   task->state != state,
												   msecs_to_jiffies((unsigned long)task->args.sendTimeout.time));
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("Locking task %d for SendTimeout sleep\n", task->pid);
		Srr_TaskLock();
		if(sleepOn != NULL)
		{
			LKDEBUG("Locking sleepOn %d for SendTimeout sleep\n", sleepOn->pid);
			Srr_WaitTaskLock(task, sleepOn);
		}
#else
		Srr_SpinLock();
#endif
	  if(!mx)
      task->args.sendTimeout.time = jiffies_to_msecs(waitresult);
    else if(mx)
      task->args.sendmxTimeout.time = jiffies_to_msecs(waitresult);
    count++;

		if (waitresult > 0)
		{
			/* If the task transitioned into REPLY from SEND before we
			 * had a chance to block, update the target state and block
			 * anyway.  A transition from SEND to REPLY is not a reason
			 * to abandon the sleep.
			 */
			if (task->state == REPLY_BLOCKED && state == SEND_BLOCKED)
			{
				/* Oops("Task transitioned from SEND to REPLY during sleep\n"); */
				state = REPLY_BLOCKED;
				continue;
			}
 		}
		else if (waitresult == -ERESTARTSYS)
		{
			//Oops ("Wait failed due to an interrupt\n");
			break;
		}
       if(waitresult > 0) {
         PDEBUG("SendTimeout() awakened by event.\n");
       } else if(waitresult < 0) {
         PDEBUG("SendTimeout() interrupted!\n");
       } else {
         PDEBUG("SendTimeout() timed out!\n");
       }
		/* The waitresult is 0 if the condition was already met.  In that
		   case another thread ran after the spin lock was released to
		   make the wait unnecessary.  We only worry about cases where
		   an error seems to have occurred. */
		if((waitresult < 0 ) && (task->waker == -1) 
          && task->kern_task && 
		   !signal_pending(task->kern_task))
		{
			if (waitonqueue)
			{
				/* We got woken up before we went to sleep */
				task->state = READY;
				KDEBUG ("No need to sleep in sendTimeout (wakeup before sleep)\n");
			}
			else
			{
				Oops("sleep failed: %s, wait %ld, state %s, args %d, waker %d, sig %#lx of %#lx, # %d\n",
					 whence, waitresult,
					 SrrTaskState2Str(task, task->state),
					 task->args.pid, task->waker,
					 signal_set(current), signal_blocked(current), count);
				continue;
			}
		}
		break; /* Timeout occured */
	}

	PDEBUG("SendTimeoutSleep woken by %d into state %s # %d\n",
		task->waker, SrrTaskState2Str(task, task->state), count);
  return waitresult;
}


void SrrTaskSleep(SrrTask* task, int state, SPIN_FLAG_TYPE _spinflags,
	const char* whence
#ifdef TASK_SPECIFIC_LOCKS
		, SrrTask* sleepOn)
#else
		)
#endif
{
	SrrTaskSleepAny (task, state, 0, _spinflags, whence
#ifdef TASK_SPECIFIC_LOCKS
		,	sleepOn);
#else
	);
#endif
}

int SrrTaskSleepTimeout(SrrTask* task, int state, SPIN_FLAG_TYPE _spinflags, const char* whence
#ifdef TASK_SPECIFIC_LOCKS
		, SrrTask *sleepOn)
#else
		)
#endif
{
  return SrrTaskSleepSendTimeout(task, state, 0, _spinflags, whence, 0
#ifdef TASK_SPECIFIC_LOCKS
		 , sleepOn);
#else
		);
#endif
}

int SrrTaskSleepmxTimeout(SrrTask* task, int state, SPIN_FLAG_TYPE _spinflags, const char* whence
#ifdef TASK_SPECIFIC_LOCKS
			,	SrrTask* sleepOn)
#else
				)
#endif

{
  return SrrTaskSleepSendTimeout(task, state, 0, _spinflags, whence, 1
#ifdef TASK_SPECIFIC_LOCKS
		  	,  sleepOn);
#else
  	  	  );
#endif
}

void SrrTaskSleepRq(SrrTask* task, int state, SPIN_FLAG_TYPE _spinflags,
	const char* whence
#ifdef TASK_SPECIFIC_LOCKS
		,	SrrTask* sleepOn)
#else
		)
#endif
{
	SrrTaskSleepAny (task, state, 1, _spinflags, whence
#ifdef TASK_SPECIFIC_LOCKS
			, sleepOn);
#else
			);
#endif

}

int SrrTaskSleepRqmxTimeout(SrrTask* task, int state, SPIN_FLAG_TYPE _spinflags, const char* whence
#ifdef TASK_SPECIFIC_LOCKS
			, SrrTask *sleepOn)
#else
			)
#endif
{
	return SrrTaskSleepRecvTimeout (task, state, 1, _spinflags, whence, 1
#ifdef TASK_SPECIFIC_LOCKS
			, sleepOn);
#else
			);
#endif
}

int SrrTaskSleepRqTimeout(SrrTask* task, int state, SPIN_FLAG_TYPE _spinflags, const char* whence
#ifdef TASK_SPECIFIC_LOCKS
						, SrrTask* sleepOn)
#else
					)
#endif
{
	return SrrTaskSleepRecvTimeout (task, state, 1, _spinflags, whence, 0
#ifdef TASK_SPECIFIC_LOCKS
			, sleepOn);
#else
			);
#endif
}

/*
 * Various functions for copying buffers and MX records between
 * kernel and user spaces.  Currently the implementation always
 * transfers to a linear intermediate buffer held by the sender
 * in a Send() call, so we only have to worry about MX records
 * coming from user space, never from kernel space.  Also, we
 * never have to copy directly from one set of mx records to
 * another.
 */

/*
 * Copy from linear user space to linear buffer space
 */

void SrrResetBuffer (SrrTask* task)
{
  task->buf.datalen = 0;
}

int SrrCopyUserLinearToBuffer (SrrTask* fromtask, void* from,
							   int fromlen, SrrTask* totask, int offset)
{
	int			count;
	
	count = SrrBuf_OffsetRead (&totask->buf, offset, from, fromlen);
	return (count);
//  return fromlen;
}

int SrrCopyBufferToUserLinear (SrrTask* fromtask, int offset,
							   SrrTask* totask, void* to, int tolen)
{

  int			count;
	
	count = SrrBuf_OffsetWrite (&fromtask->buf, offset, to, tolen);
	return (count);
 
 // return tolen;
}

int SrrCopyBufferToUserMx (SrrTask* fromtask, int offset,
						   SrrTask* totask, struct _mxfer_entry *mx,
						   int numentries)
{
	int						count = 0, size;
	struct _mxfer_entry		*mxp, *mxentries;
	int						len, i;

	/* AST - We have to copy the array of mxfer_entry into kernel
	   space because we later reference into it. */
	len = sizeof (struct _mxfer_entry) * numentries;
  MXDEBUG("(BufferToUser) Size of mxfer entries:  %d\n", len);
	mxentries = (struct _mxfer_entry*) alloca (len);
	cpy_u2k (mxentries, mx, len);

	for (i = 0; i < numentries; i++)
	{
		mxp = &mxentries[i];
		size = (int)mxp->mxfer_len;
			
		if (size > 0)
		{
			len = SrrBuf_OffsetWrite(&fromtask->buf, offset,
									 (char *)mxp->addr, (size_t)size);

			/* Return the actual number of bytes in reply buffer on
			   a per-mxfer_entry basis */
			mxp->mxfer_len = len;
			
			if (len > 0)
			{
				count += len;
				offset += len;
			}
		}
	}

	/* if we want to return the actual sizes in the mx records,
	   we need to do this: */
#ifdef RETURN_SIZES
	len = sizeof (struct _mxfer_entry) * numentries;
	cpy_k2u (mx, mxentries, len);
#endif

	return (count);
}

int SrrCopyUserMxToBuffer (SrrTask* fromtask, struct _mxfer_entry *mx,
						   int numentries, SrrTask* totask, int offset)
{
	struct _mxfer_entry		*mxp, *mxentries;
	int						len, size, i, count;
	
	count = 0;
	
	/* AST - We have to copy the array of mxfer_entry into kernel
	   space because we later reference into it. */
	len = sizeof (struct _mxfer_entry) * numentries;
  MXDEBUG("(UsertoBuffer) Size of mxEntry:  %d\n", len);
	mxentries = (struct _mxfer_entry*) alloca (len);
	cpy_u2k (mxentries, mx, len);
	
	/* We do not check the client buffer size, as we can grow the
	   buffer to match. */
	for (i = 0; i < numentries; i++)
	{
		mxp = &mxentries[i];
		size = (int)mxp->mxfer_len;
		
		PDEBUG("Sendmx: Buffer %d, offset %d, len %d\n", i, offset, size);
		len = SrrBuf_OffsetRead(&totask->buf, offset,
								(char *)mxp->addr, (size_t)size);
		if (len > 0)
		{
			count += len;
			offset += len;
		}
		if (len != size)
		{
			Oops ("SrrCopyUserMxToBuffer: could not grow buffer to %d bytes\n",
				  offset + size);
			break;
		}
	}
	PDEBUG("Sendmx: buf is now %s\n", (char*)totask->buf.data);
	
	return (count);
}

/*
 * The Send() (pseudo-)system call.
 */

/*
 * These three functions are designed to allow two entries into the
 * general Send() code, one for sends originating from outside the
 * module, and one for sends originating inside the module.  This
 * only happens when a task dies and it must send its death message
 * to all other tasks.
 */
int SrrSendReadUser (SrrBuf* buf, char* from, size_t size)
{
	/* This will need re-thought when we do a single-copy version
	   of the module.  See SrrCopyUserLinearToBuffer. */
	return (SrrBuf_Read (buf, from, size));
}

int SrrSendWriteUser (SrrBuf* buf, char* to, size_t size)
{
	/* This will need re-thought when we do a single-copy version
	   of the module.  See SrrCopyBufferToUserLinear. */
	return (SrrBuf_Write (buf, to, size));
}

int SrrModSendTimeout(SrrTm* map, SrrTask* task){
  return SrrDoSendTimeout(map, task, SrrSendReadUser, SrrSendWriteUser);
}

int SrrModSend(SrrTm* map, SrrTask* task)
{
#ifndef DONT_COPY_MESSAGES
  return SrrDoSend (map, task, SrrSendReadUser, SrrSendWriteUser);
#else
  return SrrDoSend(map, task, NULL, NULL);
#endif
}

int SrrDoSendTimeout(SrrTm* map, SrrTask* task,
		pfnSrrBufRead readfn, pfnSrrBufWrite writefn)
{
	SrrSendTimeoutArgs*	args = &task->args.sendTimeout;
	SrrTask*		server;
	SrrTask**		taskp;
	int				retval = 0;
    int       sleepResult = 1;
  	Srr_UseSpinLock();

	/* Check for interrupt, and return immediately if interrupted.
	   The rest of this function does not "return" because it
	   needs to execute the Srr_SpinUnlock() at the end. */
#ifdef TASK_SPECIFIC_LOCKS
  LKDEBUG("Interruptible TaskLocking %d for SendTimeout", task->pid);
  atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing task lock to %d (sendTimeout)\n", atomic_read(&task->mutex.users));
#ifdef LOCK_VIA_SPINLOCKS
  Srr_TaskLockInterruptible();
  if(unlikely(0))
  {
#else
  if (Srr_TaskLockInterruptible() != 0)
	{
#endif /* LOCK_VIA_SPINLOCKS */
	  LKDEBUG("%d's TaskLockInterruptible interrupted! (sendTimeout)\n", task->pid);
    atomic_dec(&task->mutex.users);
    UDEBUG("Decrementing %d's lock to %d users(sendTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#else
	if (Srr_SpinLockInterruptible() != 0)
	{
#endif /* TASK_SPECIFIC_LOCKS */
		return (-EINTR);
	}
	
	PDEBUG("SendTimeout(pid %d, ssize %d, rsize %d, timeout %ldms)\n",
		args->pid, args->ssize, args->rsize, args->time);

	if(args->pid == task->pid)
	{
		retval = -EDEADLK;
		server = NULL;
	}
	else
	{
		/* find server */
		server = SrrTm_Get(map, args->pid);
		if(!server || IS_PROXY(server))
		{
			retval = -ESRCH;
		}
		else if (server->flags & _PPF_TERMING)
		{
			KDEBUG ("Send from %d to terminating task %d\n",
					task->pid, server->pid);
			retval = -ESRCH;
		}

	}

	/* Read the data from the sender */
	if (!retval)
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("GetTaskLock for %d's SendTimeout to %d...\n", task->pid, server->pid);
		atomic_inc(&server->mutex.users);
		Srr_GetTaskLock(server);
#endif
		PDEBUG("Send: Server is %d\n", server->pid);

		/* get sent data */
#ifndef DONT_COPY_MESSAGES	
		SrrResetBuffer (task);
		/* Read function should return bytes read, or -errno.  We reset
		   this to 0 for success, or -errno */
    SMPRINT(args->smsg, args->ssize,"Send to %d.  Hexdumping...\n", args->pid);
    retval = readfn(&task->buf, args->smsg, args->ssize);
		if (retval > 0)
			retval = 0;
#endif
	}

	if (!retval)
	{
		if (IS_PROXY(task))
			Oops ("Proxy %d performing a Send to %d\n",
				  task->pid, server->pid);
		SrrRq_Queue (server, task);

		PDEBUG("Send: Task is %d\n", task->pid);

		/* Set the task state before waking the server.  Then wake the
		 * server and go to sleep ourselves.  The act of waking the
		 * server can cause rescheduling so between the wake and the
		 * sleep the server could run and change this task's state to
		 * reply blocked, or even to READY.
		 */
		task->state = SEND_BLOCKED;
		task->waker = -1;
		
		/* wake a server if it is blocked on select() otherwise try a wake
		   from recieve. */
		if (SrrWakeFromPoll (server, READY))
		{
			/* Nothing to do */
		}
		/* wake server if its recv blocked on this task, or 0 (any task) */
		else if(server->state == RECV_BLOCKED)
		{
			if(server->args.recv.pid == task->pid ||
			   server->args.recv.pid == 0)
			{
				PDEBUG("Send: wake server %d\n", server->pid);
				SrrTaskWake(server, READY, __FUNCTION__);
			}
		}

		/*
		 * In 2.6 kernels, SrrTaskWake will not only mark the woken task
		 * for scheduling, it will also invoke the scheduler.
		 */
/*
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d returning task lock to %d for SendTimeout sleep...\n", task->pid, server->pid);
		Srr_ReturnTaskLock(server);
#endif
*/
		sleepResult = SrrTaskSleepTimeout(task, SEND_BLOCKED, _spinflags, __FUNCTION__
#ifdef TASK_SPECIFIC_LOCKS
				, server);
#else
		);
#endif
/*
#ifdef TASK_SPECIFIC_LOCKS
    LKDEBUG("%d GetTaskLock from %d after sleep for SendTimeout\n", task->pid, server->pid);
		Srr_WaitTaskLock(task, server);
#endif
*/
		/* check for signal */
		if(task->kern_task && signal_pending (task->kern_task))
		{
			PDEBUG("Send() signal %#lx blocked %#lx\n",
				   signal_set(task->kern_task),
				   signal_blocked(task->kern_task));
		
			if(task->state == SEND_BLOCKED) 
			{
				/* deque ourselves */
				taskp = SrrRq_Find(server, task->pid);

				if (taskp && *taskp)
					SrrRq_Unlink(taskp);
				else
					Oops ("A: task %d is send blocked on %d, but %d not queued\n",
						  task->pid, server->pid, task->pid);
			}
		
			task->state = READY;
      retval = -EINTR;
		}

    /* If we were woken up by the wait_event_interruptible_timeout timer */
    else if(sleepResult == 0 && task->state == SEND_BLOCKED && 
        task->kern_task && !signal_pending(task->kern_task))
    {
      PDEBUG("SendTimeout() woken up by timer\n");
      if(task->state == SEND_BLOCKED)
      {
      taskp = SrrRq_Find(server, task->pid);

      if(taskp && *taskp)
        SrrRq_Unlink(taskp);
      else
        Oops ("A: task %d is send blocked on %d, but %d not queued (timeout)\n",
            task->pid, server->pid, task->pid);
      }
      task->state = READY;
			args->time = 0;
      retval = -ETIME;
    }

		/* if we're still blocked then the process we're blocked on has died */
		else if (task->state != READY)
		{
			/* sleep returned for unknown reasons. This should not happen */
			if(task->state != RELEASED)
			{
				Oops("Send(%s): send sleep failed: state %s on pid %d!\n",
					 (readfn == SrrSendReadUser ? "normal" : "on death"),
					 SrrTaskState2Str(task, task->state), args->pid);

				/* clean up - deque ourselves (if necessary) and make READY */
				if(task->state == SEND_BLOCKED)
				{
					taskp = SrrRq_Find(server, task->pid);
					if (taskp && *taskp)
						SrrRq_Unlink(taskp);
					else
						Oops ("C: task %d is send blocked on %d, but %d not queued\n",
							  task->pid, server->pid, task->pid);
				}
				task->state = READY;
				retval = -EAGAIN;
			}
			else
			{
				task->state = READY;
				retval = -ESRCH;
			}
		}
		else
		{
//			if (readfn == SrrBuf_LocalRead)
//			{
//				Oops ("Writing a task death message\n");
//			}
		
			//			Oops("Send: reply\n");
			/* get replied data */
#ifndef DONT_COPY_MESSAGES
      args->rsize = writefn (&task->buf, args->rmsg, args->rsize);
#endif
      SMPRINT(args->rmsg, args->rsize,"Reply from %d.  Hexdumping...\n", args->pid);
		}
#ifdef TASK_SPECIFIC_LOCKS
    LKDEBUG("%d returning task lock to %d after SendTimeout\n", task->pid, server->pid);
		Srr_ReturnTaskLock(server);
		atomic_dec(&server->mutex.users);
    UDEBUG("Decrementing %d's lock to %d users (sendTimeout)\n", server->pid, atomic_read(&server->mutex.users));
#endif
	}
#ifdef TASK_SPECIFIC_LOCKS
	else
	{
		if(server)
    {
      LKDEBUG("%d returning task lock to %d after buffer error on SendTimeout\n", task->pid, server->pid);
		  Srr_ReturnTaskLock(server);
		  atomic_dec(&server->mutex.users);
      UDEBUG("Decrementing %d's users to %d (sendTimeout)\n", server->pid, atomic_read(&server->mutex.users));
    }
	}
	LKDEBUG("%d unlocking after sendTimeout...\n", task->pid);
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing %d's users to %d (sendTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#else
	Srr_SpinUnlock ();
#endif
	return retval;
}

int SrrDoSend(SrrTm* map, SrrTask* task,
		pfnSrrBufRead readfn, pfnSrrBufWrite writefn)
{
	SrrSendArgs*	args = &task->args.send;
	SrrTask*		server;
	SrrTask**		taskp;
	int				retval = 0;
	Srr_UseSpinLock();

	/* Check for interrupt, and return immediately if interrupted.
	   The rest of this function does not "return" because it
	   needs to execute the Srr_SpinUnlock() at the end. */
#ifdef TASK_SPECIFIC_LOCKS
  LKPRINT("Locking self for Send\n");
  atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing task lock to %d (Send)\n", atomic_read(&task->mutex.users));
  TTP();
#ifdef LOCK_VIA_SPINLOCKS
  Srr_TaskLockInterruptible();
  if(unlikely(0))
  {
#else
  if (Srr_TaskLockInterruptible() != 0)
	{
#endif /* LOCK_VIA_SPINLOCKS */
		LKDEBUG("%d's TaskLockInterruptible interrupted (send)!\n", task->pid);
    atomic_dec(&task->mutex.users);
    UDEBUG("Decrementing %d's lock users to %d (send)\n", task->pid, atomic_read(&task->mutex.users));
#else
	if (Srr_SpinLockInterruptible() != 0)
	{
#endif /* TASK_SPECIFIC_LOCKS */
		return (-EINTR);
	}
	PDEBUG("Send(pid %d, ssize %d, rsize %d)\n",
		args->pid, args->ssize, args->rsize);

	if(args->pid == task->pid)
	{
		retval = -EDEADLK;
		server = NULL;
	}
	else
	{
		/* find server */
		server = SrrTm_Get(map, args->pid);
    if(!server || IS_PROXY(server))
		{
			retval = -ESRCH;
		}
		else if (server->flags & _PPF_TERMING)
		{
			KDEBUG ("Send from %d to terminating task %d\n",
					task->pid, server->pid);
			retval = -ESRCH;
		}
  }	


	/* Read the data from the sender */

	if (!retval)
	{
#ifdef TASK_SPECIFIC_LOCKS
    TTP();
		LKPRINT("Locking task %d for send...\n", server->pid);
    atomic_inc(&server->mutex.users);
    UDEBUG("Incrementing server lock to %d (Send)\n", atomic_read(&server->mutex.users));
    Srr_WaitTaskLock(task, server);
#endif
		PDEBUG("Send: Server is %d\n", server->pid);

		/* get sent data */
    retval = 0;
#ifndef DONT_COPY_MESSAGES
		SrrResetBuffer (task);
		/**
     * Read function should return bytes read, or -errno.  We reset
     * this to 0 for success, or -errno 
     **/
      SMPRINT(args->smsg, args->ssize,"Send to %d.  Hexdumping...\n", args->pid);
		retval = readfn(&task->buf, args->smsg, args->ssize);
		if (retval > 0)
			retval = 0;
#endif /* DONT_COPY_MESSAGES */
	}
	if (!retval)
	{
		if (IS_PROXY(task))
			Oops ("Proxy %d performing a Send to %d\n",
				  task->pid, server->pid);
		SrrRq_Queue (server, task);

		PDEBUG("Send: Task is %d\n", task->pid);

		/* Set the task state before waking the server.  Then wake the
		 * server and go to sleep ourselves.  The act of waking the
		 * server can cause rescheduling so between the wake and the
		 * sleep the server could run and change this task's state to
		 * reply blocked, or even to READY.
		 */
		task->state = SEND_BLOCKED;
		task->waker = -1;
		
		/* wake a server if it is blocked on select() otherwise try a wake
		   from recieve. */
		if (SrrWakeFromPoll (server, READY))
		{
			/* Nothing to do */
		}
		/* wake server if its recv blocked on this task, or 0 (any task) */
		else if(server->state == RECV_BLOCKED)
		{
			if(server->args.recv.pid == task->pid ||
			   server->args.recv.pid == 0)
			{
				PDEBUG("Send: wake server %d\n", server->pid);
				SrrTaskWake(server, READY, __FUNCTION__);
			}
		}

		/*
		 * In 2.6 kernels, SrrTaskWake will not only mark the woken task
		 * for scheduling, it will also invoke the scheduler.
		 */
		SrrTaskSleep(task, SEND_BLOCKED, _spinflags, __FUNCTION__
#ifdef TASK_SPECIFIC_LOCKS
      , server);
#else
      );
#endif
		/* check for signal */
		if(task->kern_task && signal_pending (task->kern_task))
		{
			PDEBUG("Send() signal %#lx blocked %#lx\n",
				   signal_set(task->kern_task),
				   signal_blocked(task->kern_task));
		
			if(task->state == SEND_BLOCKED) 
			{
				/* deque ourselves */
				taskp = SrrRq_Find(server, task->pid);

				if (taskp && *taskp)
					SrrRq_Unlink(taskp);
				else
					Oops ("A: task %d is send blocked on %d, but %d not queued\n",
						  task->pid, server->pid, task->pid);
			}
		
			task->state = READY;
			retval = -EINTR;
		}
		/* if we're still blocked then the process we're blocked on has died */
		else if (task->state != READY)
		{
			/* sleep returned for unknown reasons. This should not happen */
			if(task->state != RELEASED)
			{
				Oops("Send(%s): send sleep failed: state %s on pid %d!\n",
					 (readfn == SrrSendReadUser ? "normal" : "on death"),
					 SrrTaskState2Str(task, task->state), args->pid);

				/* clean up - deque ourselves (if necessary) and make READY */
				if(task->state == SEND_BLOCKED)
				{
					taskp = SrrRq_Find(server, task->pid);
					if (taskp && *taskp)
						SrrRq_Unlink(taskp);
					else
						Oops ("C: task %d is send blocked on %d, but %d not queued\n",
							  task->pid, server->pid, task->pid);
				}
				task->state = READY;
				retval = -EAGAIN;
			}
			else
			{
				task->state = READY;
				retval = -ESRCH;
			}
		}
		else
		{
//			if (readfn == SrrBuf_LocalRead)
//			{
//				Oops ("Writing a task death message\n");
//			}
		
			//			Oops("Send: reply\n");
			/* get replied data */
#ifndef DONT_COPY_MESSAGES
			args->rsize = writefn (&task->buf, args->rmsg, args->rsize);
#endif
      SMPRINT(args->rmsg, args->rsize,"Reply from %d.  Hexdumping...\n", args->pid);

		}
#ifdef TASK_SPECIFIC_LOCKS
    TTP();
		LKPRINT("Unlocking task %d after send completion\n", server->pid);
    Srr_ReturnTaskLock(server);
    atomic_dec(&server->mutex.users);
    UDEBUG("Decrementing server lock to %d (send)\n", atomic_read(&server->mutex.users));
#endif /* TASK_SPECIFIC_LOCKS */
   }
#ifdef TASK_SPECIFIC_LOCKS
	else
	{
    if(server)
    {
	    TTP();	
      LKPRINT("Unlocking task %d after failed send\n", server->pid);
      Srr_ReturnTaskLock(server);
      atomic_dec(&server->mutex.users);
      UDEBUG("Decrementing server lock to %d (send)\n", atomic_read(&server->mutex.users));
    }
	}
  LKPRINT("Unlocking self after send\n");
  Srr_TaskUnlock();	
	atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing task lock to %d (send)\n", atomic_read(&task->mutex.users)); 
#else
	Srr_SpinUnlock ();
#endif /* TASK_SPECIFIC_LOCKS */

	return retval;
}

int SrrRecvTimeout(SrrTm* map, SrrTask* task)
{
	SrrRecvTimeoutArgs*	args = &task->args.recvTimeout;
	SrrTask**		clientp = 0;
	SrrTask*		client = 0;
	int				retval = 0;
	int       sleepResult = 1;
  Srr_UseSpinLock();
	
	PDEBUG("RecvTimeout(pid %d, size %d, timeout %ldms)\n", args->pid, args->size, args->time);

	/* Check for interrupt, and return immediately if interrupted.
	   The rest of this function does not "return" because it
	   needs to execute the Srr_SpinUnlock() at the end. */
#ifdef TASK_SPECIFIC_LOCKS
	LKPRINT("Locking self for RecvTimeout");
	atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing %d's lock to %d users (RecvTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#ifdef LOCK_VIA_SPINLOCKS
  Srr_TaskLockInterruptible();
  if(unlikely(0))
  {
#else
  if (Srr_TaskLockInterruptible() != 0)
	{
#endif /* LOCK_VIA_SPINLOCKS */
		LKPRINT("Interrupted waiting on own lock (RecvTimeout)\n");
		atomic_dec(&task->mutex.users);
    UDEBUG("Decrementing %d's lock to %d users (RecvTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#else
	if (Srr_SpinLockInterruptible() != 0)
	{
#endif /* TASK_SPECIFIC_LOCKS */
		return (-EINTR);
	}
	
	if(args->pid == current->pid)
	{
		retval = -EDEADLK;
	}
	/* make sure pid exists (if non-zero).  If the client exists
	   but is terminating, forget it. */
	else if(args->pid != 0 && !(client = SrrTm_Get(map, args->pid)))
	{
		retval = -ESRCH;
	}
	else if (client && (client->flags & _PPF_TERMING))
	{
		KDEBUG ("ReceiveTimeout by %d on terminating task %d\n",
				task->pid, client->pid);
		retval = -ESRCH;
	}

	if (!retval)
	{
		/* search for pid on our recv queue */
		if(args->pid == 0)	/* take the first one */
			clientp = SrrRq_First (task);
		else 				/* search for pid */
			clientp = SrrRq_Find(task, args->pid);
		client = *clientp;

		/* If we are creceived return with no client */

		if (!client && args->nowait)
		{
			retval = -ENOMSG;
		}
	}

	if (!retval)
	{
		/* wait for a client, if didn't find one */
		if(!client)
		{
			task->state = RECV_BLOCKED;
			task->waker = -1;
			sleepResult = SrrTaskSleepRqTimeout (task, RECV_BLOCKED, _spinflags, __FUNCTION__
#ifdef TASK_SPECIFIC_LOCKS
					, client);
#else
					);
#endif
			/* check for signal */
			if(signal_pending (current))
			{
				PDEBUG("RecvTimeout() signal %#lx blocked %#lx\n",
					   signal_set(current), signal_blocked(current));

				task->state = READY;
				retval = -EINTR;
			}
      if(sleepResult == 0){
        PDEBUG("RecvTimeout() awakened by timer.  Setting state to READY...\n");
        task->state = READY;
        args->time = 0;
        retval = -ETIME;
      }
    }
	}

	if (!retval)
	{
		/* check if we were released because of a task death */
		if(task->state != READY)
		{
			/* sleep returned for unknown reasons... ;-( */
			if(task->state != RELEASED)
			{
				Oops("recv timeout sleep failed: state %s on pid %d!\n",
					 SrrTaskState2Str(task, task->state), args->pid);
				task->state = READY;
/*				retval = -EAGAIN;*/
            retval = -EINTR;
			}
			else
			{
				task->state = READY;
				retval = -ESRCH;
			}
		}
	}

	if (!retval)
	{
		/* search for client, if not previously found */
		if(!client)
		{
			if(args->pid == 0)		/* take the first one */
				clientp = SrrRq_First(task);
			else					/* search for pid */
				clientp = SrrRq_Find(task, args->pid);

			/* clientp should always exist */
			if (!clientp)
			{
				retval = -ESRCH;
				Oops("ReceiveTimeout succeeded, but no sender: task=%d, state=%s, target=%d\n",
					 task->pid, SrrTaskState2Str(task, task->state), args->pid);
			}
			else
			{
				client = *clientp;
			}
		}
	
		if(!client)
		{
			/* This line generates an unused variable warning if we
			   compile without debugging enabled. */
			SrrTask* waker = task->waker ? SrrTm_Get(map, task->waker) : 0;
			/*
			 * This is not an Oops.  It is normal to be woken by a
			 * sender that is READY if the sender was in the SEND
			 * state when a signal caused it to drop out.
			 */
			PDEBUG("ReceiveTimeout(%d, ...)  woke by %d, which is now %s!\n",
				   args->pid, task->waker,
				   waker ? SrrTaskState2Str(waker, waker->state) : "dead");
			// If we wanted to differentiate a broken SEND from a
			// failed receive on a non-zero PID, we would do this, but
			// it breaks the test program.  What does QNX4 do?
			// Well, QNX does not allow a Receive to fail here, so we
			// will return -EAGAIN and let the library handle it.
			if (waker)
				retval = -EAGAIN;
			else
				retval = -ESRCH;
		}
	}

	if (!retval)
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKPRINT("Getting %d's lock for RecvTimeout\n", client->pid);
		atomic_inc(&client->mutex.users);
    UDEBUG("Incrementing %d's lock to %d users (recvTimeout)\n", client->pid, atomic_read(&client->mutex.users));
		Srr_WaitTaskLock(task, client);
#endif
		PDEBUG("RecvTimeout() from %d in state %s\n",
			   client->pid, SrrTaskState2Str(client, client->state));
	
		if(client->state != SEND_BLOCKED){
			Oops("RecvTimeout() client %d is %s!\n",
				 client->pid, SrrTaskState2Str(client, client->state));
			retval = -EINTR;
		}
		else
		{
			/* We absolutely must dequeue a proxy before we get the user
			   data, or we have a window where a timer could fire again
			   when we reschedule due to user-to-kernel copying in the
			   SrrCopyBufferToUserLinear call below. */
	
			if(!IS_PROXY(client))
			{ /* deque task */
				if (clientp && *clientp)
					SrrRq_Unlink(clientp);
				else
					Oops ("D: task %d is send blocked on %d, but %d not queued\n",
						  client->pid, -1, client->pid);
				client->state = REPLY_BLOCKED;
			}
			else
			{
				/* deque proxy in any case, then re-queue if the count is
				   non-zero.  This gives other processes a fair-share
				   chance to send messages.  */

				if (clientp && *clientp)
				{
					/* Decrement the proxy count.  If it is not yet zero,
					   then re-queue the proxy onto the task.  Use the
					   Ndebug version of the queue function because we
					   know we are queuing a proxy that is not queued but
					   has a non-zero count. */

					if (!atomic_dec_and_test (&client->count))
					{
						SrrRq_UnlinkNdebug (clientp);
						SrrRq_QueueNdebug (task, client);
					}
					else
					{
						SrrRq_Unlink (clientp);
						client->state = READY;
					}
				}
				else
				{
					Oops ("E: task %d is send blocked on %d, but %d not queued\n",
						  client->pid, -1, client->pid);
					client->state = READY;
				}
			}

			/* get sent data */

//			Oops("SrrModReadMsg to len,offset %d,%d and length %d\n",
//				 client->buf.datalen, 0, args->size);
#ifndef DONT_COPY_MESSAGES
      args->size = SrrCopyBufferToUserLinear (client, 0, task,
													args->msg, args->size);
#endif
			RMPRINT(args->msg, args->size, "Hexdumping...\n");
      /* client buffer not yet written to  - 2000/04/01 - Alessandro Sala */
			client->flags &= ~_PPF_KBUF_WRITTEN;
			retval = client->pid;
		}
#ifdef TASK_SPECIFIC_LOCKS
		LKPRINT("Returning %d's lock after RecvTimeout\n", client->pid);
		Srr_ReturnTaskLock(client);
		atomic_dec(&client->mutex.users);
    UDEBUG("Decrementing %d's lock to %d users (recvTimeout)\n", client->pid, atomic_read(&client->mutex.users));
#endif
	}
#ifdef TASK_SPECIFIC_LOCKS
	LKPRINT("Unlocking self after RecvTimeout\n");
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing %d's lock to %d users (RecvTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#else
	Srr_SpinUnlock();
#endif
	return (retval);
}



/*
 * The Receive() (pseudo-)system call.
 */

int SrrRecv(SrrTm* map, SrrTask* task)
{

  SrrRecvArgs*	args = &task->args.recv;
	SrrTask**		clientp = 0;
	SrrTask*		client = 0;
	int				retval = 0;
#ifdef TASK_SPECIFIC_LOCKS
  int       hasRecvLock = 0;
#endif
	Srr_UseSpinLock();
	PDEBUG("Recv(pid %d, size %d)\n", args->pid, args->size);

	/* Check for interrupt, and return immediately if interrupted.
	   The rest of this function does not "return" because it
	   needs to execute the Srr_SpinUnlock() at the end. */
#ifdef TASK_SPECIFIC_LOCKS
  LKPRINT("Waiting on own lock for Recv\n");
  atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing task lock to %d (Recv)\n", atomic_read(&task->mutex.users));
#ifdef LOCK_VIA_SPINLOCKS
  Srr_TaskLockInterruptible();
  if(unlikely(0))
  {
#else
  if(Srr_TaskLockInterruptible() != 0)
	{
#endif /* LOCK_VIA_SPINLOCKS */
	  LKDEBUG("%d's TaskLockInterruptible interrupted (Recv)!\n", task->pid);
    atomic_dec(&task->mutex.users);
    UDEBUG("Decrementing %'d lock to %d users (Recv)\n", task->pid, atomic_read(&task->mutex.users));
#else
	if (Srr_SpinLockInterruptible() != 0)
	{
#endif /* TASK_SPECIFIC_LOCKS */
		LKPRINT("Wait on own lock interrupted! (Recv)\n");
		return (-EINTR);
	}
	
  if(args->pid == current->pid)
	{
		retval = -EDEADLK;
	}
	/* make sure pid exists (if non-zero).  If the client exists
	   but is terminating, forget it. */
	else if(args->pid != 0 && !(client = SrrTm_Get(map, args->pid)))
	{
		retval = -ESRCH;
	}
	else if (client && (client->flags & _PPF_TERMING))
	{
		KDEBUG ("Receive by %d on terminating task %d\n",
				task->pid, client->pid);
		retval = -ESRCH;
	}

	if (!retval)
	{
		/* search for pid on our recv queue */
		if(args->pid == 0)	/* take the first one */
			clientp = SrrRq_First (task);
		else 				/* search for pid */
			clientp = SrrRq_Find(task, args->pid);
		client = *clientp;

		/* If we are creceived return with no client */

		if (!client && args->nowait)
		{
			retval = -ENOMSG;
		}
	}

	if (!retval)
	{
		/* wait for a client, if didn't find one */
		if(!client)
		{
			task->state = RECV_BLOCKED;
			task->waker = -1;
			SrrTaskSleepRq (task, RECV_BLOCKED, _spinflags, __FUNCTION__
#ifdef TASK_SPECIFIC_LOCKS
					, client);
#else
			);
#endif
		
			/* check for signal */
			if(signal_pending (current))
			{
				PDEBUG("Recv() signal %#lx blocked %#lx\n",
					   signal_set(current), signal_blocked(current));

				task->state = READY;
				retval = -EINTR;
			}
		}
	}

	if (!retval)
	{
		/* check if we were released because of a task death */
		if(task->state != READY)
		{
			/* sleep returned for unknown reasons... ;-( */
			if(task->state != RELEASED)
			{
				Oops("recv sleep failed: state %s on pid %d!\n",
					 SrrTaskState2Str(task, task->state), args->pid);
				task->state = READY;
				retval = -EAGAIN;
       //     retval = -EINTR;
			}
			else
			{
				task->state = READY;
				retval = -ESRCH;
			}
		}
	}

	if (!retval)
	{
		/* search for client, if not previously found */
		if(!client)
		{
			if(args->pid == 0)		/* take the first one */
				clientp = SrrRq_First(task);
			else					/* search for pid */
				clientp = SrrRq_Find(task, args->pid);

			/* clientp should always exist */
			if (!clientp)
			{
				retval = -ESRCH;
				Oops("Receive succeeded, but no sender: task=%d, state=%s, target=%d\n",
					 task->pid, SrrTaskState2Str(task, task->state), args->pid);
			}
			else
			{
				client = *clientp;
			}
		}
	
		if(!client)
		{
			/* This line generates an unused variable warning if we
			   compile without debugging enabled. */
			SrrTask* waker = task->waker ? SrrTm_Get(map, task->waker) : 0;
			/*
			 * This is not an Oops.  It is normal to be woken by a
			 * sender that is READY if the sender was in the SEND
			 * state when a signal caused it to drop out.
			 */
			PDEBUG("Receive(%d, ...)  woke by %d, which is now %s!\n",
				   args->pid, task->waker,
				   waker ? SrrTaskState2Str(waker, waker->state) : "dead");
			// If we wanted to differentiate a broken SEND from a
			// failed receive on a non-zero PID, we would do this, but
			// it breaks the test program.  What does QNX4 do?
			// Well, QNX does not allow a Receive to fail here, so we
			// will return -EAGAIN and let the library handle it.
			if (waker)
				retval = -EAGAIN;
			else
				retval = -ESRCH;
		}
	}

	if (!retval)
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKPRINT("Locking task %d for Recv\n", client->pid);
    atomic_inc(&client->mutex.users);
  	UDEBUG("Incrementing client lock to %d (Recv)\n", atomic_read(&client->mutex.users));
		Srr_WaitTaskLock(task, client);
    hasRecvLock = 1;
#endif
		PDEBUG("Recv() from %d in state %s\n",
			   client->pid, SrrTaskState2Str(client, client->state));
	
		if(client->state != SEND_BLOCKED)
		{
			Oops("Recv() client %d is %s!\n",
				 client->pid, SrrTaskState2Str(client, client->state));
			retval = -EINTR;
		}
		else
		{
			/* We absolutely must dequeue a proxy before we get the user
			   data, or we have a window where a timer could fire again
			   when we reschedule due to user-to-kernel copying in the
			   SrrCopyBufferToUserLinear call below. */
	
			if(!IS_PROXY(client))
			{ /* deque task */
				if (clientp && *clientp)
					SrrRq_Unlink(clientp);
				else
					Oops ("D: task %d is send blocked on %d, but %d not queued\n",
						  client->pid, -1, client->pid);
				client->state = REPLY_BLOCKED;
			}
			else
			{
				/* deque proxy in any case, then re-queue if the count is
				   non-zero.  This gives other processes a fair-share
				   chance to send messages.  */

				if (clientp && *clientp)
				{
					/* Decrement the proxy count.  If it is not yet zero,
					   then re-queue the proxy onto the task.  Use the
					   Ndebug version of the queue function because we
					   know we are queuing a proxy that is not queued but
					   has a non-zero count. */

					if (!atomic_dec_and_test (&client->count))
					{
						SrrRq_UnlinkNdebug (clientp);
						SrrRq_QueueNdebug (task, client);
					}
					else
					{
						SrrRq_Unlink (clientp);
						client->state = READY;
					}
				}
				else
				{
					Oops ("E: task %d is send blocked on %d, but %d not queued\n",
						  client->pid, -1, client->pid);
					client->state = READY;
				}
			}

			/* get sent data */

//			Oops("SrrModReadMsg to len,offset %d,%d and length %d\n",
//				 client->buf.datalen, 0, args->size);
#ifndef DONT_COPY_MESSAGES
      args->size = SrrCopyBufferToUserLinear (client, 0, task,
													args->msg, args->size);
#endif
			RMPRINT(args->msg, args->size, "Hexdumping...\n");
			/* client buffer not yet written to  - 2000/04/01 - Alessandro Sala */
			client->flags &= ~_PPF_KBUF_WRITTEN;
			retval = client->pid;

		}
#ifdef TASK_SPECIFIC_LOCKS
		LKPRINT("Unlocking task %d after Recv\n", client->pid);
  	Srr_ReturnTaskLock(client);
    atomic_dec(&client->mutex.users);
    UDEBUG("Decrementing client lock to %d (Recv)\n", atomic_read(&client->mutex.users));
    hasRecvLock = 0;
#endif
	}
#ifdef TASK_SPECIFIC_LOCKS
	if(hasRecvLock)
  {
    LKPRINT("Unlocking client %d after Recv\n", client->pid);
    Srr_ReturnTaskLock(client);
    atomic_dec(&client->mutex.users);
    UDEBUG("Decrementing client lock to %d (Recv)\n", atomic_read(&client->mutex.users));
  }
  LKPRINT("Unlocking self after Recv\n");
	Srr_TaskUnlock();
  atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing task lock to %d (Recv)\n", atomic_read(&task->mutex.users));
#else
	Srr_SpinUnlock ();
#endif
	return (retval);
}

/*
 * The Reply() (pseudo-)system call.
 */

int SrrModReply(SrrTm* map, SrrTask* task)
{
	SrrReplyArgs*	args = &task->args.reply;
	SrrTask*		client;
	int				retval = 0;
	Srr_UseSpinLock();
	
	PDEBUG("Reply(pid %d, size %d)\n", args->pid, args->size);
	/* find client, confirm it's blocked on us */
	client = SrrTm_Get(map, args->pid);

	if(!client)
  {
		return -ESRCH;
	}
  /* replying to *your* proxy is a null op */
	else if(client->target == task)
  {
		return 0;
	}
  /* Replying to somebody else's proxy is just plain wrong */
	else if (client->target)
	{
    return -EINVAL;
	}
	/* Reply is not supposed to block or fail so we are going to use
	   the non-interruptible lock here.  */
#ifdef TASK_SPECIFIC_LOCKS
	LKPRINT("Locking self for Reply\n");
	atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing task lock to %d (Reply)\n", atomic_read(&task->mutex.users));
	Srr_TaskLock();
	LKPRINT("Locking task %d for Reply\n", client->pid);
	atomic_inc(&client->mutex.users);
  UDEBUG("Incrementing client lock to %d (Reply)\n", atomic_read(&client->mutex.users));
	Srr_WaitTaskLock(task, client);
#else
	Srr_SpinLock();
#endif
	
	if(client->state != REPLY_BLOCKED || client->args.send.pid != current->pid)
	{
		retval = -ESRCH;
	}
	else
	{
		/* 2000/04/01 - Alessandro Sala
		 * get reply data. use OffsetRead if the buffer already
		 * contains valid data (written by Writemsg()), otherwise use
		 * the standard Read 2002/03/15 - Andrew Thomas
		 * Use generic copy instead.
		 */
#ifndef DONT_COPY_MESSAGES
	 SrrResetBuffer (client);
    if (SrrCopyUserLinearToBuffer (task, args->msg, args->size, client, 0)
			!= args->size){
#else
    if(unlikely(0))
    {
#endif
#ifdef TASK_SPECIFIC_LOCKS
      UDEBUG("Copy fail, returning locks in Reply\n");
		Srr_ReturnTaskLock(client);
		atomic_dec(&client->mutex.users);
      Srr_TaskUnlock();
      atomic_dec(&task->mutex.users);
#endif
			return -ENOMEM;
		}
    
		/* wake up client */
		if (!retval)
		{
			SrrTaskWake(client, READY, __FUNCTION__);
		}
	}
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d returning %d's lock after Reply\n", task->pid, client->pid);
	Srr_ReturnTaskLock(client);
	atomic_dec(&client->mutex.users);
	UDEBUG("Decrementing client lock to %d (reply)\n", atomic_read(&client->mutex.users));
   LKDEBUG("%d unlocking after Reply\n", task->pid);
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing task lock to %d (reply)\n", atomic_read(&task->mutex.users));
#else
	Srr_SpinUnlock ();
#endif
  return retval;
}

/* 2005/09/05 - Len Meakin
 * The Relay() (pseudo-)system call.
 */

int SrrModRelay(SrrTm* map, SrrTask* task)
{
	SrrRelayArgs*	args 	= &task->args.relay;
	SrrTask*		target 	= NULL;
	SrrTask*		source 	= NULL;
	int				retval 	= 0;
	Srr_UseSpinLock();
	
	/*-- Debug where we are -------------------------------------------*/
	PDEBUG("Relay(source %d, target %d)\n", args->srcpid, args->tgtpid);
	
	/*-- Ensure process is not trying to Relay to its self -----------*/
	if (args->tgtpid == task->pid)
		retval = -EDEADLK;
		
	/* Relay is not supposed to block or fail so we are going to use
	   the non-interruptible lock here.  */
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d locking for relay\n", task->pid);
	atomic_inc(&task->mutex.users);
	Srr_TaskLock();
#else
	Srr_SpinLock();
#endif
	
	if (!retval)
	{	
		/*-- Find source task -----------------------------------------*/
		source = SrrTm_Get(map, args->srcpid);
		
		/*-- Ensure source located and is not a proxy -----------------*/
		if (!source || IS_PROXY(source))
		{
			retval = -ESRCH;
		}
		else if (source->flags & _PPF_TERMING)
		{
			/*-- Source task appears to be terminating ! --------------*/
			PDEBUG ("Relay from terminating task %d to [tgt:%d]\n",
					source->pid, args->tgtpid);
			retval = -ESRCH;
		}
		/*-- Ensure that the source is really reply blocked -----------*/
		if (source->state != REPLY_BLOCKED)
		{
			PDEBUG ("Relay from source %d in state %s.  Should be %s\n",
					source->pid,
					SrrTaskState2Str(source, source->state),
					SrrTaskState2Str(source, REPLY_BLOCKED));
			retval = -EINVAL;
		}
	}
	if (!retval)
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d locking source %d for relay\n", task->pid, source->pid);
		atomic_inc(&source->mutex.users);
		Srr_WaitTaskLock(task, source);
#endif
		/*-- Find target task -----------------------------------------*/
		target = SrrTm_Get(map, args->tgtpid);
		
		/*-- Ensure target located and is not a proxy -----------------*/
		if (!target || IS_PROXY(target))
		{
			retval = -ESRCH;
		}
		else if (target->flags & _PPF_TERMING)
		{	
			/*-- Target task appears to be terminating ! --------------*/
			PDEBUG ("Relay from %d to terminating task [tgt:%d]\n",
					task->pid, target->pid);
			retval = -ESRCH;
		}
	}
	
	if (!retval)
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d locking target %d for relay\n", task->pid, target->pid);
		atomic_inc(&target->mutex.users);
		Srr_WaitTaskLock(task, target);
#endif
		/*-- Add source task to the target queue ----------------------*/
		SrrRq_Queue (target, source);
		
		/*-- Change state of source task to Send Blocked --------------*/
		source->state = SEND_BLOCKED;
		
		/*-- Set source send pid to the new target pid ----------------*/
		source->args.send.pid = target->pid;
		
		/*-- Wake target if its blocked on select() -------------------*/
		/* BUG: Could wake from poll and receive here? */
		if (SrrWakeFromPoll (target, READY))
		{
			/* nothing else to do */
		}
		/*-- Wake target if its recv blocked on src task, or 0 (any) --*/
		else if (target->state == RECV_BLOCKED)
		{
			if (target->args.recv.pid == source->pid ||
			    target->args.recv.pid == 0)
			{
				PDEBUG("Relay: Wake target %d\n", target->pid); 
				/*-- Set the waker to be that of the source task ------*/
				/*-- Do this first.  SrrTaskWake can schedule ---------*/
				target->waker = source->pid;
				SrrTaskWake (target, READY, __FUNCTION__);
			}
		}
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d unlocking target %d after relay\n", task->pid, target->pid);
		Srr_ReturnTaskLock(target);
		atomic_dec(&target->mutex.users);
	  LKDEBUG("%d unlocking source %d after relay\n", task->pid, source->pid);
		Srr_ReturnTaskLock(source);
		atomic_dec(&source->mutex.users);
#endif
	}
#ifdef TASK_SPECIFIC_LOCKS
	else
	{
		LKDEBUG("%d unlocking targetless source %d after relay", task->pid, source->pid);
		Srr_ReturnTaskLock(source);
		atomic_dec(&source->mutex.users);
	}
	LKDEBUG("%d unlocking after relay\n", task->pid);
  LKDEBUG("%d unlocking after relay\n", task->pid);
	atomic_dec(&task->mutex.users);
#else
	Srr_SpinUnlock();
#endif
	return retval;
}

/* 2000/02/17 - Alessandro Sala
 * The Readmsg() (pseudo-)system call.
 */

int SrrModReadmsg(SrrTm* map, SrrTask* task)
{
	SrrReadmsgArgs*	args = &task->args.readmsg;
	SrrTask*    	client;

	PDEBUG("Readmsg(pid %d, offset %d, size %d)\n",
									args->pid, args->offset, args->size);

		/* find client, confirm it's blocked on us */
	client = SrrTm_Get(map, args->pid);

	if(!client)
		return -ESRCH;

	if (IS_PROXY(client))
	{
		/* Proxies don't have REPLY_BLOCKED state */
		if (client->state != SEND_BLOCKED || client->target != task)
			return -ESRCH;
	}
	else if (client->state != REPLY_BLOCKED ||
			 client->args.send.pid != current->pid)
		return -ESRCH;

	/* limit read to client's send buffer size: I do this check here
	 * because Writemsg() could have extended the buffer beyond its
	 * initial size
	 */
	if (args->offset + args->size > client->buf.datalen)
	{
		if (args->offset > client->buf.datalen)
			return 0;

		args->size = client->buf.datalen - args->offset;
	}

	/* get sent data from offset */
	//Oops("SrrModReadMsg to len,offset %d,%d and length %d\n",
	//	 client->buf.datalen, args->offset, args->size);
		 
	return SrrCopyBufferToUserLinear (client, args->offset, task, args->msg,
									  args->size);
}

/* 2000/03/25 - Alessandro Sala
 * The Writemsg() (pseudo-)system call.
 */

int SrrModWritemsg(SrrTm* map, SrrTask* task)
{
	SrrWritemsgArgs*	args = &task->args.writemsg;
	SrrTask*		client;
	int				count;
	
	PDEBUG("Writemsg(pid %d, offset %d, size %d)\n",
									args->pid, args->offset, args->size);
	
		/* find client, confirm it's blocked on us */
	client = SrrTm_Get(map, args->pid);
	
	if(!client)
		return -ESRCH;
	
		/* writing to *your* proxy is a null op: always success ? */
	if(client->target == task)
		return args->size;
	
	if(client->state != REPLY_BLOCKED || client->args.send.pid != current->pid)
		return -ESRCH;
	
		/* if this is the first write, invalidate the old content of the buffer
		 * and ensure the buffer can hold the maximum reply size: this avoids
		 * multiple kmalloc/memcpy/kfree cycles while Writemsg() builds the
		 * answer writing deeper and deeper into the buffer
		 */
	if (!(client->flags & _PPF_KBUF_WRITTEN)) 
	{
		if (!SrrBuf_AdjustSize(&client->buf, client->args.send.rsize))
			return -ENOMEM;

		client->buf.datalen = 0;
		client->flags |= _PPF_KBUF_WRITTEN;
	}

	/* get reply data and write it at offset */
	count = SrrCopyUserLinearToBuffer (task, args->msg, args->size,
									   client, args->offset);
	if (count == -1)
		count = -ENOMEM;
	
	return (count);
}

/*
** Begin MX
** 2000/10/16	John Green
*/

/* Copy the data from a Send() call to the internal buffer of the
   sending task, taking care about Send vs Sendmx data format. */

int SrrCopySendDataToSender (SrrTask* sender)
{
	int		count = 0;

	/* Reset the send buffer length to zero before copying in new data */
	SrrBuf_Reset (&sender->buf);
	
	if (sender->last_ioctl == SRR_IOCSEND)
	{
		SrrSendArgs*			args = &sender->args.send;
		count = SrrCopyUserLinearToBuffer (sender, args->smsg, args->ssize,
										   sender, 0);
	}
	else if (sender->last_ioctl == SRR_IOCSENDMX)
	{
		SrrSendmxArgs*			args = &sender->args.sendmx;
		count = SrrCopyUserMxToBuffer (sender, args->smsgmx, args->sparts,
									   sender, 0);
	}
   else if (sender->last_ioctl == SRR_IOCSENDTIME)
   {
		SrrSendTimeoutArgs*	args = &sender->args.sendTimeout;
		count = SrrCopyUserLinearToBuffer (sender, args->smsg, args->ssize,
										   sender, 0);
	}
	else  if (sender->last_ioctl == SRR_IOCSENDMXTIME)
   {
		SrrSendmxTimeoutArgs*	args = &sender->args.sendmxTimeout;
		count = SrrCopyUserMxToBuffer (sender, args->smsgmx, args->sparts,
										   sender, 0);
   }
	return (count);
}

/*
 * Copy data from the sender's own data buffer back to the sender's
 * user space.
 */
int SrrCopyReplyDataToSender (SrrTask* sender)
{
	int		count = 0;
	
	if (sender->last_ioctl == SRR_IOCSEND)
	{
		SrrSendArgs*			args = &sender->args.send;
		count = SrrCopyBufferToUserLinear (sender, 0, sender,
										   args->rmsg, args->rsize);
	}
	else if (sender->last_ioctl == SRR_IOCSENDMX)
	{
		SrrSendmxArgs*			args = &sender->args.sendmx;
		count = SrrCopyBufferToUserMx (sender, 0, sender, args->rmsgmx,
									   args->rparts);
	} 
	else if (sender->last_ioctl == SRR_IOCSENDTIME )
   {
		SrrSendTimeoutArgs*			args = &sender->args.sendTimeout;
		count = SrrCopyBufferToUserLinear (sender, 0, sender,
										   args->rmsg, args->rsize);
	} 
	else if ( sender->last_ioctl == SRR_IOCSENDMXTIME)
   {
		SrrSendmxTimeoutArgs*			args = &sender->args.sendmxTimeout;
		count = SrrCopyBufferToUserMx (sender, 0, sender, args->rmsgmx,
									   args->rparts);
	}
	return (count);
}

/*
 * The Sendmx() system call.
 */

int SrrModSendmx(SrrTm* map, SrrTask* task)
{
	SrrSendmxArgs*	args = &task->args.sendmx;
	SrrTask*		server = NULL;
	SrrTask**		taskp;
    size_t			count;
	int				retval = 0;
	Srr_UseSpinLock();
	
	PDEBUG("Sendmx(pid %d, sparts %d, rparts %d)\n",
		   args->pid, args->sparts, args->rparts);

	/* Check for interrupt, and return immediately if interrupted.
	   The rest of this function does not "return" because it
	   needs to execute the Srr_SpinUnlock() at the end. */
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d locking for Sendmx\n", task->pid);
	atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing %d's lock to %d (sendmx)\n", task->pid, atomic_read(&task->mutex.users));
#ifdef LOCK_VIA_SPINLOCKS
  Srr_TaskLockInterruptible();
  if(unlikely(0))
  {
#else
  if (Srr_TaskLockInterruptible() != 0)
	{
#endif /* LOCK_VIA_SPINLOCKS */
		atomic_dec(&task->mutex.users);
    UDEBUG("Decrementing %d's lock to %d (sendmx)\n", task->pid, atomic_read(&task->mutex.users));
#else
	if (Srr_SpinLockInterruptible() != 0)
	{
#endif /* TASK_SPECIFIC_LOCKS */
		return (-EINTR);
	}

	if(args->pid == task->pid)
	{
		retval = -EDEADLK;
		server = NULL;
	}
	else
	{
		/* find server */
		server = SrrTm_Get(map, args->pid);
		if(!server || IS_PROXY(server))
		{
			retval = -ESRCH;
		}
		else if (server->flags & _PPF_TERMING)
		{
			KDEBUG ("Send from %d to terminating task %d\n",
					task->pid, server->pid);
			retval = -ESRCH;
		}
	}

	if (!retval)
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d locking %d for Sendmx\n", task->pid, server->pid);
		atomic_inc(&server->mutex.users);
    UDEBUG("Incrementing %d's lock to %d (sendmx)\n", server->pid, atomic_read(&server->mutex.users));
		Srr_WaitTaskLock(task, server);
#endif
		PDEBUG("Send: Server is %d\n", server->pid);
	
		/************** read from the send buffer ******************/
#ifndef DONT_COPY_MESSAGES	
		count = SrrCopySendDataToSender (task);
#endif
		if (count == -1)
		{
			/* This is a cheat, but the alternative was some ugly code
			   formatting. */
#ifdef TASK_SPECIFIC_LOCKS
			LKDEBUG("%d unlocking %d after Sendmx data copy error\n", task->pid, server->pid);
			Srr_ReturnTaskLock(server);
			atomic_dec(&server->mutex.users);
      UDEBUG("Decrementing %d's lock to %d (sendmx)\n", server->pid, atomic_read(&server->mutex.users));
			LKDEBUG("%d unlocking after Sendmx data copy error\n", task->pid);
			Srr_TaskUnlock();
			atomic_dec(&task->mutex.users);
      UDEBUG("Decrementing %d's lock to %d (sendmx)\n", task->pid, atomic_read(&task->mutex.users));
#else
			Srr_SpinUnlock ();
#endif
			return -ENOMEM;
		}
		else
		{
			/* queue ourselves */
			if (IS_PROXY(task))
				Oops ("Proxy %d performing a Send to %d\n",
					  task->pid, server->pid);
			SrrRq_Queue (server, task);
			PDEBUG("Send: Task is %d\n", task->pid);
		}

		/* Set the task state before waking the server.  Then wake the
		 * server and go to sleep ourselves.  The act of waking the
		 * server can cause rescheduling so between the wake and the
		 * sleep the server could run and change this task's state to
		 * reply blocked, or even to READY.
		 */
		task->state = SEND_BLOCKED;
		task->waker = -1;
		
		/* wake a server if its blocked on select() */
		if (SrrWakeFromPoll (server, READY))
		{
			/* Nothing else to do */
		}
		/* wake server if its recv blocked on this task, or 0 (any task) */
		else if(server->state == RECV_BLOCKED)
		{
			if(server->args.recv.pid == task->pid ||
			   server->args.recv.pid == 0)
			{
				PDEBUG("Send: wake server %d\n", server->pid); 
        SrrTaskWake(server, READY, __FUNCTION__);
			}
		}
	
		/*
		 * As far as I can tell, wake_up will NOT cause rescheduling, so we
		 * can assume that we are still executing atomically.  Bad idea if
		 * we either have interrupts happening, or we are running in an
		 * SMP kernel.
		 */
		SrrTaskSleep(task, SEND_BLOCKED, _spinflags, __FUNCTION__
#ifdef TASK_SPECIFIC_LOCKS
				, server);
#else
				);
#endif
		/* check for signal */
		if(signal_pending (current))
		{
			PDEBUG("Send() signal %#lx blocked %#lx\n",
				   signal_set(current), signal_blocked(current));
		
			if(task->state == SEND_BLOCKED)
			{
				/* deque ourselves */
				taskp = SrrRq_Find(server, task->pid);

				if (taskp && *taskp)
					SrrRq_Unlink(taskp);
				else
					Oops ("A: task %d is send blocked on %d, but %d not queued\n",
						  task->pid, server->pid, task->pid);
			}
		
			task->state = READY;
			retval = -EINTR;
		}

		/* if we're still blocked then the process we're blocked on has died */
		else if(task->state != READY)
		{
			/* sleep returned for unknown reasons... ;-( */
			if(task->state != RELEASED)
			{
				Oops("SendMX: send sleep failed: state %s on pid %d!\n",
					 SrrTaskState2Str(task, task->state), args->pid);

				/* clean up - deque ourselves (if necessary) and make READY */
				if(task->state == SEND_BLOCKED)
				{
					taskp = SrrRq_Find(server, task->pid);
					if (taskp && *taskp)
						SrrRq_Unlink(taskp);
					else
						Oops ("C: task %d is send blocked on %d, but %d not queued\n",
							  task->pid, server->pid, task->pid);
				}
				task->state = READY;
				retval = -EAGAIN;
			}
			else
			{
				task->state = READY;
				retval = -ESRCH;
			}
		}
		else
		{
			/************** write to reply buffer *****************/

			count = SrrCopyReplyDataToSender (task);
		}
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d unlocking %d after Sendmx\n", task->pid, server->pid);
		Srr_ReturnTaskLock(server);
		atomic_dec(&server->mutex.users);
    UDEBUG("Decrementing %d's lock to %d (sendmx)\n", server->pid, atomic_read(&server->mutex.users));
#endif
	}
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d unlocking after Sendmx\n", task->pid);
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
    UDEBUG("Decrementing %d's lock to %d (sendmx)\n", task->pid, atomic_read(&task->mutex.users));
#else
	Srr_SpinUnlock ();
#endif
	return (retval);
}

int SrrModSendmxTimeout(SrrTm* map, SrrTask* task)
{
	SrrSendmxTimeoutArgs*	args = &task->args.sendmxTimeout;
	SrrTask*		server = NULL;
	SrrTask**		taskp;
    size_t			count;
	int				retval = 0;
  int       sleepResult = 1;
  Srr_UseSpinLock();
	PDEBUG("Sendmx(pid %d, sparts %d, rparts %d)\n",
		   args->pid, args->sparts, args->rparts);

	/* Check for interrupt, and return immediately if interrupted.
	   The rest of this function does not "return" because it
	   needs to execute the Srr_SpinUnlock() at the end. */
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d locking for SendmxTimeout\n", task->pid);
	atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing %d's lock to %d (sendmxTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#ifdef LOCK_VIA_SPINLOCKS
  Srr_TaskLockInterruptible();
  if(unlikely(0))
  {
#else
  if (Srr_TaskLockInterruptible() != 0)
	{
#endif /* LOCK_VIA_SPINLOCKS */
		atomic_dec(&task->mutex.users);
    UDEBUG("Decrementing %d's lock to %d (sendmxTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#else
	if (Srr_SpinLockInterruptible() != 0)
	{
#endif /* TASK_SPECIFIC_LOCKS */
		return (-EINTR);
	}

	if(args->pid == task->pid)
	{
		retval = -EDEADLK;
		server = NULL;
	}
	else
	{
		/* find server */
		server = SrrTm_Get(map, args->pid);
		if(!server || IS_PROXY(server))
		{
			retval = -ESRCH;
		}
		else if (server->flags & _PPF_TERMING)
		{
			KDEBUG ("Send from %d to terminating task %d\n",
					task->pid, server->pid);
			retval = -ESRCH;
		}
	}

	if (!retval)
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d locking %d for SendmxTimeout\n", task->pid, server->pid);
		atomic_inc(&server->mutex.users);
    UDEBUG("Incrementing %d's lock to %d (sendmxTimeout)\n", server->pid, atomic_read(&server->mutex.users));
		Srr_WaitTaskLock(task, server);
#endif
		PDEBUG("Send: Server is %d\n", server->pid);
	
		/************** read from the send buffer ******************/
	
		count = SrrCopySendDataToSender (task);
		if (count == -1)
		{
			/* This is a cheat, but the alternative was some ugly code
			   formatting. */
#ifdef TASK_SPECIFIC_LOCKS
			LKDEBUG("%d unlocking %d after SendmxTimeout copy error\n", task->pid, server->pid);
			Srr_ReturnTaskLock(server);
    UDEBUG("Decrementing %d's lock to %d (sendmxTimeout)\n", server->pid, atomic_read(&server->mutex.users));
			atomic_dec(&server->mutex.users);
		  LKDEBUG("%d unlocking after SendmxTimeout copy error\n", task->pid);
			Srr_TaskUnlock();
			atomic_dec(&task->mutex.users);
    UDEBUG("Decrementing %d's lock to %d (sendmxTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#else
			Srr_SpinUnlock ();
#endif
			return -ENOMEM;
		}
		else
		{
			/* queue ourselves */
			if (IS_PROXY(task))
				Oops ("Proxy %d performing a Send to %d\n",
					  task->pid, server->pid);
			SrrRq_Queue (server, task);
			PDEBUG("Send: Task is %d\n", task->pid);
		}

		/* Set the task state before waking the server.  Then wake the
		 * server and go to sleep ourselves.  The act of waking the
		 * server can cause rescheduling so between the wake and the
		 * sleep the server could run and change this task's state to
		 * reply blocked, or even to READY.
		 */
		task->state = SEND_BLOCKED;
		task->waker = -1;
		
		/* wake a server if its blocked on select() */
		if (SrrWakeFromPoll (server, READY))
		{
			/* Nothing else to do */
		}
		/* wake server if its recv blocked on this task, or 0 (any task) */
		else if(server->state == RECV_BLOCKED)
		{
			if(server->args.recv.pid == task->pid ||
			   server->args.recv.pid == 0)
			{
				PDEBUG("Send: wake server %d\n", server->pid);
				SrrTaskWake(server, READY, __FUNCTION__);
			}
		}
	
		/*
		 * As far as I can tell, wake_up will NOT cause rescheduling, so we
		 * can assume that we are still executing atomically.  Bad idea if
		 * we either have interrupts happening, or we are running in an
		 * SMP kernel.
		 */
		sleepResult = SrrTaskSleepmxTimeout(task, SEND_BLOCKED, _spinflags, __FUNCTION__
#ifdef TASK_SPECIFIC_LOCKS
						, server);
#else
		);
#endif
		/* check for signal */
		if(signal_pending (current))
		{
			PDEBUG("Send() signal %#lx blocked %#lx\n",
				   signal_set(current), signal_blocked(current));
		
			if(task->state == SEND_BLOCKED)
			{
				/* deque ourselves */
				taskp = SrrRq_Find(server, task->pid);

				if (taskp && *taskp)
					SrrRq_Unlink(taskp);
				else
					Oops ("A: task %d is send blocked on %d, but %d not queued\n",
						  task->pid, server->pid, task->pid);
			}
		
			task->state = READY;
			retval = -EINTR;
		}
    
    /* If we were woken up by the wait_event_interruptible_timeout timer */
    else if(sleepResult == 0 && task->state == SEND_BLOCKED && 
        task->kern_task && !signal_pending(task->kern_task))
    {
      PDEBUG("SendTimeout() woken up by timer\n");
      if(task->state == SEND_BLOCKED)
      {
      taskp = SrrRq_Find(server, task->pid);

      if(taskp && *taskp)
        SrrRq_Unlink(taskp);
      else
        Oops ("A: task %d is send blocked on %d, but %d not queued (timeout)\n",
            task->pid, server->pid, task->pid);
      }
      task->state = READY;
      args->time = 0;
      retval = -ETIME;
    }

		/* if we're still blocked then the process we're blocked on has died */
		else if(task->state != READY)
		{
			/* sleep returned for unknown reasons... ;-( */
			if(task->state != RELEASED)
			{
				Oops("SendMX: send sleep failed: state %s on pid %d!\n",
					 SrrTaskState2Str(task, task->state), args->pid);

				/* clean up - deque ourselves (if necessary) and make READY */
				if(task->state == SEND_BLOCKED)
				{
					taskp = SrrRq_Find(server, task->pid);
					if (taskp && *taskp)
						SrrRq_Unlink(taskp);
					else
						Oops ("C: task %d is send blocked on %d, but %d not queued\n",
							  task->pid, server->pid, task->pid);
				}
				task->state = READY;
				retval = -EAGAIN;
			}
			else
			{
				task->state = READY;
				retval = -ESRCH;
			}
		}
		else
		{
			/************** write to reply buffer *****************/

			count = SrrCopyReplyDataToSender (task);
		}
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d unlocking %d after SendmxTimeout\n", task->pid, server->pid);
		Srr_ReturnTaskLock(server);
		atomic_dec(&server->mutex.users);
    UDEBUG("Decrementing %d's lock to %d (sendmxTimeout)\n", server->pid, atomic_read(&server->mutex.users));
#endif
	}

#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d unlocking after SendmxTimeout\n", task->pid);
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
#else
    UDEBUG("Decrementing %d's lock to %d (sendmxTimeout)\n", task->pid, atomic_read(&task->mutex.users));
	Srr_SpinUnlock ();
#endif

	return (retval);
}



int SrrRecvmxTimeout(SrrTm* map, SrrTask* task)
{
	SrrRecvmxTimeoutArgs*	args = &task->args.recvmxTimeout;
	SrrTask**		clientp = 0;
	SrrTask*		client = 0;
	size_t			count;
	int 				retval = 0;
	int         sleepResult = 1;
	Srr_UseSpinLock();
	
	PDEBUG("Recvmx(pid %d, parts %d)\n", args->pid, args->parts);

	/* Check for interrupt, and return immediately if interrupted.
	   The rest of this function does not "return" because it
	   needs to execute the Srr_SpinUnlock() at the end. */
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d locking for RecvmxTimeout\n", task->pid);
	atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing %d's lock to %d(RecvMxTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#ifdef LOCK_VIA_SPINLOCKS
  Srr_TaskLockInterruptible();
  if(unlikely(0))
  {
#else
	if (Srr_TaskLockInterruptible() != 0)
	{
#endif /* LOCK_VIA_SPINLOCKS */
		atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing %d's lock to %d\n", task->pid, atomic_read(&task->mutex.users));
#else
	if (Srr_SpinLockInterruptible() != 0)
	{
#endif /* TASK_SPECIFIC_LOCKS */
		return (-EINTR);
	}
	
	if(args->pid == current->pid)
	{
		retval = -EDEADLK;
	}
	/* make sure pid exists (if non-zero).  If the client exists
	   but is terminating, forget it. */
	else if (args->pid != 0 && !(client = SrrTm_Get(map, args->pid)))
	{
		retval = -ESRCH;
	}
	else if (client && (client->flags & _PPF_TERMING))
	{
		KDEBUG ("Receive by %d on terminating task %d\n",
				task->pid, client->pid);
		retval = -ESRCH;
	}
	else /* if (!retval) */
	{
		/* search for pid on our recv queue */
		if(args->pid == 0)	/* take the first one */
			clientp = SrrRq_First(task);
		else 				/* search for pid */
			clientp = SrrRq_Find(task, args->pid);
		client = *clientp;

		/* If we are creceived return with no client */

		if (!client && args->nowait)
		{
			retval = -ENOMSG;
		}
	}

	if (!retval)
	{
		/* wait for a client, if didn't find one */
		if(!client)
		{
			task->state = RECV_BLOCKED;
			task->waker = -1;
			sleepResult = SrrTaskSleepRqmxTimeout (task, RECV_BLOCKED, _spinflags, __FUNCTION__
#ifdef TASK_SPECIFIC_LOCKS
						, NULL);
#else
					);
#endif
			/* check for signal */
			if(signal_pending (current))
			{
				PDEBUG("Recv() signal %#lx blocked %#lx\n",
					   signal_set(current), signal_blocked(current));
				task->state = READY;
				retval = -EINTR;
			}
      if(sleepResult == 0){
        PDEBUG("RecvmxTimeout() awakened by timer.  Seting state to READY...\n");
        task->state = READY;
        args->time = 0;
        retval = -ETIME;
      }
		
    }

		if (!retval)
		{
			/* check if we were released because of a task death */
			if(task->state != READY)
			{
				/* sleep returned for unknown reasons... ;-( */
				if(task->state != RELEASED) {
					Oops("recv sleep failed: state %s on pid %d!\n",
						 SrrTaskState2Str(task, task->state), args->pid);
					task->state = READY;
/*					retval = -EAGAIN;*/
               retval = -EINTR;
				}
				else
				{
					task->state = READY;
					retval = -ESRCH;
				}
			}
		}

		if (!retval)
		{
			/* search for client, if not previously found */
			if(!client)
			{
				if(args->pid == 0)		/* take the first one */
					clientp = SrrRq_First(task);
				else					/* search for pid */
					clientp = SrrRq_Find(task, args->pid);
				/* clientp should always exist */
				if (!clientp)
				{
					retval = -ESRCH;
					Oops("Receive succeeded, but no sender: task=%d, state=%s, target=%d\n",
						 task->pid, SrrTaskState2Str(task, task->state), args->pid);
				}
				else
				{
					client = *clientp;
				}
			}
	
			if(!client)
			{
				/* This line generates an unused variable warning if we
				   compile without debugging enabled. */
				SrrTask* waker = task->waker ? SrrTm_Get(map, task->waker) : 0;
				/*
				 * This is not an Oops.  It is normal to be woken by a
				 * sender that is READY if the sender was in the SEND
				 * state when a signal caused it to drop out.
				 */
				PDEBUG("Receive(%d, ...)  woke by %d, which is now %s!\n",
					   args->pid,
					   task->waker,
					   waker ? SrrTaskState2Str(waker, waker->state) : "dead"
					   );
				// If we wanted to differentiate a broken SEND from a
				// failed receive on a non-zero PID, we would do this, but
				// it breaks the test program.  What does QNX4 do?
				// Well, QNX does not allow a Receive to fail here, so we
				// will return -EAGAIN and let the library handle it.
				if (waker)
					retval = -EAGAIN;
				else
					retval = -ESRCH;
			}
		}

		if (!retval)
		{
#ifdef TASK_SPECIFIC_LOCKS
			LKDEBUG("%d locking %d for RecvmxTimeout\n", task->pid, client->pid);
			atomic_inc(&client->mutex.users);
      UDEBUG("Incrementing %d's lock to %d (RecvmxTimeout)\n", client->pid, atomic_read(&client->mutex.users));
			Srr_WaitTaskLock(task, client);
#endif
			PDEBUG("Recv() from %d in state %s\n",
				   client->pid, SrrTaskState2Str(client, client->state));
	
			if(client->state != SEND_BLOCKED)
			{
				Oops("Recv() client %d is %s!\n",
					 client->pid, SrrTaskState2Str(client, client->state));
				retval = -EINTR;
			}
			else
			{
				/* We absolutely must dequeue a proxy before we get the user
				   data, or we have a window where a timer could fire again
				   when we reschedule due to user-to-kernel copying in the
				   SrrCopyBufferToUserLinear call below. */

				if(!IS_PROXY(client))
				{ /* deque task */
					if (clientp && *clientp)
						SrrRq_Unlink(clientp);
					else
						Oops ("D: task %d is send blocked on %d, but %d not queued\n",
							  client->pid, -1, client->pid);
					client->state = REPLY_BLOCKED;
				}
				else
				{
					/* deque proxy in any case, then re-queue if the count is
					   non-zero.  This gives other processes a fair-share
					   chance to send messages.  */

					if (clientp && *clientp)
					{
						/* Decrement the proxy count.  If it is not yet zero,
						   then re-queue the proxy onto the task.  Use the
						   Ndebug version of the queue function because we
						   know we are queuing a proxy that is not queued but
						   has a non-zero count. */

						if (!atomic_dec_and_test (&client->count))
						{
							SrrRq_UnlinkNdebug (clientp);
							SrrRq_QueueNdebug (task, client);
						}
						else
						{
							SrrRq_Unlink (clientp);
							client->state = READY;
						}
					}
					else
					{
						Oops ("E: task %d is send blocked on %d, but %d not queued\n",
							  client->pid, -1, client->pid);
						client->state = READY;
					}
				}
		
				/************** read from send buf *****************/

				count = SrrCopyBufferToUserMx (client, 0, task, args->msgmx,
											   args->parts);
	
				/********** end read from send buf less the return *************/

				/* client buffer not yet written to  - 2000/04/01 - Alessandro Sala */
				client->flags &= ~_PPF_KBUF_WRITTEN;
				retval = client->pid;
			}
#ifdef TASK_SPECIFIC_LOCKS
			LKDEBUG("%d unlocking %d after RecvmxTimeout\n", task->pid, client->pid);
			Srr_ReturnTaskLock(client);
			atomic_dec(&client->mutex.users);
      UDEBUG("Decrementing %d's lock to %d (RecvMxTimeout)\n", client->pid, atomic_read(&client->mutex.users));
#endif
		}
	}
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d unlocking after RecvmxTimeout\n", task->pid);
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing %d's lock to %d (RecvMxTimeout)\n", task->pid, atomic_read(&task->mutex.users));
#else
	Srr_SpinUnlock ();
#endif
	return retval;
}

/*
 * The Receive() (pseudo-)system call.
 */

int SrrRecvmx(SrrTm* map, SrrTask* task)
{
	SrrRecvmxArgs*	args = &task->args.recvmx;
	SrrTask**		clientp = 0;
	SrrTask*		client = 0;
    size_t			count;
    int				retval = 0;
	Srr_UseSpinLock();
	
	PDEBUG("Recvmx(pid %d, parts %d)\n", args->pid, args->parts);

	/* Check for interrupt, and return immediately if interrupted.
	   The rest of this function does not "return" because it
	   needs to execute the Srr_SpinUnlock() at the end. */
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d locking for Recvmx\n", task->pid);
	atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing %d's lock to %d (RecvMx)\n", task->pid, atomic_read(&task->mutex.users));
#ifdef LOCK_VIA_SPINLOCKS
  Srr_TaskLockInterruptible();
  if(unlikely(0))
  {
#else
  if (Srr_TaskLockInterruptible() != 0)
	{
#endif /* LOCK_VIA_SPINLOCKS */
		atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing %d's lock to %d (RecvMx)\n", task->pid, atomic_read(&task->mutex.users));
#else
	if (Srr_SpinLockInterruptible() != 0)
	{
#endif /* TASK_SPECIFIC_LOCKS */
		return (-EINTR);
	}
	
	if(args->pid == current->pid)
	{
		retval = -EDEADLK;
	}
	/* make sure pid exists (if non-zero).  If the client exists
	   but is terminating, forget it. */
	else if (args->pid != 0 && !(client = SrrTm_Get(map, args->pid)))
	{
		retval = -ESRCH;
	}
	else if (client && (client->flags & _PPF_TERMING))
	{
		KDEBUG ("Receive by %d on terminating task %d\n",
				task->pid, client->pid);
		retval = -ESRCH;
	}
	else /* if (!retval) */
	{
		/* search for pid on our recv queue */
		if(args->pid == 0)	/* take the first one */
			clientp = SrrRq_First(task);
		else 				/* search for pid */
			clientp = SrrRq_Find(task, args->pid);
		client = *clientp;

		/* If we are creceived return with no client */

		if (!client && args->nowait)
		{
			retval = -ENOMSG;
		}
	}

	if (!retval)
	{
		/* wait for a client, if didn't find one */
		if(!client)
		{
			task->state = RECV_BLOCKED;
			task->waker = -1;
			SrrTaskSleepRq (task, RECV_BLOCKED, _spinflags, __FUNCTION__
#ifdef TASK_SPECIFIC_LOCKS
						, NULL);
#else
				);
#endif
		
			/* check for signal */
			if(signal_pending (current))
			{
				PDEBUG("Recv() signal %#lx blocked %#lx\n",
					   signal_set(current), signal_blocked(current));
				task->state = READY;
				retval = -EINTR;
			}
		}

		if (!retval)
		{
			/* check if we were released because of a task death */
			if(task->state != READY)
			{
				/* sleep returned for unknown reasons... ;-( */
				if(task->state != RELEASED) {
					Oops("recv sleep failed: state %s on pid %d!\n",
						 SrrTaskState2Str(task, task->state), args->pid);
					task->state = READY;
					retval = -EAGAIN;
          //retval = -EINTR;
				}
				else
				{
					task->state = READY;
					retval = -ESRCH;
				}
			}
		}

		if (!retval)
		{
			/* search for client, if not previously found */
			if(!client)
			{
				if(args->pid == 0)		/* take the first one */
					clientp = SrrRq_First(task);
				else					/* search for pid */
					clientp = SrrRq_Find(task, args->pid);
				/* clientp should always exist */
				if (!clientp)
				{
					retval = -ESRCH;
					Oops("Receive succeeded, but no sender: task=%d, state=%s, target=%d\n",
						 task->pid, SrrTaskState2Str(task, task->state), args->pid);
				}
				else
				{
					client = *clientp;
				}
			}
	
			if(!client)
			{
				/* This line generates an unused variable warning if we
				   compile without debugging enabled. */
				SrrTask* waker = task->waker ? SrrTm_Get(map, task->waker) : 0;
				/*
				 * This is not an Oops.  It is normal to be woken by a
				 * sender that is READY if the sender was in the SEND
				 * state when a signal caused it to drop out.
				 */
				PDEBUG("Receive(%d, ...)  woke by %d, which is now %s!\n",
					   args->pid,
					   task->waker,
					   waker ? SrrTaskState2Str(waker, waker->state) : "dead"
					   );
				// If we wanted to differentiate a broken SEND from a
				// failed receive on a non-zero PID, we would do this, but
				// it breaks the test program.  What does QNX4 do?
				// Well, QNX does not allow a Receive to fail here, so we
				// will return -EAGAIN and let the library handle it.
				if (waker)
					retval = -EAGAIN;
				else
					retval = -ESRCH;
			}
		}

		if (!retval)
		{
#ifdef TASK_SPECIFIC_LOCKS
			LKDEBUG("%d locking %d for Recvmx\n", task->pid, client->pid);
			atomic_inc(&client->mutex.users);
      UDEBUG("Incrementing %d's lock to %d (RecvMx)\n", client->pid, atomic_read(&client->mutex.users));
			Srr_WaitTaskLock(task, client);
#endif
			PDEBUG("Recv() from %d in state %s\n",
				   client->pid, SrrTaskState2Str(client, client->state));
	
			if(client->state != SEND_BLOCKED)
			{
				Oops("Recv() client %d is %s!\n",
					 client->pid, SrrTaskState2Str(client, client->state));
				retval = -EINTR;
			}
			else
			{
				/* We absolutely must dequeue a proxy before we get the user
				   data, or we have a window where a timer could fire again
				   when we reschedule due to user-to-kernel copying in the
				   SrrCopyBufferToUserLinear call below. */

				if(!IS_PROXY(client))
				{ /* deque task */
					if (clientp && *clientp)
						SrrRq_Unlink(clientp);
					else
						Oops ("D: task %d is send blocked on %d, but %d not queued\n",
							  client->pid, -1, client->pid);
					client->state = REPLY_BLOCKED;
				}
				else
				{
					/* deque proxy in any case, then re-queue if the count is
					   non-zero.  This gives other processes a fair-share
					   chance to send messages.  */

					if (clientp && *clientp)
					{
						/* Decrement the proxy count.  If it is not yet zero,
						   then re-queue the proxy onto the task.  Use the
						   Ndebug version of the queue function because we
						   know we are queuing a proxy that is not queued but
						   has a non-zero count. */

						if (!atomic_dec_and_test (&client->count))
						{
							SrrRq_UnlinkNdebug (clientp);
							SrrRq_QueueNdebug (task, client);
						}
						else
						{
							SrrRq_Unlink (clientp);
							client->state = READY;
						}
					}
					else
					{
						Oops ("E: task %d is send blocked on %d, but %d not queued\n",
							  client->pid, -1, client->pid);
						client->state = READY;
					}
				}
		
				/************** read from send buf *****************/

				count = SrrCopyBufferToUserMx (client, 0, task, args->msgmx,
											   args->parts);
	
				/********** end read from send buf less the return *************/

				/* client buffer not yet written to  - 2000/04/01 - Alessandro Sala */
				client->flags &= ~_PPF_KBUF_WRITTEN;
				retval = client->pid;
			}
#ifdef TASK_SPECIFIC_LOCKS
			LKDEBUG("%d unlocking %d after Recvmx\n", task->pid, client->pid);
			Srr_ReturnTaskLock(client);
			atomic_dec(&client->mutex.users);
      UDEBUG("Decrementing %d's lock to %d (RecvMx)\n", client->pid, atomic_read(&client->mutex.users));
#endif
		}
	}
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d unlocking after Recvmx\n", task->pid);
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing %d's lock to %d (RecvMx)\n", task->pid, atomic_read(&task->mutex.users));
#else
	Srr_SpinUnlock ();
#endif
	return retval;
}

/*
 * The Reply() (pseudo-)system call.
 */

int SrrModReplymx(SrrTm* map, SrrTask* task)
{
	SrrReplymxArgs*	args = &task->args.replymx;
	SrrTask*		client;
    size_t			count;
    int			retval = 0;
	Srr_UseSpinLock();
	
	PDEBUG("Replymx(pid %d, parts %d)\n", args->pid, args->parts);
	
	/* Reply is not supposed to block or fail so we are going to use
	   the non-interruptible lock here.  */
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d locking for Replymx\n", task->pid);
	atomic_inc(&task->mutex.users);
  UDEBUG("Incrementing %d's lock to %d (ReplyMx)\n", task->pid, atomic_read(&task->mutex.users));
	Srr_TaskLock();
#else
	Srr_SpinLock();
#endif
	
	/* find client, confirm it's blocked on us */
	client = SrrTm_Get(map, args->pid);
	
	if(!client)
	{
		retval = -ESRCH;
	}
	else if(client->target == task)
	{
		/* replying to *your* proxy is a null op */
		retval = 0;
	}
	else if(client->state != REPLY_BLOCKED ||
			client->args.send.pid != current->pid)
	{
		retval = -ESRCH;
	}
	else
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d locking %d for Replymx\n", task->pid, client->pid);
		atomic_inc(&client->mutex.users);
    UDEBUG("Incrementing %d's lock to %d (ReplyMx)\n", client->pid, atomic_read(&client->mutex.users));
		Srr_WaitTaskLock(task, client);
#endif
		/************** write to reply buf less the return *****************/

		count = SrrCopyUserMxToBuffer (task, args->msgmx, args->parts,
									   client, 0);

		/************** end write to reply buf ******************/

		if (!retval)
		{
			/* wake up client */
			SrrTaskWake(client, READY, __FUNCTION__);
		}
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("%d unlocking %d after Replymx\n", task->pid, client->pid);
		Srr_ReturnTaskLock(client);
		atomic_dec(&client->mutex.users);
    UDEBUG("Decrementing %d's lock to %d (ReplyMx)\n", client->pid, atomic_read(&client->mutex.users));
#endif
	}
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("%d unlocking after Replymx\n", task->pid);
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
  UDEBUG("Decrementing %d's lock to %d (ReplyMx)\n", task->pid, atomic_read(&task->mutex.users));
#else
	Srr_SpinUnlock ();
#endif
	return retval;
}

/* 2000/02/17 - Alessandro Sala
 * The Readmsg() (pseudo-)system call.
 */

int SrrModReadmsgmx(SrrTm* map, SrrTask* task)
{
	SrrReadmsgmxArgs*   args = &task->args.readmsgmx;
	SrrTask*    client;
    size_t		offset;
    size_t		count;
	
	PDEBUG("Readmsgmx(pid %d, offset %d, parts %d)\n",
		   args->pid, args->offset, args->parts);
	
	/* find client, confirm it's blocked on us */
	client = SrrTm_Get(map, args->pid);
	
	if(!client)
		return -ESRCH;
	
	if (IS_PROXY(client)) {
		/* Proxies don't have REPLY_BLOCKED state */
		if ((client->state != SEND_BLOCKED) || (client->target != task)) {
			return -ESRCH;
		}
	} else if (   (client->state != REPLY_BLOCKED)
			   || (client->args.send.pid != current->pid)) {
		return -ESRCH;
	}
	
	/* limit read to client's send buffer size */
	/* get sent data from offset */
	
	/************** read from send buf *****************/

    offset = args->offset;
	count = SrrCopyBufferToUserMx (client, offset, task,
								   args->msgmx, args->parts);
	
    return count;
}

/* 2000/03/25 - Alessandro Sala
 * The Writemsg() (pseudo-)system call.
 */

int SrrModWritemsgmx(SrrTm* map, SrrTask* task)
{
	SrrWritemsgmxArgs*	args = &task->args.writemsgmx;
	SrrTask*		client;
    size_t		offset;
    size_t		count;
    size_t		i;
	
	PDEBUG("Writemsgmx(pid %d, offset %d, parts %d)\n",
		   args->pid, args->offset, args->parts);
	
	/* find client, confirm it's blocked on us */
	client = SrrTm_Get(map, args->pid);
	
	if(!client)
		return -ESRCH;
	
	/* writing to *your* proxy is a null op: always success ? */
	if(client->target == task) {
		count = 0;
		for (i = 0; i < args->parts; i++) {
			count += (int)((struct _mxfer_entry *)args->msgmx)[i].mxfer_len;
		}
		return count;
	}
	
	if(client->state != REPLY_BLOCKED || client->args.send.pid != current->pid)
		return -ESRCH;
	
	/* if this is the first write, invalidate the old content of the buffer
	 * and ensure the buffer can hold the maximum reply size: this avoids
	 * multiple kmalloc/memcpy/kfree cycles while Writemsg() builds the
	 * answer writing deeper and deeper into the buffer
	 */
	if (!(client->flags & _PPF_KBUF_WRITTEN)) 
	{
		if (!SrrBuf_AdjustSize(&client->buf, client->args.send.rsize))
			return -ENOMEM;
		
		client->buf.datalen = 0;
		client->flags |= _PPF_KBUF_WRITTEN;
	}
	
	/************** write to reply buf ***************************/

    offset = args->offset;
	count = SrrCopyUserMxToBuffer (task, args->msgmx, args->parts,
								   client, offset);

    return count;
}

/*
 ** End MX
 */

/*
 * The NameAttach() (pseudo-)system call.
 */
int SrrModNameAttach(SrrTm* map, SrrTask* task)
{
	char				*name;
	int					result;
	SrrNameAttachArgs	*args = &task->args.name_a;
	
	if(args->size == 0)
	{
		PDEBUG("NameAttach(size %d)\n", args->size);
		return -EINVAL;
	}
	
	name = str_cpy_u2k(args->name, args->size);
	
	PDEBUG("NameAttach(size %d, name %s)\n", args->size,
		   name ? name : "<0>");
	
	if(!name)
		return -ENOMEM;
	
	result = SrrNm_Attach(task, name);
	// SrrNm_Attach holds the name buffer after the attach call,
	// but only if the attach succeeded.
	if (result < 0)
		str_cpy_release(name);
	return (result);
}

/*
 * The NameLocate() (pseudo-)system call.
 */
int SrrModNameLocate(SrrTm* map, SrrTask* task)
{
	char		*name;
	int			result;
	SrrNameLocateArgs*	args = &task->args.name_l;
	
	if(args->size == 0)
	{
		PDEBUG("NameLocate(size %d)\n", args->size);
		return -EINVAL;
	}
	
	name = str_cpy_u2k(args->name, args->size);
	
	PDEBUG("NameLocate(size %d, name %s)\n", args->size,
		   name ? name : "<0>");
	
	if(!name)
		return -ENOMEM;
	
	result = SrrNm_Locate(map, name);
	str_cpy_release(name);
	return (result);
}

/*
 * The NameDetach() (pseudo-)system call.
 */
int SrrModNameDetach(SrrTm* map, SrrTask* task)
{
	SrrNameDetachArgs*	args = &task->args.name_d;
	
	PDEBUG("NameDetach(name_id %d)\n", args->name_id);
	
	return SrrNm_Detach(task, args->name_id);
}

/*
 * The ProxyAttach() (pseudo-)system call.
 */
int SrrModProxyAttach(struct SrrTm* map, struct SrrTask* task)
{
	SrrProxyAttachArgs*	args = &task->args.proxy_a;
	SrrTask* proxy = 0;
	SrrTask* target = 0;
	int result;
	Srr_UseSpinLock();

	PDEBUG("ProxyAttach(pid %d) task (pid %d)\n", args->pid, task->pid);
#ifdef TASK_SPECIFIC_LOCKS
	LKDEBUG("Locking %d for ProxyAttach\n", task->pid);
	atomic_inc(&task->mutex.users);
	Srr_TaskLock();
#else
	Srr_SpinLock();
#endif
	/* find pid to attach proxy to */
	if(args->pid == 0)
	{
		target = task;			// attach to ourselves
	}
	else
	{
		target = SrrTm_Get(map, args->pid);
	}
	if(!target)
	{
		result = -ESRCH;
	}
	/* make sure a proxy isn't being attached to a proxy */
	else if(target->target)
	{
		result = -ESRCH;
	}
	else
	{
#ifdef TASK_SPECIFIC_LOCKS
		if(target != task)
		{
			LKDEBUG("%d locking target %d for proxyAttach\n", task->pid, target->pid);
			atomic_inc(&target->mutex.users);
			Srr_WaitTaskLock(task, target);
		}
#endif
		/* allocate a task struct for the proxy */
		proxy = SrrTm_ProxyAlloc(map);
	
		if(!proxy)
		{
      			TDEBUG("Proxy allocation for %d failed!\n", target->pid);
      			result = -EAGAIN;
#ifdef TASK_SPECIFIC_LOCKS
			if(target != task)
			{
				LKDEBUG("%d unlocking target %d after proxyAlloc error in proxyAttach\n", task->pid, target->pid);
				Srr_ReturnTaskLock(target);
				atomic_dec(&target->mutex.users);
			}
			LKDEBUG("%d unlocking after proxyAlloc error in proxyAttach\n", task->pid);
			Srr_TaskUnlock();
			atomic_dec(&task->mutex.users);
      			TDEBUG("Return result for failure is %d.\n", result);
      			return result;
#endif
		}
		/* read the proxies mesg */

		else if(SrrCopyUserLinearToBuffer (task, args->data, args->nbytes,
										   proxy, 0) == -1)
		{
			SrrTm_Free(map, proxy->pid);
			result = -ENOMEM;
		}
		else
		{
			/* attach proxy to its target */
			proxy->target = target;
			proxy->args.pid = target->pid;
			result = proxy->pid;
		}
#ifdef TASK_SPECIFIC_LOCKS
		if(target != task)
		{
			LKDEBUG("%d unlocking target %d after proxyAttach\n", task->pid, target->pid);
			Srr_ReturnTaskLock(target);
			atomic_dec(&target->mutex.users);
		}
#endif
	}

#ifdef TASK_SPECIFIC_LOCKS
  	LKDEBUG("%d unlocking after proxyAttach\n", task->pid);
	Srr_TaskUnlock();
	atomic_dec(&task->mutex.users);
#else
	Srr_SpinUnlock();
#endif

	return result;
}

/*
 * The ProxyDetach() (pseudo-)system call.
 */
int SrrModProxyDetach(struct SrrTm* map, struct SrrTask* task)
{
	SrrProxyDetachArgs*	args = &task->args.proxy_d;
	int			result;
	
	result = SrrTm_Free(map, args->pid);
	return (result ? 0 : -ESRCH);
}

/*
 * The Trigger() (pseudo-)system call.
 */

EXPORT_SYMBOL( SrrDoTrigger );

int SrrDoTrigger (struct SrrTm* map, pid_t pid, int have_lock)
{
	SrrTask*	proxy;
	SrrTask*        target;
	int		retval = 0;
#ifdef PROXY_DEBUG
	SrrTask**	taskp;
#endif
	Srr_UseSpinLock();

	/* This is not supposed to block or fail so we are going to use
	   the non-interruptible lock here.  */
#ifndef TASK_SPECIFIC_LOCKS
	if (!have_lock)
		Srr_SpinLock();
#endif
	/* find proxy */
	proxy = SrrTm_Get(map, pid);

	if(!proxy || !IS_PROXY(proxy))
	{
		if (!proxy)
		{
			/* This is not an oops.  This can happen when a timer fires,
			   causing the bottom half handler to be scheduled.  Then,
			   the proxy is deleted before the bottom half handler runs,
			   which might happen because another task can be scheduled
			   ahead of the bottom half handler. (AST) */
			PDEBUG ("Proxy %d triggered but proxy does not exist\n", pid);
		}
		else
		{
			Oops ("Proxy triggered for task %d, but task is not a proxy\n",
				  pid);
		}
		retval = -ESRCH; // not a proxy!
	}
	else
	{
#ifdef TASK_SPECIFIC_LOCKS
		LKDEBUG("Locking proxy %d for DoTrigger\n", proxy->pid);
		atomic_inc(&proxy->mutex.users);
		UDEBUG("DoTrigger proxy users incremented to %d\n", atomic_read(&proxy->mutex.users));
    		Srr_GetTaskLock(proxy);
#endif
		target = proxy->target;
#ifdef SRR_TRACEPOINT 
    		trace_srripc_start_event(target->pid );
#endif
#ifdef TASK_SPECIFIC_LOCKS
    		if((target->pid != pid) &&  ( target != proxy)) // Don't lock if target == proxy 
    		{
      			LKDEBUG("Locking target %d for DoTrigger\n", target->pid);
      			atomic_inc(&target->mutex.users);
			UDEBUG("DoTrigger target users incremented to %d\n", atomic_read(&target->mutex.users));
      			Srr_GetTaskLock(target);
    		}
#endif
		/* queue ourselves if not already */
		if(proxy->state != SEND_BLOCKED)
		{
			SrrRq_Queue (target, proxy);
			proxy->state = SEND_BLOCKED;
		}

#ifdef PROXY_DEBUG
		else if (!(taskp = SrrRq_Find(target, pid)) || !*taskp)
		{
			Oops ("Proxy %d SEND_BLOCKED, count %d, not queued link=%p\n",
				  pid, proxy->count, proxy->recv_l);
		}
#endif
		atomic_inc(&proxy->count);
		proxy->firecount++;
		

		/* wake the receiver if blocked on poll (select) */
		if (target->flags & _PPF_POLLING)
		{
			//if (!have_lock)
			//	Srr_SpinUnlock ();
			//have_lock = 1;
#ifdef TASK_SPECIFIC_LOCKS
    			if((target->pid != pid) && (  target != proxy )) // Don't lock if task == proxy 
    			{
      				LKDEBUG("Unlocking target %d for PollWake in DoTrigger\n", target->pid);
      				Srr_ReturnTaskLock(target);
      				atomic_dec(&target->mutex.users);
      				UDEBUG("DoTrigger target users decremented to %d\n", atomic_read(&target->mutex.users));
    			}
    			LKDEBUG("Unlocking proxy %d for PollWake in DoTrigger\n", proxy->pid);
			Srr_ReturnTaskLock(proxy);
    			atomic_dec(&proxy->mutex.users);
    			UDEBUG("DoTrigger proxy users decremented to %d\n", atomic_read(&proxy->mutex.users));
#else
			Srr_SpinUnlock ();
#endif
			SrrWakeFromPoll (target, READY);

#ifndef TASK_SPECIFIC_LOCKS
			Srr_SpinLock ();
#endif

		}
		/* wake target if recv blocked on this or "any" */
		else if(target->state == RECV_BLOCKED)
		{
			if(target->args.recv.pid == proxy->pid ||
			   target->args.recv.pid == 0)
			{
				//if (!have_lock)
				//	Srr_SpinUnlock ();
				//have_lock = 1;
				/* Do not wake with the lock held.  The lock could be held
				   within the interrupt bottom half handler, which I suspect
				   of causing issues. */
#ifdef	TASK_SPECIFIC_LOCKS
			    	if((target->pid != pid) && ( target != proxy ))
    				{
					LKDEBUG("Unlocking target %d for taskWake in DoTrigger\n", target->pid);
					Srr_ReturnTaskLock(target);
        				atomic_dec(&target->mutex.users);
        				UDEBUG("DoTrigger target users decremented to %d\n", atomic_read(&target->mutex.users));
				}
    				LKDEBUG("Unlocking proxy %d for taskWake in DoTrigger\n", proxy->pid);
				Srr_ReturnTaskLock(proxy);
    				atomic_dec(&proxy->mutex.users);
    				UDEBUG("DoTrigger proxy users decremented to %d\n", atomic_read(&proxy->mutex.users));
#else
				Srr_SpinUnlock ();
#endif
				SrrTaskWake(target, READY, __FUNCTION__);

#ifndef TASK_SPECIFIC_LOCKS
				Srr_SpinLock ();
#endif
			}
		}
#ifdef TASK_SPECIFIC_LOCKS
	    	else
    		{
    			if((target->pid != pid) && ( target != proxy ))
    			{
      				LKDEBUG("Unlocking target %d after DoTrigger\n", target->pid);
		  		Srr_ReturnTaskLock(target);
		  		atomic_dec(&target->mutex.users);
      				UDEBUG("DoTrigger target users decremented to %d\n", atomic_read(&target->mutex.users));
    			}
    			LKDEBUG("Unlocking proxy %d after DoTrigger\n", proxy->pid);
			Srr_ReturnTaskLock(proxy);
			atomic_dec(&proxy->mutex.users);
    			UDEBUG("DoTrigger proxy users decremented to %d\n", atomic_read(&proxy->mutex.users));
    		}
#endif
	}
#ifndef TASK_SPECIFIC_LOCKS
	if (!have_lock)
		Srr_SpinUnlock ();
#endif
#ifdef  SRR_TRACEPOINT
        trace_srripc_stop_event( pid );
#endif	
	return retval;
}

int SrrModTrigger(struct SrrTm* map, struct SrrTask* task)
{
	int				retval = 0;

	SrrTriggerArgs* args = &task->args.trigger;
	
	PDEBUG("Trigger(proxy %d)\n", args->pid);
	retval = SrrDoTrigger (map, args->pid, 0);
	return (retval);
}

/*
 * The PFlags() (pseudo-)system call.
 */
int SrrPFlags(struct SrrTm* map, struct SrrTask* task)
{
	SrrPFlagsArgs* args = &task->args.pflags;
	
	PDEBUG("PFlags(flags %d, mask %d)\n", args->flags, args->mask);
	task->flags = (task->flags & ~args->mask) | args->flags;
	return 0;
}

/*
 * The Proc File System read call.
 */

/* This define is an attempt to guess what the maximum length of a single
 * sprintf() will be, before actually printing it to the proc buffer.
 * This is stupid, of course, what we really need are *nprintf() calls that
 * won't print more than n characters into the buffer, maybe in the next
 * version of "C"?
 */
#define MAXSINGLELINE 128

#define Sprintf(f, a...) \
			len += sprintf(&buf[len], f, ##a); \
			if(len >= (size - MAXSINGLELINE)) return len;

#define PQ(T) \
		do { \
			 while(T) { \
				Sprintf(" %d", T->pid); \
				T = T->recv_l; \
			} \
		 } while(0)

extern SrrPt srr_pt[NR_TIMERS];
extern int SrrTimerCount;
	 
#ifdef PROC_SRR_TARGET
int SrrDumpState(SrrTm* map, char* buf, size_t size, int *totlen)
{
	int	len = 0;
	
	Sprintf ("%-5.5s %-5.5s %-5.5s %-5.5s\n",
			 "PID", "STATE", "BLOCK", "FLAGS");
	SrrDumpSin (map, buf, size, &len);
	
	Sprintf ("\n");
	Sprintf ("%-5.5s %-5.5s %-5.5s %-5.5s\n",
			 "PROXY", "STATE", "BLOCK", "COUNT");
	SrrDumpProxy (map, buf, size, &len);
	
	Sprintf ("\n");
	Sprintf ("%-5.5s %-s\n",
			 "PID", "NAME");
	SrrDumpNames (map, buf, size, &len);

#ifndef NO_IRQ	
	Sprintf ("\n");
	Sprintf ("%-5.5s %-3.3s %-5.5s %-5.5s\n",
			 "PID", "IRQ", "PROXY", "FLAGS");
	SrrDumpIrq (map, buf, size, &len);
#endif

	Sprintf ("\n");
	Sprintf ("Version\n")
	SrrDumpVersion (map, buf, size, &len);

	Sprintf ("\n");
	Sprintf ("Tasks blocked:\n");
	SrrDumpTaskBlocks (map, buf, size, totlen);

	Sprintf ("\n");
	Sprintf ("Timers: %d fired total\n", SrrTimerCount);
	Sprintf ("Owner       Clock Flags Event    Interval    Usec     Arm    Fire Trigger Pending  Queued  QuFail\n");
	SrrDumpTimers (map, buf, size, totlen);
	
	*totlen += len;
	return *totlen;
}
#endif /* PROC_SRR_TARGET */

static int SrrBlockedOn (SrrTm* map, SrrTask* t)
{
	switch (t->state)
	{
	  case RELEASED:		return 0;
	  case READY:			return 0;
	  case RECV_BLOCKED:			// return (t->recv_l ? t->recv_l->pid : 0);
	  case REPLY_BLOCKED:
	  case SEND_BLOCKED:
		  return (t->args.pid);
	  default:				return 0;
	}
}

int SrrDumpProxy (SrrTm* map, char* buf, size_t size, int *totlen)
{
	int	len = *totlen;
	int i;
	
	for(i = 0; i < NR_TASKS; ++i)
	{
		SrrTask* 	t = map->tasks[i];
		if(!t || !IS_PROXY(t)) continue;
		
		Sprintf("%5d %5.5s %5d %5d %10d\n",
				t->pid, SrrTaskState2Str(t, t->state), t->target->pid,
				t->count.counter, t->firecount);
	}
	*totlen = len;
	return *totlen;
}

int SrrDumpIrq (SrrTm* map, char* buf, size_t size, int *totlen)
{
	int	len = *totlen;
#ifndef NO_IRQ
	int i, irq;
	
	for(i = 0; i < NR_TASKS; ++i)
	{
		SrrTask* 	t = map->tasks[i];
		if(!t || IS_PROXY(t)) continue;

		for (irq=0; irq<8; irq++)
		{
			if (t->IRQArray[irq])
			{
				Sprintf("%5d %3d %5d  0x%02x %10d %10d\n",
						t->pid,
						t->IRQArray[irq]->irq,
						t->IRQArray[irq]->proxy,
						t->IRQArray[irq]->flags,
						t->IRQArray[irq]->count.counter,
						t->IRQArray[irq]->bhcount.counter
						);
			}
		}
	}
#endif
	*totlen = len;
	return *totlen;
}

int SrrDumpTaskBlocks (SrrTm* map, char* buf, size_t size, int *totlen)
{
	int			len = *totlen;
	int			i;
	SrrTask 	*t, *block;

	for(i = 0; i < NR_TASKS; ++i)
	{
		t = map->tasks[i];
		
		if(t && t->recv_q)
		{
			Sprintf ("Task: %d\n", t->pid);
			
			for (block = t->recv_q; block; block = block->recv_l)
			{
				Sprintf ("  Blocked: %d\n", block->pid);
			}
		}
	}

	*totlen = len;
	return *totlen;
}

int SrrDumpTimers (SrrTm* map, char* buf, size_t size, int *totlen)
{
	int			len = *totlen;
	int			i;

	Sprintf ("Owner Flags  Event Fire-in(sec)  Repeat(sec)    Usec     Arm    Fire Trigger Pending  Queued  ReQueu BotHalf\n");
	for(i = 0; i < NR_TIMERS; ++i)
	{
		if (srr_pt[i].owner)
		{
			long int		remaining = srr_pt[i].expires - jiffies;
			struct timespec		fire, rpt;
			jiffies_to_timespec (remaining < 0 ? 0 : remaining, &fire);
			jiffies_to_timespec (srr_pt[i].interval, &rpt);

			/* show 4 decimal places for fractional seconds in case we
			   see 100us tick sizes in future.  2.6 uses 1ms tick size. */
			Sprintf ("%5d  0x%02x %6d %7ld.%04ld %7ld.%04ld %7lu %7d %7d %7d %7d %7d %7d %7d\n",
					srr_pt[i].owner, srr_pt[i].flags,
					 srr_pt[i].event.sigev_signo,
					 fire.tv_sec, fire.tv_nsec/100000,
					 rpt.tv_sec, rpt.tv_nsec/100000,
					 srr_pt[i].usec, srr_pt[i].arm_count, srr_pt[i].fire_count,
					 srr_pt[i].trigger_count, srr_pt[i].qcount.counter,
					 srr_pt[i].queued, srr_pt[i].queue_fail,
					 srr_pt[i].bottom_halfs
				);
		}
	}

	*totlen = len;
	return *totlen;
}

int SrrDumpVersion (SrrTm* map, char* buf, size_t size, int *totlen)
{
	int	len = *totlen;

	Sprintf("%s\n", VERSION);

	*totlen = len;

	return *totlen;
}

int SrrDumpSin(SrrTm* map, char* buf, size_t size, int* totlen)
{
	int	len = *totlen;
	int i;
	
	for(i = 0; i < NR_TASKS; ++i)
	{
		SrrTask* 	t = map->tasks[i];
		
		if(!t || IS_PROXY(t)) continue;
		
		Sprintf("%5d %5.5s %5d %5lx\n",
				t->pid,
				SrrTaskState2Str(t, t->state),
				SrrBlockedOn(map, t),
				t->flags);
	}
	*totlen = len;
	return *totlen;
}

int SrrDumpNames(SrrTm* map, char* buf, size_t size, int* totlen)
{
	int	len = *totlen;
	int	name_id;
	
	for(name_id = 0; name_id < SRR_MAX_NAMES; ++name_id)
	{
		if(srr_NameIdMap[name_id].task)
		{
			Sprintf ("%5d %s\n",
					 srr_NameIdMap[name_id].task->pid,
					 srr_NameIdMap[name_id].name);
		}
	}
	*totlen = len;
	return *totlen;
}
