/*
 * srr_taskLocks.h:
 *
 *  Linux Loadable Module Implementation of QNX Send/Receive/Reply
 *	message passing, and related function.
 *
 *	QNX is a trademark of QSSL, Ltd. <www.qnx.com>
 *
 * Copyright (C) 2000 Cogent Real-Time Systems Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software in a file called 'COPYING'; if not, write to:
 * 	Free Software Foundation, Inc.,
 * 	59 Temple Place, Suite 330,
 * 	Boston, MA 
 * 	02111-1307  USA
 *
 * $Id: srr_taskLocks.h 3612 2011-08-17 08:34:32Z meaton $
 */


/**
 * TASK_SPECIFIC_LOCKS -- Michael Eaton
 *
 * By default in v1.4.44, SRR has a single mutex which allows only one process to use 
 * the module's functions at a time.
 *
 * If TASK_SPECIFIC_LOCKS is defined in srr_taskLocks.h, the module will instead assign
 * mutexes to each process' entry in the task table.  Only processes that are attempting
 * to send/receive/reply to the same target process will lock each other out of SRR.
 * Essentially, whenever a process wishes to communicate with another process, it must
 * retreive its own lock and the target's lock.  Upon successfully obtaining both locks
 * the process will begin its operation.
 *
 * This has also required a large number of changes and locking to other areas of the
 * module.  The taskmap used to track relevant data about active processes in SRR now
 * has its own readers/writer lock (implemented rather crudely via a set of two atomic_t
 * counters).  Task deletion now waits until all tasks that depend upon the terminating
 * task's lock are completed, then deletes the task -- this is acheived by removing the
 * task structure from the task map and pushing it onto a delete queue, which is cleaned
 * every time another task is deleted.
 *
 * Task lock dependency is tracked by an atomic_t called "users", which is incremented at the
 * beginning of mutex-dependent tasks, and decremented at the end of mutex-dependent tasks.
 * If this value is not <= zero upon Task_delete, the task is pushed to the delete queue.
 * When the delete queue is being cleaned, if a task's "users" value is once again not <=
 * zero, the task will be passed over.  This is present to prevent a task taking a lock for
 * a function, having the lock be deleted during a sleep, then attempting to relock the
 * now-null lock.
 *
 * In situations where there are a small number of heavily interrelated processes, this patch
 * does produce a small performance hit, but when there are a large number of processes,
 * heavily related or no, this patch will produce large performance boosts.
 *
 * Current bugs include:
 *    -- *Timeout functions do not properly pass time remaining on timer back to the task
 *    struct
 *
 * Things to change/improve:
 *    -- Convert the current map locking scheme to an actual readers/writer lock rather than
 *    the way it's implemented now, which is kind of kludge-y. 
 **/

/**
 * TASK_SPECIFIC_LOCKS Locking Conventions
 *
 * When locking functions where you increment and decrement the structure's users value is 
 * extremely important -- because of the now-concurrent nature of the module, performing
 * this out of order can lead to null pointer exceptions and panics.
 *
 * The order for locking is:
 *  1.  Increment users
 *  2.  Lock task
 *
 * The order for unlocking is:
 *  1.  Unlock task
 *  2.  Decrement users.
 *
 * In Messaging functions (send/recieve/reply, DoTrigger, Relay, etc. -- essentially, those
 * found in srr_sys.c), the locking order is:
 *  1.  Calling process' task struct
 *  2.  Retreived structs in order
 *
 * The above order is subject to the function's purpose -- for example, in relay, there is no
 * need to lock the calling process' task struct as it is not modified.
 *
 * All other locks are fairly simple to deal with -- the map lock is a simple readers/writer
 * lock and the ProxyLock is only really dealt with in ProxyAlloc, and should not be an issue.
 **/


/**
 * Macros and their explanations -- Michael Eaton
 *
 * Fucntionality:
 *
 * LOCK_VIA_SPINLOCKS                           Replaces all mutexes from TASK_SPECIFIC_LOCKS with
 *                                              spinlocks.  Requires equivalent definition in srr_tt.h
 *
 * DONT_COPY_MESSAGES                           Removes all data-passing from messages:  Only allows
 *                                              recv_q and locking functionality (essentially wakeup
 *                                              sleep from calling functions)
 *
 * ALWAYS_PUSH_TO_QUEUE                         Always push a task to the delete queue on attempted
 *                                              deletion
 *
 * Debug output:
 *
 * DEBUG_MAP_LOCK                               Print owners/readers-writers on the maplock
 *
 * DEBUG_LOCKS                                  Print owners of TASK_SPECIFIC_LOCKS on taking and release
 *  
 * DEBUG_LOCK_USERS                             Print the value of a lock's USERS upon incementing
 *                                              and decrementing
 *
 * MX_DEBUG                                     Print MX struct sizes upon send and receive of a message
 *
 * DELETE_QUEUE_DEBUG                           Print the contents of the debug queue on addition and
 *                                              removal from the queue
 *
 * DEBUG_TASK_CREATIONDELETION                  Print a message on task creation and deletion
 *
 * HASH_ACCESS_DEBUG                            Print data relevant to task access by hash
 *
 * PRINT_FUNCTION_CALLS                         Allow usage of FCE and FCL to print a message on entering
 *                                              and leaving a function.
 *
 * PRINT_TIMEOUT_SLEEP_REMAINDER                What it says on the tin (for *Timeout functions).
 *
 * DEBUG_TIMERS                                 Various debug outputs related to Srr's timers
 *
 * PRINT_TASK_TABLE                             What it says on the tin.
 * 
 * DUMP_MESSAGES                                Perform a hexdump of message contents upon sending and
 *                                              upon reception. 
 **/


//#define TASK_SPECIFIC_LOCKS

/* Functionality macros */
//#define LOCK_VIA_SPINLOCKS
//#define DONT_COPY_MESSAGES
//#define ALWAYS_PUSH_TO_QUEUE

/* Debug macros */
//#define DEBUG_MAP_LOCK
//#define DEBUG_LOCKS
//#define DEBUG_LOCK_USERS
//#define MX_DEBUG
//#define DELETE_QUEUE_DEBUG
//#define DEBUG_TASK_CREATIONDELETION
//#define HASH_ACCESS_DEBUG
//#define PRINT_TASK_TABLE
//#define DEBUG_TIMERS
//#define PRINT_TIMEOUT_SLEEP_REMAINDER
//#define DUMP_MESSAGES

/* Task Locks */
#ifdef TASK_SPECIFIC_LOCKS
#define MSECS_TO_WAIT_ON_LOCK 1

/**
 * Macros for spinlock task locks
 **/
#ifdef LOCK_VIA_SPINLOCKS 
#define Srr_GetTaskLock(a)		\
      spin_lock_irqsave(&(a)->mutex.lock, (a)->mutex.flags)
      
#define Srr_ReturnTaskLock(a)	\
    spin_unlock_irqrestore(&(a)->mutex.lock, (a)->mutex.flags)
    
#define Srr_TaskLock()		    spin_lock_irqsave(&task->mutex.lock, task->mutex.flags)

#define Srr_TaskLockInterruptible()	spin_lock_irqsave(&task->mutex.lock, task->mutex.flags)

#define Srr_TaskUnlock()		  spin_unlock_irqrestore(&task->mutex.lock, task->mutex.flags)

#define Srr_CheckTaskLock(a)	mutex_is_locked(&(a)->mutex.mutex)

#define Srr_WaitTaskLock(a , b)	\
	LKPRINT("Task %d Waiting on %d's lock\n", (a)->pid, (b)->pid); \
	while (!spin_trylock(&(b)->mutex.lock)) \
	{ \
		Srr_ReturnTaskLock((a)); \
		wait_event_interruptible_timeout((a)->mutex.wq, 0, msecs_to_jiffies(MSECS_TO_WAIT_ON_LOCK)); \
	    Srr_GetTaskLock((a)); \
	} \
	LKPRINT("Task %d done waiting on %d's lock\n", (a)->pid, (b)->pid)

//Exclusively for use in Tm_Kill to prevent early death
#define Srr_WaitUntilUnlocked(a) \
		if(mutex_is_locked(&(a)->mutex.mutex)) \
			wait_event_interruptible((a)->mutex.wq, !mutex_is_locked(&(a)->mutex.mutex));

#else
/**
 * Macros for mutex task locks
 **/
#define Srr_GetTaskLock(a)		\
      mutex_lock(&(a)->mutex.mutex) 

#define Srr_ReturnTaskLock(a)	\
    mutex_unlock(&(a)->mutex.mutex)

#define Srr_TaskLock()		    Srr_GetTaskLock(task)

#define Srr_TaskLockInterruptible()	mutex_lock_interruptible(&task->mutex.mutex)

#define Srr_TaskUnlock()		  Srr_ReturnTaskLock(task)

#define Srr_CheckTaskLock(a)	mutex_is_locked(&(a)->mutex.mutex)

#define Srr_WaitTaskLock(a , b)	\
	LKPRINT("Task %d Waiting on %d's lock\n", (a)->pid, (b)->pid); \
	while (!mutex_trylock(&(b)->mutex.mutex)) \
	{ \
		Srr_ReturnTaskLock((a)); \
		wait_event_interruptible_timeout((a)->mutex.wq, !mutex_is_locked(&(b)->mutex.mutex), msecs_to_jiffies(MSECS_TO_WAIT_ON_LOCK)); \
	    Srr_GetTaskLock((a)); \
	} \
	LKPRINT("Task %d done waiting on %d's lock\n", (a)->pid, (b)->pid)

//Exclusively for use in Tm_Kill to prevent early death
#define Srr_WaitUntilUnlocked(a) \
		if(mutex_is_locked(&(a)->mutex.mutex)) \
			wait_event_interruptible((a)->mutex.wq, !mutex_is_locked(&(a)->mutex.mutex));

#endif /* LOCK_VIA_SPINLOCKS */

/**
 * Blank macros for monolithic locking.  Paranoid.
 **/
#else
#define Srr_GetTaskLock(a) 
#define Srr_ReturnTaskLock(a) 
#define Srr_TaskLock()
#define Srr_TaskLockInterruptible()
#define Srr_TaskUnlock()
#define Srr_CheckTaskLock(a)
#define Srr_WaitTaskLock(a , b)
#define Srr_WaitUntilUnlocked(a)
#endif /* TASK_SPECIFIC_LOCKS */


/* Taskmap lock */
#ifdef TASK_SPECIFIC_LOCKS
#ifndef DEBUG_MAP_LOCK
#define AddMapWriter(); \
  while(atomic_read(&map->map_writers) || atomic_read(&map->map_readers)) \
    wait_event_timeout(map->Map_wq, !atomic_read(&map->map_readers) && !atomic_read(&map->map_writers), jiffies_to_msecs(1)); \
  atomic_inc(&map->map_writers);
#define RemoveMapWriter(); \
  atomic_dec(&map->map_writers);
#define AddMapReader(); \
  while(atomic_read(&map->map_writers)) \
    wait_event_timeout(map->Map_wq, !atomic_read(&map->map_writers), jiffies_to_msecs(1)); \
  atomic_inc(&map->map_readers);
#define RemoveMapReader(); \
  atomic_dec(&map->map_readers);
#else
#define AddMapWriter(); \
  printk(KERN_DEBUG "Process %d attempting get writer MapLock in %s\n", current->pid, __FUNCTION__); \
  while(atomic_read(&map->map_writers) || atomic_read(&map->map_readers)) \
    wait_event_timeout(map->Map_wq, !atomic_read(&map->map_readers) && !atomic_read(&map->map_writers), jiffies_to_msecs(1)); \
  atomic_inc(&map->map_writers); \
  printk(KERN_DEBUG "Process %d has writer MapLock in %s\n", current->pid, __FUNCTION__);
#define RemoveMapWriter(); \
  printk(KERN_DEBUG "Process %d relinquishing writer MapLock in %s\n", current->pid, __FUNCTION__); \
  atomic_dec(&map->map_writers);
#define AddMapReader(); \
  printk("Process %d adding reader to MapLock in %s\n", current->pid, __FUNCTION__); \
  while(atomic_read(&map->map_writers)) \
    wait_event_timeout(map->Map_wq, !atomic_read(&map->map_writers), jiffies_to_msecs(1)); \
  atomic_inc(&map->map_readers); \
  printk("Process %d added reader to MapLock (now %d readers) in %s\n", current->pid, atomic_read(&map->map_readers), __FUNCTION__);
#define RemoveMapReader(); \
  printk("Process %d relinquishing reader on MapLock (%d readers) in %s\n", current->pid, atomic_read(&map->map_readers) - 1, __FUNCTION__); \
  atomic_dec(&map->map_readers);
#endif /* DEBUG_MAP_LOCK (R/W) */

/* ProxyAlloc lock.  Needed to prevent multiple proxies
 * from claiming the same pid */
#define ProxyAllocLock(); \
  mutex_lock(&proxyLock);
#define ProxyAllocUnlock(); \
  mutex_unlock(&proxyLock);

#else
#define AddMapWriter();
#define RemoveMapWriter();
#define AddMapReader();
#define RemoveMapReader();
#define ProxyAllocLock();
#define ProxyAllocUnlock();
#endif /* TASK_SPECIFIC_LOCKS */

/* Debug implementations */
#ifdef DUMP_MESSAGES
#define RMPRINT(message, length, fmt, args...); \
  do{ \
    printk(KERN_DEBUG "RecvPrint (%d) : " fmt, current->pid, ##args); \
    hexdump(message, length); \
  } while (0)
#define SMPRINT(message, length, fmt, args...); \
  do{ \
    printk(KERN_DEBUG "SendPrint (%d) : " fmt, current->pid, ##args); \
    hexdump(message, length); \
  } while (0)
#else
#define RMPRINT(message, length, fmt, args...);
#define SMPRINT(message, length, fmt, args...);
#endif

#ifdef PRINT_TIMEOUT_SLEEP_REMAINDER
#define TSPRINT(fmt, args...) \
  printk(KERN_DEBUG "%s (%d) : " fmt, __FUNCTION__ , current->pid, ##args);
#else
#define TSPRINT(fmt, args...)
#endif

#ifdef DEBUG_TIMERS
#define TDEBUG(fmt, args...) \
  printk(KERN_DEBUG "Timer Debug (%d) : " fmt, current->pid, ##args);
#else
#define TDEBUG(fmt, args...)
#endif

#ifdef PRINT_TASK_TABLE
#define TTP(); \
  printTm(map);
#else
#define TTP(); 
#endif

#ifdef DEBUG_TASK_CREATIONDELETION
#define TCDEBUG(fmt, args...); \
  printk(KERN_DEBUG "TaskCreate Debug (%d) : " fmt, current->pid, ##args);
#else
#define TCDEBUG(fmt, args...);
#endif

#ifdef HASH_ACCESS_DEBUG
#define HASHDEBUG(fmt, args...); \
  prink(KERN_DEBUG "Hash Access Debug (%d) : " fmt, current->pid, ##args);
#else
#define HASHDEBUG(fmt, args...);
#endif


#ifdef DELETE_QUEUE_DEBUG
#define DQDEBUG(fmt, args...); \
  printk(KERN_DEBUG "DeleteQueueDebug (%d): " fmt, current->pid, ##args);
#define PRINT_QUEUE_STATE(); \
  DQDEBUG("Printing Queue state...\n"); \
  DQDEBUG("Queue head is %d\n", (map->deleteQueue->head ? map->deleteQueue->head->task->pid : 0)); \
  struct DeleteNode *pnode = map->deleteQueue->head; \
  int pnodeCount = 0; \
  while(pnode) \
  { \
    DQDEBUG("Queue node %d is pid %d, has %d users\n", pnodeCount, pnode->task->pid, atomic_read(&pnode->task->mutex.users)); \
    pnode = pnode->next; \
    pnodeCount++; \
  }
#else
#define DQDEBUG(fmt, args...);
#define PRINT_QUEUE_STATE(); 
#endif

#if defined(MX_DEBUG) && defined(SRR_DEBUG)
#define MXDEBUG(fmt, args...); \
  if ( srr_debug_level > 3 ) \
  printk(KERN_DEBUG "Mx Debug (%d): " fmt, current->pid, ##args);
#else
#define MXDEBUG(fmt, args...);
#endif

#if defined(DEBUG_LOCKS) && defined(SRR_DEBUG) 
#define LKPRINT(fmt, args...); \
  if ( srr_debug_level > 4 ) \
  printk(KERN_DEBUG "Lock Debug(%d): " fmt, current->pid, ##args);
#define LKDEBUG(fmt, args...);  LKPRINT(fmt, ##args);
#else
#define LKPRINT(fmt, args...);
#define LKDEBUG(fmt, args...);
#endif

#ifdef DEBUG_LOCK_USERS
#define UDEBUG(fmt, args...) \
  printk(KERN_DEBUG "LockUsersDebug(%d):  " fmt, current->pid, ##args)
#else
#define UDEBUG(fmt, args...)
#endif


